// assemble_ex.cpp
//
// The example program in this file shows how to assemble a message by
// walking the document tree and serializing them to a buffer.
//
// The single command line parameter required is the name of a file that
// contains a single MIME document.
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mimepp/mimepp.h>


//
// This StringWrapper class wraps a DwString object.  It has an important
// purpose though: if we set a NULL string, then it will just compute the
// required length.  This allows us to allocate enough memory in advance
// for the DwString, and it means we can use the same code for measuring
// the length and for appending.
//
class StringWrapper
{
public:
    StringWrapper(DwString* str) : mString(str), mLength(0) {}
    void append(const DwString& str);
    void append(const char* cstr);
    int mLength;
    DwString* mString;
};

void StringWrapper::append(const DwString& str)
{
    if (mString != 0) {
        mString->append(str);
    }
    mLength += str.length();
}

void StringWrapper::append(const char* cstr)
{
    if (mString != 0) {
        mString->append(cstr);
    }
    mLength += strlen(cstr);
}

//
// Assembles all the components of the entity (message or body part) into
// the string wrapper.  Uses recursion if there are nested entities.
//
static void assembleTo_(DwEntity& aEntity, StringWrapper& aStr)
{
    //
    // Iterate over list of header fields.  Serialize each one.
    //
    DwHeaders& headers = aEntity.Headers();
    int numHeaderFields = headers.NumFields();
    for (int i=0; i < numHeaderFields; ++i) {
        DwField& field = headers.FieldAt(i);
        field.FieldBody()->Assemble();
        aStr.append(field.FieldNameStr());
        aStr.append(": ");
        if (field.FieldBody() != 0) {
            aStr.append(field.FieldBody()->AsString());
        }
        else /* if (field.FieldBody() == 0) */ {
            aStr.append(field.FieldBodyStr());
        }
        aStr.append(DW_EOL);
    }
    //
    // Append blank line, which marks the end of the header fields
    //
    aStr.append(DW_EOL);
    //
    // Get the content type
    //
    int contentType = DwMime::kTypeText;
    int contentSubtype = DwMime::kSubtypePlain;
    if (headers.HasContentType()) {
        contentType = headers.ContentType().Type();
        contentSubtype = headers.ContentType().Subtype();
    }
    DwBody& body = aEntity.Body();
    //
    // If it's a multipart type, get the boundary and preamble.  Then
    // iterate over each body part, and serialize each one by calling
    // this function recursively.  Append the boundary to separate body
    // parts.
    //
    if (contentType == DwMime::kTypeMultipart) {
        const DwString& boundary = headers.ContentType().Boundary();
        const DwString& preamble = body.Preamble();
        if (preamble.length() > 0) {
            aStr.append(preamble);
        }
        int numParts = body.NumBodyParts();
        for (int i=0; i < numParts; ++i) {
            DwBodyPart& part = body.BodyPartAt(i);
            aStr.append(DW_EOL "--");
            aStr.append(boundary);
            aStr.append(DW_EOL);
            assembleTo_(part, aStr);
        }
        aStr.append(DW_EOL "--");
        aStr.append(boundary);
        aStr.append("--" DW_EOL);
    }
    //
    // If it's a message/rfc822 type, then call this function recursively
    // to serialize the encapsulated message
    //
    else if (contentType == DwMime::kTypeMessage
        && contentSubtype == DwMime::kSubtypeRfc822) {
        DwMessage* message = body.Message();
        if (message != 0) {
             assembleTo_(*message, aStr);
        }
        else {
            //
            // If we arrive here, there something is wrong.  This should
            // recover.
            //
            body.Assemble();
            aStr.append(body.AsString());
        }
    }
    //
    // If it's a discrete body, then just append its string contents
    //
    else {
        aStr.append(body.AsString());
    }
}


void assembleTo(DwMessage& message, DwString& str)
{
    //
    // First pass: get the length
    //
    StringWrapper str1(0);
    assembleTo_(message, str1);
    int length = str1.mLength;
    //
    // Second pass: serialize everything to the string
    //
    str.reserve(length);
    StringWrapper str2(&str);
    assembleTo_(message, str2);
}


int main(int argc, char** argv)
{
    //
    // Get a message from a file.  The file name is the first commandline
    // parameter.
    //
    if (argc < 2) {
        printf("Yikes! No input file!\n");
        return 1;
    }
    FILE* in = fopen(argv[1], "r");
    if (in == NULL) {
        printf("Whoa! I can't open the file \"%s\".\n", argv[1]);
    }
    fseek(in, 0, SEEK_END);
    size_t fileLen = ftell(in);
    fseek(in, 0, SEEK_SET);
    DwString messageStr;
    messageStr.reserve(fileLen);
    char* buffer = (char*) malloc(8192);
    do {
        size_t count = fread(buffer, 1, 8192, in);
        messageStr.append(buffer, count);
    } while (! feof(in));
    free(buffer);
    buffer = 0;
    fclose(in);
    in = 0;
    //
    // Create a message object and parse the message
    //
    DwMessage message(messageStr);
    message.Parse();
    //
    // Just to be sure that we are assembling correctly, let's make a
    // change to the subject line
    //
    DwString subj = message.Headers().Subject().AsString();
    subj = DwString("FLAG: ") + subj;
    message.Headers().Subject().FromString(subj);
    //
    // Assemble the message
    //
    DwString str;
    assembleTo(message, str);
    //
    // If we want to clean up some memory and restore the tree of
    // DwMessageComponents to a nice, consistent state, we could parse
    // the message all over again, like this:
    //
    //     message.FromString(str);
    //     message.Parse();
    //
    // If we parse the message again, as shown above, then all the message
    // components will contain strings that share the same internal buffer
    // space.  If the message was originally created from parts, or was
    // modified in a significant way, then this should reduce the amount
    // of memory used.
    //

    //
    // Write the assembled message to a file
    //
    FILE* out = fopen("out.txt", "w");
    fwrite(str.data(), 1, str.length(), out);
    fclose(out);
    out = 0;

    return 0;   
}
