// doctree.cpp

#include <mimepp/mimepp.h>
#include <iostream>
#include <stdlib.h>

using std::cout;
using std::endl;

void PrintBodyPart(const DwString&, const DwString&, const DwBodyPart&);


// Print tree node that has no child nodes

void PrintDefault(const DwString& aIndent, const DwMessageComponent& aObject)
{
    // Print this node
    const char* s = aObject.ClassName();
    cout << aIndent << "+-" << s << endl;
}


// Print tree node for DwMailboxList object

void PrintMailboxList(const DwString& aIndent, const DwString& aIndentAdd,
    const DwMailboxList& aList)
{
    // Print this node
    cout << aIndent << "+-DwMailboxList" << endl;

    DwString indent = aIndent + aIndentAdd;

    // Print the DwMailbox objects
    int numMailboxes = aList.NumMailboxes();
    for (int j=0; j < numMailboxes; ++j) {
        const DwMailbox& mbox = aList.MailboxAt(j);
        PrintDefault(indent, mbox);
    }
}


// Print tree node for DwGroup object

void PrintGroup(const DwString& aIndent, const DwString& aIndentAdd,
    const DwGroup& aGroup)
{
    // Print this node
    cout << aIndent << "+-DwGroup" << endl;

    DwString indent = aIndent + aIndentAdd;

    // Print the DwMailboxList object
    DwString indentAdd = "  ";
    const DwMailboxList& mboxList = aGroup.MailboxList();
    PrintMailboxList(indent, indentAdd, mboxList);
}


// Print tree node for DwAddressList object

void PrintAddressList(const DwString& aIndent, const DwString& aIndentAdd,
    const DwAddressList& aList)
{
    // Print this node
    cout << aIndent << "+-DwAddressList" << endl;

    DwString indent = aIndent + aIndentAdd;

    // Print the DwAddress objects
    int numAddresses = aList.NumAddresses();
    for (int j=0; j < numAddresses; ++j) {
        const DwAddress& addr = aList.AddressAt(j);
        if (addr.IsGroup()) {
            DwString indentAdd;
            if (j < numAddresses - 1) {
                indentAdd = "| ";
            }
            else {
                indentAdd = "  ";
            }
            PrintGroup(indent, indentAdd, (const DwGroup&) addr);
        }
        else {
            PrintDefault(indent, addr);
        }
    }
}


// Print tree node for DwDispositionType object

void PrintDispositionType(const DwString& aIndent, const DwString& aIndentAdd,
    const DwDispositionType& aDtype)
{
    // Print this node
    cout << aIndent << "+-DwDispositionType" << endl;

    DwString indent = aIndent + aIndentAdd;

    // Print the DwParameter objects
    int numParameters = aDtype.NumParameters();
    for (int j=0; j < numParameters; ++j) {
        const DwParameter& param = aDtype.ParameterAt(j);
        PrintDefault(indent, param);
    }
}


// Print tree node for DwMediaType object

void PrintMediaType(const DwString& aIndent, const DwString& aIndentAdd,
    const DwMediaType& aMtype)
{
    // Print this node
    cout << aIndent << "+-DwMediaType" << endl;

    DwString indent = aIndent + aIndentAdd;

    // Print the DwParameter objects
    int numParameters = aMtype.NumParameters();
    for (int j=0; j < numParameters; ++j) {
        const DwParameter& param = aMtype.ParameterAt(j);
        PrintDefault(indent, param);
    }
}


// Print tree node for DwText object

void PrintText(const DwString& aIndent, const DwString& aIndentAdd,
    const DwText& aText)
{
    // Print this node
    cout << aIndent << "+-DwText" << endl;

    DwString indent = aIndent + aIndentAdd;

    // Print the DwEncodedWords objects
    int numEncodedWords = aText.NumEncodedWords();
    for (int j=0; j < numEncodedWords; ++j) {
        const DwEncodedWord& word = aText.EncodedWordAt(j);
        PrintDefault(indent, word);
    }
}


// Print tree node for DwField object

void PrintField(const DwString& aIndent, const DwString& aIndentAdd,
    const DwField& aField)
{
    // Print this node
    cout << aIndent << "+-DwField (" << aField.FieldNameStr() << ")" << endl;

    DwString indent = aIndent + aIndentAdd;
    DwString indentAdd = "  ";

    // Print the DwFieldBody object
    const DwFieldBody* fieldBody = aField.FieldBody();
    if (fieldBody != 0) {
        int id = fieldBody->ClassId();
        switch (id) {
        case DwMessageComponent::kCidAddressList:
            {
                const DwAddressList* addrListFieldBody = (DwAddressList*)
                    fieldBody;
                PrintAddressList(indent, indentAdd, *addrListFieldBody);
            }
            break;
        case DwMessageComponent::kCidDispositionType:
            {
                const DwDispositionType* dtypeFieldBody = (DwDispositionType*)
                    fieldBody;
                PrintDispositionType(indent, indentAdd, *dtypeFieldBody);
            }
            break;
        case DwMessageComponent::kCidMediaType:
            {
                const DwMediaType* mtypeFieldBody = (DwMediaType*)
                    fieldBody;
                PrintMediaType(indent, indentAdd, *mtypeFieldBody);
            }
            break;
        case DwMessageComponent::kCidMailboxList:
            {
                const DwMailboxList* mboxListFieldBody = (DwMailboxList*)
                    fieldBody;
                PrintMailboxList(indent, indentAdd, *mboxListFieldBody);
            }
            break;
        case DwMessageComponent::kCidText:
            {
                const DwText* textFieldBody = (DwText*) fieldBody;
                PrintText(indent, indentAdd, *textFieldBody);
            }
            break;
        default:
            {
                PrintDefault(indent, *fieldBody);
            }
            break;
        }
    }
}


// Print tree node for DwHeaders object

void PrintHeaders(const DwString& aIndent, const DwString& aIndentAdd,
    const DwHeaders& aHeaders)
{
    // Print this node
    cout << aIndent << "+-DwHeaders" << endl;

    DwString indent = aIndent + aIndentAdd;

    // Print all the DwField objects
    int numFields = aHeaders.NumFields();
    for (int j=0; j < numFields; ++j) {
        DwString indentAdd;
        if (j < numFields - 1) {
            indentAdd = "| ";
        }
        else {
            indentAdd = "  ";
        }
        const DwField& field = aHeaders.FieldAt(j);
        PrintField(indent, indentAdd, field);
    }
}


// Print tree node for DwHeaders object

void PrintBody(const DwString& aIndent, const DwString& aIndentAdd,
    const DwBody& aBody)
{
    // Print this node
    cout << aIndent << "+-DwBody" << endl;

    DwString indent = aIndent + aIndentAdd;

    // Print all the DwBodyPart objects, if this is a multipart message
    int numBodyParts = aBody.NumBodyParts();
    int j;
    for (j=0; j < numBodyParts; ++j) {
        DwString indentAdd;
        if (j < numBodyParts - 1) {
            indentAdd = "| ";
        }
        else {
            indentAdd = "  ";
        }
        const DwBodyPart& part = aBody.BodyPartAt(j);
        PrintBodyPart(indent, indentAdd, part);
    }
}


// Print tree node for DwBodyPart object

void PrintBodyPart(const DwString& aIndent, const DwString& aIndentAdd,
    const DwBodyPart& aPart)
{
    // Print this node
    DwString type = "text";
    DwString subtype = "plain";
    if (aPart.Headers().HasContentType()) {
        type = aPart.Headers().ContentType().TypeStr();
        subtype = aPart.Headers().ContentType().SubtypeStr();
    }
    cout << aIndent << "+-DwBodyPart (" << type << "/" << subtype
       << ")" << endl;

    DwString indent = aIndent + aIndentAdd;

    // Print the Headers object
    DwString indentAdd = "| ";
    PrintHeaders(indent, indentAdd, aPart.Headers());

    // Print the DwBody object
    indentAdd = "  ";
    PrintBody(indent, indentAdd, aPart.Body());
}


// Print tree node for DwMessage object

void PrintMessage(const DwString& aIndent, const DwString& aIndentAdd,
    const DwMessage& aMessage)
{
    // Print this node
    DwString type = "text";
    DwString subtype = "plain";
    if (aMessage.Headers().HasContentType()) {
        type = aMessage.Headers().ContentType().TypeStr();
        subtype = aMessage.Headers().ContentType().SubtypeStr();
    }
    cout << aIndent << "+-DwMessage (" << type << "/" << subtype
       << ")" << endl;

    DwString indent = aIndent + aIndentAdd;

    // Print the Headers object
    DwString indentAdd = "| ";
    PrintHeaders(indent, indentAdd, aMessage.Headers());

    // Print the DwBody object
    indentAdd = "  ";
    PrintBody(indent, indentAdd, aMessage.Body());
}


int main(int argc, char **argv)
{
    if (argc < 2) {
        printf("Yikes! No input file!\n");
        exit(EXIT_FAILURE);
    }
    FILE* in = fopen(argv[1], "rb");
    if (in == NULL) {
        printf("Whoa! I can't open the file \"%s\".\n", argv[1]);
        exit(EXIT_FAILURE);
    }
    fseek(in, 0, SEEK_END);
    size_t fileLen = ftell(in);
    fseek(in, 0, SEEK_SET);
    DwString messageStr;
    messageStr.reserve(fileLen);
    char* buffer = (char*) malloc(8192);
    do {
        size_t count = fread(buffer, 1, 8192, in);
        messageStr.append(buffer, count);
    } while (! feof(in));
    free(buffer);
    fclose(in);
    DwMessage message(messageStr);
    message.Parse();
    DwString indent = "";
    DwString indentAdd = "  ";
    PrintMessage(indent, indentAdd, message);
    return 0;
}
