// headers_ex.cpp
//
// This example program shows how to receive all the article headers from a
// newsgroup.
//
// Two command line parameters are required: the name of the news server
// and the name of the newsgroup.  The output lists a one line summary of
// all the articles in a newsgroup, where the summary lines are created by
// parsing the article headers.
//

#include <mimepp/mimepp.h>
#include <stdio.h>
#include <stdlib.h>

#if defined(WIN32)
#define snprintf _snprintf
#endif

void printSummary(int articleNum, DwHeaders& headers)
{
    // Print a summary on a single line, including the article number,
    // date, author, and subject

    DwString line;

    // Article Number
    char num[32];
    size_t numSize = sizeof(num);
    snprintf(num, numSize, "%-7d", articleNum);
    line = num;

    // Date
    line += " ";
    time_t caltime = headers.Date().AsCalendarTime();
    struct tm* stime = localtime(&caltime);
    char date[16];
    size_t dateSize = sizeof(date);
    snprintf(date, dateSize, "%4d-%02d-%02d", 1900+stime->tm_year,
        stime->tm_mon+1, stime->tm_mday);
    line += date;

    // Author
    line += " ";
    DwString from;
    if (headers.From().NumMailboxes() > 0) {
        if (headers.From().MailboxAt(0).FullName().length() > 0) {
            from = headers.From().MailboxAt(0).FullName();
        }
        else {
            from = headers.From().MailboxAt(0).LocalPart();
            from += "@";
            from += headers.From().MailboxAt(0).Domain();
        }
    }
    from.resize(30, ' ');
    line += from;

    // Subject
    line += " ";
    line += headers.Subject().AsString();

    line.resize(79, ' ');

    puts(line.c_str());
}


int main(int argc, char** argv)
{
    // Two command line parameters are required: the name of the news
    // server and the name of the newsgroup

    if (argc < 3) {
        printf("Usage: %s server newsgroup\n", argv[0]);
        exit(EXIT_FAILURE);
    }
    const char* server = argv[1];
    const char* newsgroup = argv[2];

    // Create an NNTP client object

    DwNntpClient nntp;

    // Open a connection to the news server. Check for a 2xx reply code,
    // which indicates success.

    int replyCode = nntp.Open(server, 119);
    if ((replyCode/100%10) == 2) {

        // Send a GROUP command, with the newsgroup name as a parameter.
        // Check for a 2xx reply code, which indicates success.

        replyCode = nntp.Group(newsgroup);
        if ((nntp.ReplyCode()/100%10) == 2) {

            // Parse the server's full response, which is of this form:
            //
            // c n f l s
            //
            // c = response code
            // n = estimated number of articles
            // f = first article number
            // l = last article number
            // s = name of the newsgroup
            // (the fields are separated by a space)
            //
            // We want to get the first article number

            const DwString& response = nntp.StatusResponse();
            int p = 3;
            p = response.find_first_not_of(" ", p);
            p = response.find_first_of(" ", p);
            int articleNum = strtol(response.c_str()+p, 0, 10);

            // Send a STAT command with the number of the first article.
            // This will set the "current article pointer".

            replyCode = nntp.Stat(articleNum);
            do {

                // Send a HEAD command to get the headers for the current
                // article.  Check for a 2xx reply code, which indicates
                // success.

                replyCode = nntp.Head();
                if ((replyCode/100%10) == 2) {

                    // If the HEAD command was successful, get the server's
                    // response, which contains the article headers.  Then
                    // parse the headers and print the summary information.

                    const DwString& textResponse = nntp.TextResponse();
                    DwHeaders headers(textResponse);
                    headers.Parse();
                    printSummary(articleNum, headers);
                }

                // Send a NEXT command to advance the "current article
                // pointer".  Check for a 2xx reply code, which indicates
                // success, and that there is yet another article to
                // retrieve.

                replyCode = nntp.Next();
                const char* resp = nntp.StatusResponse().c_str();
                articleNum = strtol(&resp[4], NULL, 10);
            } while ((replyCode/100%10) == 2);
        }

        // Send a QUIT command

        nntp.Quit();
    }
    else /* if ((replyCode/100%10) != 2) */ {
        printf("NNTP Error: %s\n", nntp.StatusResponse().c_str());
    }

    // Close the connection

    nntp.Close();

    return 0;
}
