// list_ex.cpp
//
// This example program shows how to list all the newsgroups available at
// a news server.  The list of newsgroups is written to an output file.
//
// The single command line parameter required is the name or address of the
// server.  The name of the output file is coded directly.
//

#include <mimepp/mimepp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


// This subclass of DwObserver handles callback from the NNTP protocol client
// object.  DwObserver is a interface class (a pure abstract class).

class MyObserver : public DwObserver
{
private:
    const DwNntpClient& mClient;
    FILE* mOut;
public:
    MyObserver(const DwNntpClient& client, FILE* out);
    virtual void Notify();
};


MyObserver::MyObserver(const DwNntpClient& client, FILE* out)
  : mClient(client)
{
    mOut = out;
}


// This is the callback function.

void MyObserver::Notify()
{
    // Get the last line received from the server, extract the information
    // we want, and write it to the output file

    const DwString& line = mClient.TextResponse();
    const char* s = line.c_str();

    // The format of the line is this (space separated):
    //    newsgroup-name last-msg-num first-msg-num is-posting-allowed

    int groupNameLen = strcspn(s, " ");
    char* ptr = (char*) &s[groupNameLen];
    int lastMsgNum = strtol(ptr, &ptr, 10);
    int firstMsgNum = strtol(ptr, &ptr, 10);
    fwrite(s, 1, groupNameLen, mOut);
    fprintf(mOut, " %d-%d\n", firstMsgNum, lastMsgNum);
}


int main(int argc, char** argv)
{
    // A single command line argument is required, which is the name or
    // address of the news server

    if (argc < 2) {
        printf("Usage: %s server\n", argv[0]);
        exit(EXIT_FAILURE);
    }
    const char* server = argv[1];

    // Open a file for output.  The list of newsgroups is written to this
    // file.

    const char* filename = "newsgroup-list.txt";
    FILE* out = fopen(filename, "w");
    if (out != 0) {

        // Create an NNTP client object

        DwNntpClient nntp;

        // Open a connection to the server.  Check for a 2xx reply code,
        // which indicates success.

        nntp.Open(server, 119);

        if ((nntp.ReplyCode()/100%10) == 2) {

            // Set our observer object, which handles callbacks from the
            // NNTP client object

            MyObserver observer(nntp, out);
            nntp.SetObserver(&observer);

            // Send a LIST command.  While this command executes, our
            // observer object handles callbacks, one callback for each
            // line received from the server.

            nntp.List();

            // Check for a 2xx reply code, which indicates success

            if ((nntp.ReplyCode()/100%10) == 2) {
                printf("Success!\n");
            }

            // Send a QUIT command

            nntp.Quit();
        }
        else /* if ((nntp.ReplyCode()/100%10) != 2) */ {
            printf("NNTP Error: %s\n", nntp.StatusResponse().c_str());
        }

        // Close the connection

        nntp.Close();
    }
    else /* if (out == 0) */ {
        printf("Failure opening output file\n");
    }
    return 0;
}
