// post_ex.cpp
//
// This example program shows how to post an article to a newsgroup.
//
// The single command line parameter required is the name or address of the
// server.  The actual article posted, the newsgroup to post to, and other
// parameters are written directly in the code.  You may want to make
// changes to these parameters.
//

#include <mimepp/mimepp.h>
#include <stdlib.h>


void formatMessage(
    const char* from,
    const char* subject,
    const char* newsgroup,
    const char* body,
    DwString& messageStr)
{
    messageStr = "";

    // Create a DwMessage object

    DwMessage msg(messageStr);

    // Set the message header fields

    msg.Headers().From().FromString(from);
    msg.Headers().Date().FromCalendarTime(time(0));
    msg.Headers().Newsgroups().FromString(newsgroup);
    msg.Headers().Subject().FromString(subject);
    msg.Headers().MessageId().CreateDefault();
    msg.Headers().Path().FromString("");

    // Set the message body

    msg.Body().FromString(body);

    // Serialize the message

    msg.Assemble();

    // And return the message as a string

    messageStr = msg.AsString();
}


int postMessage(const char* server, const DwString& messageStr)
{
    int retval = -1;

    // Post a message

    // The required steps are these:
    //   1. Connect to NNTP server
    //   2. Send a POST command
    //   3. Check that we got the 340 reply code (or just 3xx)
    //   4. Send the message, ending with CR LF '.' CR LF character sequence
    //   5. Send a QUIT command
    //   6. Close the connection

    // In order to post to a certain newsgroup, the message itself must
    // contain the name of the newsgroup in its "Newsgroups" header field.
    // There is no command in the NNTP protocol to tell the server which
    // newsgroup to post the message to.

    // Create an NNTP client object

    DwNntpClient nntp;

    // Open a connection to the server. Check for a 2xx reply code, which
    // indicates success.

    int replyCode = nntp.Open(server, 119);
    if ((replyCode/100%10) == 2) {

        // Send a POST command.  Check for a 3xx reply code, which
        // indicates success.

        replyCode = nntp.Post();
        if ((replyCode/100%10) == 3) {

            // Send the message.  Check for a 2xx reply code, which
            // indicates success.

            replyCode = nntp.SendData(messageStr);
            if (replyCode/100%10 == 2) {
                printf("Success!\n");
                retval = 0;
            }
            else /* if (replyCode/100%10 != 2) */ {
                printf("NNTP Error: %s\n", nntp.StatusResponse().c_str());
            }
        }
        else /* if ((replyCode/100%10) != 3) */ {
            printf("NNTP Error: %s\n", nntp.StatusResponse().c_str());
        }

        // Send a QUIT command

        nntp.Quit();
    }
    else if (replyCode == 0) {
        printf("Can't connect to host %s\n", server);
    }
    else {
        printf("NNTP Error: %s\n", nntp.StatusResponse().c_str());
    }

    // Close the connection

    nntp.Close();

    return retval;
}


int main(int argc, char** argv)
{
    // A single command line argument is required, which is the name or
    // address of the news server

    if (argc < 2) {
        printf("Usage: %s server\n", argv[0]);
        exit(EXIT_FAILURE);
    }

    int retval = 1;

    const char* server = argv[1];

    // Other parameters are coded directly

    const char* from = "snerd@testonly.org (Mortimer Snerd)";
    const char* subject = "Test message from Mort";
    const char* newsgroup = "test";
    const char* body = "It worked!\n";

    // Format the message

    DwString messageStr;
    formatMessage(from, subject, newsgroup, body, messageStr);

    // Post the message

    int err = postMessage(server, messageStr);
    if (! err) {
        retval = 0;
    }

    return retval;
}
