// smtp.cpp

#include "smtp.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

static void AppendMailboxes(DwAddressList& aSrcList, DwMailboxList& aDestList);


Smtp::Smtp(const char* aServer)
{
    mServer = strdup(aServer);
    mErrorHandler = 0;
}


Smtp::~Smtp()
{
    if (mServer != 0) {
        free(mServer);
        mServer = 0;
    }
    if (mErrorHandler != 0) {
        delete mErrorHandler;
        mErrorHandler = 0;
    }
}


void Smtp::SetErrorHandler(SmtpErrorHandler* aHandler)
{
    mErrorHandler = aHandler;
}


int Smtp::SendMessage(
    DwMessage& aMessage)
{
    DwHeaders& headers = aMessage.Headers();

    // Get the sender.  If there is a 'sender' header field, use that.
    // Otherwise, get the 'from' header field.

    DwMailbox sender;

    if (headers.HasSender()) {
        sender = headers.Sender();
    }
    else /* if (! headers.HasSender()) */ {
        if (headers.HasFrom()) {
            DwMailboxList from = headers.From();
            int numMailboxes = from.NumMailboxes();
            if (numMailboxes > 0) {
                sender = from.MailboxAt(0);
            }
        }
    }

    // Get all 'To' recipients

    DwMailboxList recipients;

    if (headers.HasTo()) {
        DwAddressList& to = headers.To();
        AppendMailboxes(to, recipients);
    }

    // Get all 'Cc' recipients

    if (headers.HasCc()) {
        DwAddressList& cc = headers.Cc();
        AppendMailboxes(cc, recipients);
    }

    // Get all 'Bcc' recipients

    if (headers.HasBcc()) {
        DwAddressList& bcc = headers.Bcc();
        AppendMailboxes(bcc, recipients);
    }

    // Remove all the 'Bcc' header fields

    DwMessage messageNoBcc(aMessage);
    if (messageNoBcc.Headers().HasBcc()) {
        DwHeaders& headers = messageNoBcc.Headers();
        // We decrement the indexes, so that we can safely remove
        // addresses from the list as we iterate
        int numFields = headers.NumFields();
        for (int i=numFields-1; i > 0; --i) {
            const DwString& fieldName = headers.FieldAt(i).FieldNameStr();
            if (DwStrcasecmp(fieldName, "bcc") == 0) {
                DwField* field = headers.RemoveFieldAt(i);
                delete field;
            }
        }
    }
    messageNoBcc.Assemble();
    DwString messageStr = messageNoBcc.AsString();

    int retval = SendMessage(messageStr, sender, recipients);
    return retval;
}


int Smtp::SendMessage(
    const DwString& aMessageStr,
    DwMailbox& aSender,
    DwMailboxList& aRecipients)
{
    int retval = UNKNOWN;

    // Convert end of line characters to CR LF

    DwString messageStr;
    DwToCrLfEol(aMessageStr, messageStr);

    // Open connection to server

    DwSmtpClient smtpClient;
    int replyCode = smtpClient.Open(mServer);

    // Check for 2xx reply code, which indicates success

    if (replyCode / 100 % 10 == 2) {
        retval = SendHeloCommand(smtpClient, messageStr, aRecipients, aSender);
    }
    else /* if (replyCode / 100 % 10 != 2) */ {
        // <error>
        if (mErrorHandler != 0) {
            const char* replyMessage = smtpClient.Response().c_str();
            mErrorHandler->Error("", replyCode, replyMessage, "");
        }
        retval = FAILURE;
    }
    smtpClient.Close();

    return retval;
}


int Smtp::SendHeloCommand(
    DwSmtpClient& aSmtpClient,
    const DwString& aMessageStr,
    DwMailboxList& aRecipients,
    DwMailbox& aSender)
{
    int retval = UNKNOWN;

    // Send HELO command

    int replyCode = aSmtpClient.Helo();

    // Check for 2xx reply code, which indicates success

    if (replyCode / 100 % 10 == 2) {
        retval = SendMailCommand(aSmtpClient, aMessageStr, aRecipients,
            aSender);
    }
    else /* if (replyCode / 100 % 10 != 2) */ {
        // <error>
        if (mErrorHandler != 0) {
            const char* replyMessage = aSmtpClient.Response().c_str();
            mErrorHandler->Error("HELO", replyCode, replyMessage, "");
        }
        retval = FAILURE;
    }
    return retval;
}


int Smtp::SendMailCommand(
    DwSmtpClient& aSmtpClient,
    const DwString& aMessageStr,
    DwMailboxList& aRecipients,
    DwMailbox& aSender)
{
    int retval = UNKNOWN;

    // Parse the address to be sure we get the local part and
    // domain part correctly
    aSender.Assemble();
    aSender.Parse();
    DwString sender = aSender.LocalPart();
    sender += "@";
    sender += aSender.Domain();

    // Send MAIL command

    int replyCode = aSmtpClient.Mail(sender.c_str());

    // Check for 2xx reply code, which indicates success

    if (replyCode / 100 % 10 == 2) {
        retval = SendRcptCommand(aSmtpClient, aMessageStr, aRecipients);
    }
    else /* if (replyCode / 100 % 10 != 2) */ {
        // <error>
        if (mErrorHandler != 0) {
            const char* replyMessage = aSmtpClient.Response().c_str();
            mErrorHandler->Error("MAIL", replyCode, replyMessage, "");
        }
        retval = FAILURE;
    }
    return retval;
}


int Smtp::SendRcptCommand(
    DwSmtpClient& aSmtpClient,
    const DwString& aMessageStr,
    DwMailboxList& aRecipients)
{
    int retval = UNKNOWN;

    int successCount = 0;
    int numRecipients = aRecipients.NumMailboxes();
    for (int i=0; i < numRecipients; ++i) {
        DwMailbox& mbox = aRecipients.MailboxAt(i);
        mbox.Parse();
        mbox.Assemble();
        DwString rcpt = mbox.LocalPart();
        rcpt += "@";
        rcpt += mbox.Domain();

        // Send RCPT command

        int replyCode = aSmtpClient.Rcpt(rcpt.c_str());

        // Check for 2xx reply code, which indicates success

        if (replyCode / 100 % 10 == 2) {
            ++successCount;
        }
        else /* if (replyCode / 100 % 10 != 2) */ {
            // <error>
            if (mErrorHandler != 0) {
                const char* replyMessage = aSmtpClient.Response().c_str();
                mErrorHandler->Error("RCPT", replyCode, replyMessage,
                    rcpt.c_str());
            }
        }
    }
    if (successCount > 0) {

        // Send DATA command

        int retval2 = SendDataCommand(aSmtpClient, aMessageStr);
        if (retval2 == SUCCESS) {
            if (successCount == numRecipients) {
                retval = SUCCESS;
            }
            else /* if (successCount < numRecipients) */ {
                retval = PARTIAL_SUCCESS;
            }
        }
        else /* if (retval2 != SUCCESS) */ {
            retval = retval2;
        }
    }
    else /* if (successCount == 0) */ {
        retval = FAILURE;
    }
    return retval;
}


int Smtp::SendDataCommand(
    DwSmtpClient& aSmtpClient,
    const DwString& aMessageStr)
{
    int retval = UNKNOWN;

    // Send DATA command

    int replyCode = aSmtpClient.Data();

    // Check for 3xx reply code, which indicates success

    if (replyCode / 100 % 10 == 3) {

        // Send the message

        int replyCode = aSmtpClient.SendData(aMessageStr);
        if (replyCode / 100 % 10 == 2) {
            retval = SUCCESS;
        }
        else /* if (replyCode / 100 % 10 != 2) */ {
            // <error>
            if (mErrorHandler != 0) {
                const char* replyMessage = aSmtpClient.Response().c_str();
                mErrorHandler->Error("DATA", replyCode, replyMessage,"");
            }
            retval = FAILURE;
        }
    }
    else /* if (replyCode / 100 % 10 != 3) */ {
        // <error>
        if (mErrorHandler != 0) {
            const char* replyMessage = aSmtpClient.Response().c_str();
            mErrorHandler->Error("DATA", replyCode, replyMessage, "");
        }
        retval = FAILURE;
    }
    return retval;
}


static void AppendMailboxes(DwAddressList& aSrcList, DwMailboxList& aDestList)
{
    int numAddresses = aSrcList.NumAddresses();
    for (int i=0; i < numAddresses; ++i) {
        DwAddress& addr = aSrcList.AddressAt(i);
        if (addr.IsMailbox()) {
            DwMailbox* pMailbox = (DwMailbox*) addr.Clone();
            aDestList.AddMailbox(pMailbox);
        }
        else /* if (addr.IsGroup()) */ {
            DwGroup& group = (DwGroup&) addr;
            DwMailboxList& list = group.MailboxList();
            int n = list.NumMailboxes();
            for (int j=0; j < n; ++j) {
                DwMailbox& mailbox = list.MailboxAt(j);
                DwMailbox* pMailbox = (DwMailbox*) mailbox.Clone();
                aDestList.AddMailbox(pMailbox);
            }
        }
    }
}
