/*=============================================================================
    multi_pass.hpp

    Spirit V1.2
    URL: http://spirit.sourceforge.net/

    Copyright (c) 2001, Daniel C. Nuffer

    This software is provided 'as-is', without any express or implied
    warranty. In no event will the copyright holder be held liable for
    any damages arising from the use of this software.

    Permission is granted to anyone to use this software for any purpose,
    including commercial applications, and to alter it and redistribute
    it freely, subject to the following restrictions:

    1.  The origin of this software must not be misrepresented; you must
        not claim that you wrote the original software. If you use this
        software in a product, an acknowledgment in the product documentation
        would be appreciated but is not required.

    2.  Altered source versions must be plainly marked as such, and must
        not be misrepresented as being the original software.

    3.  This notice may not be removed or altered from any source
        distribution.
=============================================================================*/
#ifndef BOOST_MULTI_PASS
#define BOOST_MULTI_PASS

#include <deque>
#include <iterator>

#include <iostream>
#include "boost/fixed_size_queue.hpp"

namespace boost
{

namespace multi_pass_policies
{

///////////////////////////////////////////////////////////////////////////////
// class ref_counted
// Implementation of an OwnershipPolicy used by multi_pass.
//
// Implementation modified from RefCounted class from the Loki library by
// Andrei Alexandrescu
///////////////////////////////////////////////////////////////////////////////
class ref_counted
{
    public:
        ref_counted()
            : count(new unsigned int(1))
        {}

        ref_counted(ref_counted const& x)
            : count(x.count)
        {}

        // clone is called when a copy of the iterator is made, so increment
        // the ref-count.
        void clone()
        {
            ++*count;
        }

        // called when a copy is deleted.  Decrement the ref-count.  Return
        // value of true indicates that the last copy has been released.
        bool release()
        {
            if (!--*count)
            {
                delete count;
                count = 0;
                return true;
            }
            return false;
        }

        void swap(ref_counted& x)
        {
            std::swap(count, x.count);
        }

        // returns true if there is only one iterator in existence.
        // std_dequeu StoragePolicy will free it's buffered data if this
        // returns true.
        bool unique() const
        {
            return *count == 1;
        }

    private:
        unsigned int* count;
};

///////////////////////////////////////////////////////////////////////////////
// class first_owner
// Implementation of an OwnershipPolicy used by multi_pass
// This ownership policy dictates that the first iterator created will
// determine the lifespan of the shared components.  This works well for
// spirit, since no dynamic allocation of iterators is done, and all copies
// are make on the stack.
//
// There is a caveat about using this policy together with the std_deque
// StoragePolicy. Since first_owner always returns false from unique(),
// std_deque will only release the queued data if clear_queue() is called.
///////////////////////////////////////////////////////////////////////////////
class first_owner
{
    public:
        first_owner()
            : first(true)
        {}

        first_owner(first_owner const& x)
            : first(false)
        {}

        void clone()
        {
        }

        // return true to indicate deletion of resources
        bool release()
        {
            return first;
        }

        void swap(first_owner& x)
        {
            // if we're the first, we still remain the first, even if assigned
            // to, so don't swap first_.  swap is only called from operator=
        }

        bool unique() const
        {
            return false; // no way to know, so always return false
        }

    private:
        bool first;
};

///////////////////////////////////////////////////////////////////////////////
// class illegal_backtracking
// thrown by buf_id_check CheckingPolicy if an instance of an iterator is 
// used after another one has invalidated the queue
///////////////////////////////////////////////////////////////////////////////
class illegal_backtracking // derive from std::exception?
{
};

///////////////////////////////////////////////////////////////////////////////
// class buf_id_check
// Implementation of the CheckingPolicy used by multi_pass
// This policy is most effective when used together with the std_deque 
// StoragePolicy.
// If used with the fixed_size_queue StoragePolicy, it will not detect
// iterator derefereces that are out of the range of the queue.
///////////////////////////////////////////////////////////////////////////////
class buf_id_check
{
    public:
        buf_id_check()
            : shared_buf_id(new unsigned long(0))
            , buf_id(0)
        {}

        buf_id_check(buf_id_check const& x)
            : shared_buf_id(x.shared_buf_id)
            , buf_id(x.buf_id)
        {}

        // will be called from the destructor of the last iterator.
        void destroy()
        {
            delete shared_buf_id;
            shared_buf_id = 0;
        }

        void swap(buf_id_check& x)
        {
            std::swap(shared_buf_id, x.shared_buf_id);
            std::swap(buf_id, x.buf_id);
        }

        // called to verify that everything is okay.
        void check() const
        {
            if (buf_id != *shared_buf_id)
            {
                throw illegal_backtracking();
            }
        }

        // called from multi_pass::clear_queue, so we can increment the count
        void clear_queue()
        {
            ++*shared_buf_id;
            ++buf_id;
        }

    private:
        unsigned long* shared_buf_id;
        unsigned long buf_id;
};

///////////////////////////////////////////////////////////////////////////////
// class no_check
// Implementation of the CheckingPolicy used by multi_pass
// It does not do anything :-)
///////////////////////////////////////////////////////////////////////////////
class no_check
{
    public:
        no_check() {}
        no_check(no_check const& x) {}
        void destroy() {}
        void swap(no_check& x) {}
        void check() const {}
        void clear_queue() {}
};

///////////////////////////////////////////////////////////////////////////////
// class std_deque
// Implementation of the StoragePolicy used by multi_pass
// This stores all data in a std::deque, and keeps an offset to the current
// position. It stores all the data unless there is only one
// iterator using the queue.
// Note: a position is used instead of an iterator, because a push_back on
// a deque can invalidate any iterators.
///////////////////////////////////////////////////////////////////////////////
class std_deque
{
    public:

template <typename ValueT, ::std::size_t N>
class inner
{
    private:
        typedef std::deque<ValueT> queue_type;
        queue_type* queuedElements;
        mutable typename queue_type::size_type queuePosition;

    public:
        inner()
            : queuedElements(new queue_type)
            , queuePosition(0)
        {}

        inner(inner const& x)
            : queuedElements(x.queuedElements)
            , queuePosition(x.queuePosition)
        {}

        // will be called from the destructor of the last iterator.
        void destroy()
        {
            assert(queuedElements);
            delete queuedElements;
            queuedElements = 0;
        }

        void swap(inner& x)
        {
            std::swap(queuedElements, x.queuedElements);
            std::swap(queuePosition, x.queuePosition);
        }

        // This is called when the iterator is dereferenced.  It's a template
        // method so we can recover the type of the multi_pass iterator
        // and call unique and access the m_input data member.
        template <typename MultiPassT>
        ValueT dereference(MultiPassT const& mp) const
        {
            if (queuePosition == queuedElements->size())
            {
                // check if this is the only iterator
                if (mp.unique())
                {
                    // free up the memory used by the queue.
                    if (queuedElements->size() > 0)
                    {
                        queuedElements->clear();
                        queuePosition = 0;
                    }
                }
                return **mp.m_input;
            }
            else
            {
                return (*queuedElements)[queuePosition];
            }
        }

        // This is called when the iterator is incremented.  It's a template
        // method so we can recover the type of the multi_pass iterator
        // and call unique and access the m_input data member.
        template <typename MultiPassT>
        void increment(MultiPassT& mp)
        {
            if (queuePosition == queuedElements->size())
            {
                // check if this is the only iterator
                if (mp.unique())
                {
                    // free up the memory used by the queue.
                    if (queuedElements->size() > 0)
                    {
                        queuedElements->clear();
                        queuePosition = 0;
                    }
                }
                else
                {
                    queuedElements->push_back(**mp.m_input);
                    ++queuePosition;
                }
                ++*mp.m_input;
            }
            else
            {
                ++queuePosition;
            }

        }

        // called to forcibly clear the queue
        void clear_queue()
        {
            queuedElements->clear();
            queuePosition = 0;
        }

        // called to determine whether the iterator is an eof iterator
        template <typename MultiPassT>
        bool is_eof(MultiPassT const& mp) const
        {
            return queuePosition == queuedElements->size() && 
                *mp.m_input == typename MultiPassT::iterator_type();
        }

        // called by operator==
        bool equal_to(inner const& x) const
        {
            return queuePosition == x.queuePosition;
        }

        // called by operator<
        bool less_than(inner const& x) const
        {
            return queuePosition < x.queuePosition;
        }
}; // class inner

}; // class std_deque


///////////////////////////////////////////////////////////////////////////////
// class fixed_size_queue
// Implementation of the StoragePolicy used by multi_pass
// fixed_size_queue keeps a circular buffer (implemented by 
// boost::fixed_size_queue class) that is size N+1 and stores N elements.
// It is up to the user to ensure that there is enough look ahead for their
// grammar.  Currently there is no way to tell if an iterator is pointing
// to forgotten data.  The leading iterator will put an item in the queue
// and remove one when it is incremented.  No dynamic allocation is done,
// except on creation of the queue (fixed_size_queue constructor).
///////////////////////////////////////////////////////////////////////////////
class fixed_size_queue
{
    public:

template <typename ValueT, ::std::size_t N>
class inner
{
    private:
        typedef ::boost::fixed_size_queue<ValueT, N> queue_type;
        queue_type * queuedElements;
        mutable typename queue_type::iterator queuePosition;

    public:
        inner()
            : queuedElements(new queue_type)
            , queuePosition(queuedElements->begin())
        {}

        inner(inner const& x)
            : queuedElements(x.queuedElements)
            , queuePosition(x.queuePosition)
        {}

        // will be called from the destructor of the last iterator.
        void destroy()
        {
            assert(queuedElements);
            delete queuedElements;
            queuedElements = 0;
        }

        void swap(inner& x)
        {
            std::swap(queuedElements, x.queuedElements);
            std::swap(queuePosition, x.queuePosition);
        }

        // This is called when the iterator is dereferenced.  It's a template
        // method so we can recover the type of the multi_pass iterator
        // and access the m_input data member.
        template <typename MultiPassT>
        ValueT dereference(MultiPassT const& mp) const
        {
            if (queuePosition == queuedElements->end())
            {
                return **mp.m_input;
            }
            else
            {
                return *queuePosition;
            }
        }

        // This is called when the iterator is incremented.  It's a template
        // method so we can recover the type of the multi_pass iterator
        // and access the m_input data member.
        template <typename MultiPassT>
        void increment(MultiPassT& mp)
        {
            if (queuePosition == queuedElements->end())
            {
                // don't let the queue get larger than N
                if (queuedElements->size() >= N)
                    queuedElements->pop_front();

                queuedElements->push_back(**mp.m_input);
                ++*mp.m_input;
            }
            ++queuePosition;
        }

        // no-op
        void clear_queue()
        {}

        // called to determine whether the iterator is an eof iterator
        template <typename MultiPassT>
        bool is_eof(MultiPassT const& mp) const
        {
            return queuePosition == queuedElements->end() && 
                *mp.m_input == typename MultiPassT::iterator_type();
        }

        // called by operator==
        bool equal_to(inner const& x) const
        {
            return queuePosition == x.queuePosition;
        }

        // called by operator<
        bool less_than(inner const& x) const
        {
            return queuePosition < x.queuePosition;
        }
}; // class inner

}; // class fixed_size_queue


} // namespace multi_pass_policies

///////////////////////////////////////////////////////////////////////////////
// class template multi_pass (declaration)
///////////////////////////////////////////////////////////////////////////////
template
<
    typename InputIteratorT,
    typename OwnershipPolicy = multi_pass_policies::ref_counted,
    typename CheckingPolicy = multi_pass_policies::buf_id_check,
    typename StoragePolicy = multi_pass_policies::std_deque,
    ::std::size_t N = -1
>
class multi_pass;

// The default multi_pass instantiation uses a ref-counted std_deque scheme.

// One note about the last size_t template parameter:
// This parameter does nothing when used with the std_deque StoragePolicy.
// It applies to the fixed_size_queue StoragePolicy.  It specifies the amount
// of lookahead the queue will have.  If fixed_size_queue is used, the N
// parameter must be set, otherwise the code will not compile. (because a 0
// size array is invalid.)

///////////////////////////////////////////////////////////////////////////////
// class template multi_pass (definition)
///////////////////////////////////////////////////////////////////////////////
template 
<
    typename InputIteratorT,
    typename OwnershipPolicy,
    typename CheckingPolicy,
    typename StoragePolicy,
    std::size_t N
>
class multi_pass
    : public OwnershipPolicy
    , public CheckingPolicy
    , public StoragePolicy::inner<
                typename std::iterator_traits<InputIteratorT>::value_type, N>
{
        typedef OwnershipPolicy OP;
        typedef CheckingPolicy CHP;
        typedef typename StoragePolicy::inner<
            typename std::iterator_traits<InputIteratorT>::value_type, N> SP;

    public:
        typedef std::forward_iterator_tag iterator_category;
        typedef std::iterator_traits<InputIteratorT>::value_type value_type;
        typedef std::iterator_traits<InputIteratorT>::difference_type 
            difference_type;
        typedef std::iterator_traits<InputIteratorT>::pointer pointer;
        typedef std::iterator_traits<InputIteratorT>::reference reference;

        typedef InputIteratorT iterator_type;

        multi_pass();
        explicit multi_pass(InputIteratorT input);
        
        ~multi_pass();

        multi_pass(multi_pass const&);
        multi_pass& operator=(multi_pass const&);

        void swap(multi_pass& x);
        
        // for some strange reason gcc won't accept just plain value_type here
        multi_pass<
            InputIteratorT, 
            OwnershipPolicy, 
            CheckingPolicy, 
            StoragePolicy, 
            N
        >::value_type operator*() const;
        pointer operator->() const;
        multi_pass& operator++();
        multi_pass operator++(int);

        void clear_queue();

        bool operator==(const multi_pass& y) const;
        bool operator<(const multi_pass& y) const;

        // not private so that the StoragePolicy can access it.
        // if template friends worked on more compilers, we could make it a
        // friend...
        InputIteratorT* m_input;

    private: // helper functions
        bool is_eof() const;
};

template 
<
    typename InputIteratorT, 
    typename OwnershipPolicy, 
    typename CheckingPolicy, 
    typename StoragePolicy, 
    ::std::size_t N
>
inline 
multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, StoragePolicy, N>::
multi_pass()
    : OP()
    , CHP()
    , SP()
    , m_input(new InputIteratorT())
{
}

template 
<
    typename InputIteratorT, 
    typename OwnershipPolicy, 
    typename CheckingPolicy, 
    typename StoragePolicy, 
    ::std::size_t N
>
inline 
multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, StoragePolicy, N>::
multi_pass(InputIteratorT input)
    : OP()
    , CHP()
    , SP()
    , m_input(new InputIteratorT(input))
{
}

template 
<
    typename InputIteratorT, 
    typename OwnershipPolicy, 
    typename CheckingPolicy, 
    typename StoragePolicy, 
    ::std::size_t N
>
inline 
multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, StoragePolicy, N>::
~multi_pass()
{
    if (OP::release())
    { 
        CHP::destroy();
        SP::destroy();
        delete m_input;
        m_input = 0;
    }
}

template 
<
    typename InputIteratorT, 
    typename OwnershipPolicy, 
    typename CheckingPolicy, 
    typename StoragePolicy, 
    ::std::size_t N
>
inline 
multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, StoragePolicy, N>::
multi_pass(
        multi_pass const& x)
    : OP(x)
    , CHP(x)
    , SP(x)
    , m_input(x.m_input)
{
    OP::clone();
}

template 
<
    typename InputIteratorT, 
    typename OwnershipPolicy, 
    typename CheckingPolicy, 
    typename StoragePolicy, 
    ::std::size_t N
>
inline 
multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, StoragePolicy, N>&
multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, StoragePolicy, N>::
operator=(
        multi_pass const& x)
{
    multi_pass temp(x);
    temp.swap(*this);
    return *this;
}

template 
<
    typename InputIteratorT, 
    typename OwnershipPolicy, 
    typename CheckingPolicy, 
    typename StoragePolicy, 
    ::std::size_t N
>
inline void
multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, StoragePolicy, N>::
swap(multi_pass& x)
{
    OP::swap(x);
    CHP::swap(x);
    SP::swap(x);
    std::swap(m_input, x.m_input);
}

template 
<
    typename InputIteratorT, 
    typename OwnershipPolicy, 
    typename CheckingPolicy, 
    typename StoragePolicy, 
    ::std::size_t N
>
inline 
multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, StoragePolicy, N>::
value_type
multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, StoragePolicy, N>::
operator*() const
{
    CHP::check();
    return SP::dereference(*this);
}

template 
<
    typename InputIteratorT, 
    typename OwnershipPolicy, 
    typename CheckingPolicy, 
    typename StoragePolicy, 
    ::std::size_t N
>
inline 
multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, StoragePolicy, N>::
pointer
multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, StoragePolicy, N>::
operator->() const
{
    return &(operator*());
}

template 
<
    typename InputIteratorT, 
    typename OwnershipPolicy, 
    typename CheckingPolicy, 
    typename StoragePolicy, 
    ::std::size_t N
>
inline 
multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, StoragePolicy, N>&
multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, StoragePolicy, N>::
operator++()
{
    CHP::check();
    SP::increment(*this);
    return *this;
}

template 
<
    typename InputIteratorT, 
    typename OwnershipPolicy, 
    typename CheckingPolicy, 
    typename StoragePolicy, 
    ::std::size_t N
>
inline 
multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, StoragePolicy, N>
multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, StoragePolicy, N>::
operator++(int)
{
    multi_pass
    <
        InputIteratorT, 
        OwnershipPolicy, 
        CheckingPolicy, 
        StoragePolicy, 
        N
    > tmp(*this);

    ++*this;

    return tmp;
}

template 
<
    typename InputIteratorT, 
    typename OwnershipPolicy, 
    typename CheckingPolicy, 
    typename StoragePolicy, 
    ::std::size_t N
>
inline void 
multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, StoragePolicy, N>::
clear_queue()
{
    SP::clear_queue();
    CHP::clear_queue();
}

template 
<
    typename InputIteratorT, 
    typename OwnershipPolicy, 
    typename CheckingPolicy, 
    typename StoragePolicy, 
    ::std::size_t N
>
inline bool 
multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, StoragePolicy, N>::
is_eof() const
{
    return SP::is_eof(*this);
}

///// Comparisons
template 
<
    typename InputIteratorT, 
    typename OwnershipPolicy, 
    typename CheckingPolicy, 
    typename StoragePolicy, 
    ::std::size_t N
>
inline bool 
multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, StoragePolicy, N>::
operator==(const multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, 
        StoragePolicy, N>& y) const
{
    if (is_eof() && y.is_eof())
    {
        return true;  // both are EOF
    }
    else if (is_eof() ^ y.is_eof())
    {
        return false; // one is EOF, one isn't
    }
    else if (m_input != y.m_input)
    {
        return false;
    }
    else
    {
        return SP::equal_to(y);
    }
}

template 
<
    typename InputIteratorT, 
    typename OwnershipPolicy, 
    typename CheckingPolicy, 
    typename StoragePolicy, 
    ::std::size_t N
>
inline bool 
multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, StoragePolicy, N>::
operator<(const multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, 
        StoragePolicy, N>& y) const
{
    return SP::less_than(y);
}

template 
<
    typename InputIteratorT, 
    typename OwnershipPolicy, 
    typename CheckingPolicy, 
    typename StoragePolicy, 
    ::std::size_t N
>
inline
bool operator!=(
        const multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, 
                        StoragePolicy, N>& x,
        const multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, 
                        StoragePolicy, N>& y)
{
    return !(x == y);
}

template 
<
    typename InputIteratorT, 
    typename OwnershipPolicy, 
    typename CheckingPolicy, 
    typename StoragePolicy, 
    ::std::size_t N
>
inline
bool operator>(
        const multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, 
                        StoragePolicy, N>& x,
        const multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, 
                        StoragePolicy, N>& y)
{
    return y < x;
}

template 
<
    typename InputIteratorT, 
    typename OwnershipPolicy, 
    typename CheckingPolicy, 
    typename StoragePolicy, 
    ::std::size_t N
>
inline
bool operator>=(
        const multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, 
                        StoragePolicy, N>& x,
        const multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, 
                        StoragePolicy, N>& y)
{
    return !(x < y);
}

template 
<
    typename InputIteratorT, 
    typename OwnershipPolicy, 
    typename CheckingPolicy, 
    typename StoragePolicy, 
    ::std::size_t N
>
inline
bool operator<=(
        const multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, 
                        StoragePolicy, N>& x,
        const multi_pass<InputIteratorT, OwnershipPolicy, CheckingPolicy, 
                        StoragePolicy, N>& y)
{
    return !(y < x);
}

///// Generator function
template <typename InputIteratorT>
inline multi_pass<InputIteratorT>
make_multi_pass(InputIteratorT i)
{
    return multi_pass<InputIteratorT>(i);
}

// this could be a template typedef, since such a thing doesn't
// exist in C++, we'll use inheritance to accomplish the same thing.

template <typename InputIteratorT, ::std::size_t N>
class look_ahead : 
    public multi_pass<
        InputIteratorT, 
        multi_pass_policies::first_owner,
        multi_pass_policies::no_check,
        multi_pass_policies::fixed_size_queue,
        N>
{
        typedef multi_pass<
            InputIteratorT,
            multi_pass_policies::first_owner,
            multi_pass_policies::no_check,
            multi_pass_policies::fixed_size_queue,
            N> base_t;
    public:
        look_ahead()
            : base_t() {}

        explicit look_ahead(InputIteratorT x)
            : base_t(x) {}
        
        look_ahead(look_ahead const& x)
            : base_t(x) {}

    // default generated operators destructor and assignment operator are okay.
};



} // namespace boost

#endif

