/*=============================================================================
    Attributes rules

    Spirit V1.2
    Copyright (c) 2001, Hartmut Kaiser

    This software is provided 'as-is', without any express or implied
    warranty. In no event will the copyright holder be held liable for
    any damages arising from the use of this software.

    Permission is granted to anyone to use this software for any purpose,
    including commercial applications, and to alter it and redistribute
    it freely, subject to the following restrictions:

    1.  The origin of this software must not be misrepresented; you must
        not claim that you wrote the original software. If you use this
        software in a product, an acknowledgment in the product documentation
        would be appreciated but is not required.

    2.  Altered source versions must be plainly marked as such, and must
        not be misrepresented as being the original software.

    3.  This notice may not be removed or altered from any source
        distribution.

    Acknowledgements:

        Special thanks to Dan Nuffer, John (EBo) David, Chris Uzdavinis,
        and Doug Gregor. These people are most instrumental in steering
        Spirit in the right direction.

        Special thanks also to people who have contributed to the code base
        and sample code, ported Spirit to various platforms and compilers,
        gave suggestions, reported and provided bug fixes. Alexander
        Hirner, Andy Elvey, Bogdan Kushnir, Brett Calcott, Bruce Florman,
        Changzhe Han, Colin McPhail, Hakki Dogusan, Jan Bares, Joseph
        Smith, Martijn W. van der Lee, Raghavendra Satish, Remi Delcos, Tom
        Spilman, Vladimir Prus, W. Scott Dillman, David A. Greene, Bob
        Bailey, Hartmut Kaiser.

        Finally special thanks also to people who gave feedback and
        valuable comments, particularly members of Spirit's Source Forge
        mailing list and boost.org.

    URL: http://spirit.sourceforge.net/

=============================================================================*/
#ifndef SPIRIT_ATTR_RULE_HPP
#define SPIRIT_ATTR_RULE_HPP

///////////////////////////////////////////////////////////////////////////////

#include "boost/spirit/closure.hpp"

///////////////////////////////////////////////////////////////////////////////
namespace spirit {

///////////////////////////////////////////////////////////////////////////////
// 
//  attr_match class
//
//      Is returned by the parse function of a closure_parser. Contains 
//      additionally the value of the closure_member<0>, which by design is 
//      used as the return value of a closure_parser.
//
//      To access this value the code has to look like this:
//
//          closure<string> loc;
//          closure_member<0> str;
//          rule<> r = loc[ 
//                  ... some rule definition which refers to 'loc(str)' ...
//              ][actor];
//
//      Where 'actor' is some action function or actor object, which takes
//      one argument of the type of closure_member<0> (here: 'string').
//
//      For closures embedded in a grammar the return value is automatically
//      propagated to the related grammar object, where it is saved for later 
//      access.
//
///////////////////////////////////////////////////////////////////////////////
template <typename TypeT, typename BaseT>
class attr_match :
    public BaseT
{
public:
    attr_match();
    explicit attr_match(unsigned length);
    explicit attr_match(BaseT const &hit);
    template <typename OtherTypeT>
    attr_match(attr_match<OtherTypeT, BaseT> const &hit);

//    explicit attr_match(match const &hit);

    attr_match &operator += (match const& other);

    TypeT const &attr_value() const;
    void attr_value(TypeT const &val);

private:
    TypeT value;
};

///////////////////////////////////////////////////////////////////////////////
// friend operators
template <typename TypeT, typename BaseT>
attr_match<TypeT, BaseT>
operator+ (attr_match<TypeT, BaseT> const& a, match const& b);

template <typename TypeT, typename BaseT>
attr_match<TypeT, BaseT>
operator+ (match const& a, attr_match<TypeT, BaseT> const& b);

///////////////////////////////////////////////////////////////////////////////
// Specialization for type 'null_type', which is essentially equivalent to the 
// original type 'match'

template <typename BaseT>
class attr_match<closure_null_type, BaseT> : 
    public BaseT
{
public:
// Boland 5.5 complaines, if these functions are not inlined
    attr_match() {}
    explicit attr_match(unsigned length) : BaseT(length) {}
    explicit attr_match(BaseT const &hit) : BaseT(hit) {}
    template <typename OtherTypeT>
    attr_match(attr_match<OtherTypeT, BaseT> const &hit) : BaseT(hit) {}
//    explicit attr_match(match const &hit) : BaseT(hit.length()) {}

    attr_match &operator += (match const& other)
    {
        assert(*this && other);
        data += other.length();
        return *this;
    }
    closure_null_type const &attr_value() const
        { return closure_null_type(); }
    void attr_value(closure_null_type const &val) {}
};

///////////////////////////////////////////////////////////////////////////////
// 
//  param_parser class
//
//      Helper template for initializing different closure_members before the
//      entering the parsing context of the corresponding parser.
//
///////////////////////////////////////////////////////////////////////////////
template <typename TupleT, typename ParserT, typename InitTupleT>
struct param_parser :
    public unary<ParserT>, 
    public parser<param_parser<TupleT, ParserT, InitTupleT> >
{
    typedef TupleT tuple_t;
    typedef typename element<0, TupleT>::type return_t;

    typedef param_parser_category parser_category;

    param_parser(InitTupleT const &init, TupleT *&ptr, ParserT const& parser);

    template <typename IteratorT>
    attr_match<return_t>
    parse(IteratorT& first, IteratorT const& last) const;

    template <typename IteratorT, typename MatchTraitsT>
    attr_match<return_t, typename MatchTraitsT::match_t>
    ast_parse(IteratorT& first, IteratorT const& last, 
            MatchTraitsT const&) const;

    template <typename ActionT>
    closure_parser_action<param_parser<TupleT, ParserT, InitTupleT>, ActionT>
    operator[](ActionT const& actor) const
    {
        //  Borland 5.5 reports an internal compiler
        //  error if this is not defined here.
        typedef closure_parser_action<
                    param_parser<TupleT, ParserT, InitTupleT>, ActionT> 
                actor_t;
        return actor_t(*this, actor);
    }

private:
    TupleT *&ptr;
    InitTupleT init;
};

///////////////////////////////////////////////////////////////////////////////
// 
//  parameters for the parsing process of the rule
// 
///////////////////////////////////////////////////////////////////////////////
template <typename TupleT, typename ParserT>
struct param_parser_gen
{
    param_parser_gen(TupleT *& ptr_, ParserT const &parser_)
        : parser(parser_), ptr(ptr_) {}

    template <typename T0>
    param_parser<
        TupleT, ParserT, 
        typename boost::tuples::detail::make_tuple_mapper<T0>::type
    > 
    operator()(T0 const &t0);

    template <typename T0, typename T1>
    param_parser<
        TupleT, ParserT, 
        typename boost::tuples::detail::make_tuple_mapper<T0, T1>::type
    > 
    operator()(T0 const &t0, T1 const &t1);

    template <typename T0, typename T1, typename T2>
    param_parser<
        TupleT, ParserT, 
        typename boost::tuples::detail::make_tuple_mapper<T0, T1, T2>::type
    > 
    operator()(T0 const &t0, T1 const &t1, T2 const &t2);

    template <typename T0, typename T1, typename T2, typename T3>
    param_parser<
        TupleT, ParserT, 
        typename boost::tuples::detail::make_tuple_mapper<T0, T1, T2, T3>::type
    > 
    operator()(T0 const &t0, T1 const &t1, T2 const &t2, T3 const &t3);

    template <typename T0, typename T1, typename T2, typename T3, typename T4>
    param_parser<
        TupleT, ParserT, 
        typename boost::tuples::detail::make_tuple_mapper<
            T0, T1, T2, T3, T4>::type
    > 
    operator()(T0 const &t0, T1 const &t1, T2 const &t2, T3 const &t3, 
        T4 const &t4);

    template <
        typename T0, typename T1, typename T2, typename T3, typename T4,
        typename T5
    >
    param_parser<
        TupleT, ParserT, 
        typename boost::tuples::detail::make_tuple_mapper<
            T0, T1, T2, T3, T4, T5>::type
    > 
    operator()(T0 const &t0, T1 const &t1, T2 const &t2, T3 const &t3, 
        T4 const &t4, T5 const &t5);

    template <
        typename T0, typename T1, typename T2, typename T3, typename T4,
        typename T5, typename T6
    >
    param_parser<
        TupleT, ParserT, 
        typename boost::tuples::detail::make_tuple_mapper<
            T0, T1, T2, T3, T4, T5, T6>::type
    > 
    operator()(T0 const &t0, T1 const &t1, T2 const &t2, T3 const &t3, 
        T4 const &t4, T5 const &t5, T6 const &t6);

    template <
        typename T0, typename T1, typename T2, typename T3, typename T4,
        typename T5, typename T6, typename T7
    >
    param_parser<
        TupleT, ParserT, 
        typename boost::tuples::detail::make_tuple_mapper<
            T0, T1, T2, T3, T4, T5, T6, T7>::type
    > 
    operator()(T0 const &t0, T1 const &t1, T2 const &t2, T3 const &t3, 
        T4 const &t4, T5 const &t5, T6 const &t6, T7 const &t7);

    template <
        typename T0, typename T1, typename T2, typename T3, typename T4,
        typename T5, typename T6, typename T7, typename T8
    >
    param_parser<
        TupleT, ParserT, 
        typename boost::tuples::detail::make_tuple_mapper<
            T0, T1, T2, T3, T4, T5, T6, T7, T8>::type
    > 
    operator()(T0 const &t0, T1 const &t1, T2 const &t2, T3 const &t3, 
        T4 const &t4, T5 const &t5, T6 const &t6, T7 const &t7, T8 const &t8);

    template <
        typename T0, typename T1, typename T2, typename T3, typename T4,
        typename T5, typename T6, typename T7, typename T8, typename T9
    >
    param_parser<
        TupleT, ParserT, 
        typename boost::tuples::detail::make_tuple_mapper<
            T0, T1, T2, T3, T4, T5, T6, T7, T8, T9>::type
    > 
    operator()(T0 const &t0, T1 const &t1, T2 const &t2, T3 const &t3, 
        T4 const &t4, T5 const &t5, T6 const &t6, T7 const &t7, T8 const &t8,
        T9 const &t9);

private:
    TupleT *& ptr;
    ParserT const &parser;
};

///////////////////////////////////////////////////////////////////////////////
// 
//  attr_rule class
//
//      This class is essentially a closure wrapped around a rule, so that the
//      closure members are accessible all the way down the parser hierarchy.
//
///////////////////////////////////////////////////////////////////////////////

template <typename ClosureT, typename IteratorT, typename MatchTraitsT>
struct attr_rule :
    public base_rule<IteratorT, MatchTraitsT, 
        attr_rule<ClosureT, IteratorT, MatchTraitsT> >,
    public ClosureT::base_closure_t
{
    typedef attr_rule<ClosureT, IteratorT, MatchTraitsT> this_t;
    typedef base_rule<IteratorT, MatchTraitsT, this_t> rule_t;

    typedef typename ClosureT::base_closure_t closure_t;
    typedef typename closure_t::tuple_t tuple_t;
    typedef typename element<0, tuple_t>::type return_t;

    typedef closure_parser_category parser_category;

    attr_rule();
    attr_rule(this_t const& other);
    attr_rule(rule_t const& other);
    template <typename ParserT>
    attr_rule(ParserT const& parser);

    ~attr_rule();

// parsing functions
    attr_match<return_t>
    parse(IteratorT& first, IteratorT const& last) const;

    attr_match<return_t, typename MatchTraitsT::match_t>
    ast_parse(IteratorT& first, IteratorT const& last, 
            MatchTraitsT const&) const;

// rule specific functions
    attr_rule& operator=(attr_rule const& other);
    attr_rule& operator=(rule_t const& other);
    template <typename ParserT>
    attr_rule& operator=(ParserT const& parser);

// closure specifics
    template <int N>
    reference_wrapper<local<tuple_t, N> >
    operator()(closure_member<N> cm)
        { return closure_t::operator()(cm); }

// generate different types of action parsers
    template <typename T>
    closure_parser_action<this_t, reference_wrapper<T> > // one argument actors
    operator[](reference_wrapper<T> const& wrap) const
    {
        //  Borland 5.5 reports an internal compiler
        //  error if this is not defined here.
        typedef
            closure_parser_action<this_t, reference_wrapper<T> >
            actor_t;
        return actor_t(*this, wrap);
    }

    template <typename ActionT>
    action<rule_t, ActionT>
    operator[](ActionT const& actor) const              // two argument actors
        { return rule_t::operator[](actor); }

// direct access of rule base member
    rule_t const &rule_base() const 
        { return *static_cast<rule_t const *>(this); }

// support for parser initialization
    param_parser_gen<tuple_t, rule_t> init;
};

///////////////////////////////////////////////////////////////////////////////

}   //  namespace Spirit

// include corresponding implementation header
// Can't do this because of dependencies on closure.ipp - DN
//#include "boost/spirit/impl/attr_rule.ipp"

#endif
