/*=============================================================================
    Numeric parsers

    Spirit V1.2
    Copyright (c) 2001, Joel de Guzman
    Copyright (c) 2001, Hartmut Kaiser

    This software is provided 'as-is', without any express or implied
    warranty. In no event will the copyright holder be held liable for
    any damages arising from the use of this software.

    Permission is granted to anyone to use this software for any purpose,
    including commercial applications, and to alter it and redistribute
    it freely, subject to the following restrictions:

    1.  The origin of this software must not be misrepresented; you must
        not claim that you wrote the original software. If you use this
        software in a product, an acknowledgment in the product documentation
        would be appreciated but is not required.

    2.  Altered source versions must be plainly marked as such, and must
        not be misrepresented as being the original software.

    3.  This notice may not be removed or altered from any source
        distribution.

    Acknowledgements:

        Special thanks to Dan Nuffer, John (EBo) David, Chris Uzdavinis,
        and Doug Gregor. These people are most instrumental in steering
        Spirit in the right direction.

        Special thanks also to people who have contributed to the code Radix
        and sample code, ported Spirit to various platforms and compilers,
        reported and provided bug fixes. Alexander Hirner, Andy Elvey,
        Bogdan Kushnir, Brett Calcott, Bruce Florman, Changzhe Han, Colin
        McPhail, Hakki Dogusan, Jan Bares, Joseph Smith, Martijn W. van der
        Lee, Raghavendra Satish, Remi Delcos, Tom Spilman, Vladimir Prus,
        W. Scott Dillman.

        Finally special thanks also to people who gave feedback and
        valuable comments, particularly members of Spirit's Source Forge
        mailing list and boost.org.

    URL: http://spirit.sourceforge.net/

=============================================================================*/
#ifndef SPIRIT_NUMERICS_IPP
#define SPIRIT_NUMERICS_IPP

///////////////////////////////////////////////////////////////////////////////

#include <cmath>
#include <functional>
#include <iterator>

#include <assert.h>

#include "boost/spirit/primitives.hpp"
#include "boost/spirit/iterators.hpp"

///////////////////////////////////////////////////////////////////////////////
namespace spirit {

//////////////////////////////////
namespace impl {

    //////////////////////////////////
    template <typename IteratorT>
    bool
    extract_sign(IteratorT& iter, unsigned& count)
    {
        //  Extract the sign

        count = 0;
        bool neg = *iter == '-';
        if (neg || (*iter == '+'))
        {
            ++iter;
            ++count;
            return neg;
        }

        return false;
    }

///////////////////////////////////////////////////////////////////////////////
// Traits class for number conversion
    template<const int Radix>
    struct radix_traits
    {
    // Tests validity of a single character in the context of the given
    // radix (template parameter Radix)
        template<typename CharT> static bool isok(CharT ch);

    // Converts a digit from character representation to binary representation
    // in the context of the given radix (template parameter Radix)
        template<typename CharT> static int uint(CharT ch);
    };

    template<>
    struct radix_traits<2> {
        template<typename CharT>
        static bool isok(CharT ch) { return ('0' == ch || '1' == ch); }

        template<typename CharT>
        static int uint(CharT ch) { return ch - '0'; }
    };

    template<>
    struct radix_traits<8> {
        template<typename CharT>
        static bool isok(CharT ch) { return ('0' <= ch && ch <= '7'); }

        template<typename CharT>
        static int uint(CharT ch) { return ch - '0'; }
    };

    template<>
    struct radix_traits<10> {
        template<typename CharT>
        static bool isok(CharT ch) { return isdigit(ch); }

        template<typename CharT>
        static int uint(CharT ch) { return ch - '0'; }
    };

    template<>
    struct radix_traits<16> {
        template<typename CharT>
        static bool isok(CharT ch) { return isxdigit(ch); }

        template<typename CharT>
        static int uint(CharT ch)
        {
            if (isdigit(ch))
                return ch - '0';
            return tolower(ch) - 'a' + 10;
        }
    };

///////////////////////////////////////////////////////////////////////////////
// Helper templates for abstracting the compare methods for different 
// numeric_parser_traits classes
    template<typename IteratorT>
    struct trait_type_cmp
    {
    // trait is given as a parser (template parameter ParserT)
        template <typename ParserT>
        static match parse (parser<ParserT> p, IteratorT &it1, IteratorT const &it2) 
        {
            return p.derived().parse(it1, it2);
        }
        template <typename ParserT>
        static match parse2 (parser<ParserT> p, IteratorT &it1, IteratorT const &it2) 
        {
            return parse (p.derived(), it1, it2);
        }

    // trait is given as a single character
        typedef typename std::iterator_traits<IteratorT>::value_type CharT;
        static match parse (CharT ch, IteratorT &it1, IteratorT const &/*it2*/) 
        {
            if (ch == *it1) {
                ++it1;
                return match(1);
            }
            return match();
        }
        static match parse2 (CharT ch, IteratorT &it1, IteratorT const &it2) 
        {
            return nocase[ch].parse(it1, it2);
        }

    // trait is given as a string
        typedef typename std::iterator_traits<IteratorT>::pointer PointerT;
        static match parse (PointerT pch, IteratorT &it1, IteratorT const &it2) 
        {
            typedef typename std::iterator_traits<IteratorT>::value_type CharT;
            return strlit<CharT>(pch).parse(it1, it2);
        }
        static match parse2 (PointerT pch, IteratorT &it1, 
                IteratorT const &it2) 
        {
            return parse (pch, it1, it2);
        }
    };
/*
    template<typename IteratorT, typename ParserT>
    struct trait_type_cmp
    {
    // trait is given as a parser (template parameter ParserT)
        static match parse (ParserT p, IteratorT &it1, IteratorT const &it2) 
        {
            return p.parse(it1, it2);
        }
        static match parse2 (ParserT p, IteratorT &it1, IteratorT const &it2) 
        {
            return parse (p, it1, it2);
        }
    };

    template<typename IteratorT>
    struct trait_type_cmp<IteratorT, const int> 
    {
    // trait is given as a single character
        typedef typename std::iterator_traits<IteratorT>::value_type CharT;
        static match parse (CharT ch, IteratorT &it1, IteratorT const &) 
        {
            if (ch == *it1) {
                ++it1;
                return match(1);
            }
            return match();
        }
        static match parse2 (CharT ch, IteratorT &it1, IteratorT const &it2) 
        {
            return nocase[ch].parse(it1, it2);
        }
    };

    template<typename IteratorT>
    struct trait_type_cmp<IteratorT, 
        const std::iterator_traits<IteratorT>::pointer> 
    {
    // trait is given as a string
        typedef typename std::iterator_traits<IteratorT>::pointer PointerT;
        static match parse (PointerT pch, IteratorT &it1, IteratorT const &it2) 
        {
            typedef typename std::iterator_traits<IteratorT>::value_type CharT;
            return strlit<CharT>(pch).parse(it1, it2);
        }
        static match parse2 (PointerT pch, IteratorT &it1, 
                IteratorT const &it2) 
        {
            return parse (pch, it1, it2);
        }
    };
*/
    template<typename TraitsT>
    struct traits_cmp {
    // tests for decimal separator
        template<typename IteratorT>
        static bool is_ds(IteratorT &it1, IteratorT const &it2, unsigned &count)
        {
            typedef trait_type_cmp<IteratorT> 
                trait_type;
            match mexpr = 
                trait_type::parse(TraitsT::decimal_sep_char, it1, it2);
            if (mexpr) {
                count += mexpr.length();
                return true;
            }
            return false;
        }

    // tests for thousands separator
        template<typename IteratorT>
        static bool is_ts(IteratorT &it1, IteratorT const &it2, unsigned &count)
        {
            if (TraitsT::thousands_sep_char != 0) {
                typedef trait_type_cmp<IteratorT> trait_type;
                match mexpr = 
                    trait_type::parse(TraitsT::thousands_sep_char, it1, it2);
                if (mexpr) {
                    count += mexpr.length();
                    return true;
                }
            }
            return false;
        }

    // tests for delimiter before exponent
        template<typename IteratorT>
        static bool is_exp(IteratorT &it1, IteratorT const &it2, 
                unsigned &count)
        {
            typedef trait_type_cmp<IteratorT> trait_type;
            match mexpr = trait_type::parse2(TraitsT::exponent_char, it1, it2);
            if (mexpr) {
                count += mexpr.length();
                return true;
            }
            return false;
        }
    };

///////////////////////////////////////////////////////////////////////////////
// Helper template functors for distinction of overflow testing for signed and
// unsigned data types. The template parameter T represents the parsed data type.
    template<typename T>
    struct nochange : std::unary_function<T, T>
    {
          T operator()(const T& x) const { return x; }
    };

    template<typename T>
    struct decrement : std::unary_function<T, T>
    {
          T operator()(const T& x) const { return x-1; }
    };

// Helper template for encapsulation of radix specific conversion of an input 
// string to an integral integer-like value.
//
// The template parameter Radix represents the radix of the number contained
// in the parsed string
    template <const int Radix, typename TraitsT = numeric_parser_traits<> >
    struct extract {
    // Extract a signed integer
    // Returns a non-match, if the number to parse overflows the used integral 
    // type.
    //
    // BEWARE: the parameters 'n' and 'count' should be proper initialized 
    // before calling this function.
        template <typename IteratorT, typename T>
        static bool sint(IteratorT& first, IteratorT const& last, T& n, 
                unsigned& count, bool neg)
        {
            typedef typename remove_wrap<T>::type arg_type;
            if (!neg)
                return uint(first, last, n, count, nochange<arg_type>());

            // Intel V5.0.1 chokes without explicit template parameters
            return uint<IteratorT, T, decrement<arg_type> >(first, last, n, 
                    count, decrement<arg_type>());
        }

    // Extract an unsigned integer
    // Return a non-match, if the number to parse overflows the used integral 
    // type.
    //
    // BEWARE: the parameters 'n' and 'count' should be proper initialized 
    // before calling this functions.
        template <typename IteratorT, typename T>
        static bool uint(IteratorT& first, IteratorT const& last, T& n, 
                unsigned& count)
        {
            typedef typename remove_wrap<T>::type arg_type;
            return uint(first, last, n, count, nochange<arg_type>());
        }

        template <typename IteratorT, typename T, typename PredT>
        static bool uint(IteratorT& first, IteratorT const& last, T& n, 
                unsigned& count, const PredT &pred)
        {
            typedef remove_wrap<T>::type arg_type;

        const arg_type lim = std::numeric_limits<arg_type>::max()/Radix;

            for (/**/; first != last; ++first, ++count)
            {
            // skip thousands separator if required
            unsigned ts = 0;

                if (traits_cmp<TraitsT>::is_ts(first, last, ts)) {
                    --first;        // roll back for loop reinit
                    count += ts-1;
                    continue;
                }

            // handle only 'valid' (in the context of the actual radix) digits
                if (radix_traits<Radix>::isok(*first)) {
                    if (n > lim) {
                    // overflow! return non-match
                        n = 0;
                        count = 0;
                        return false;
                    }
                    n *= Radix;            // operation is now safe

                arg_type nextdigit = radix_traits<Radix>::uint(*first);

                    if (pred(nextdigit) > std::numeric_limits<arg_type>::max() - n) {
                    // overflow! return non-match
                        n = 0;
                        count = 0;
                        return false;
                    }
                    n += nextdigit;            // operation is now safe too
                } else
                    break;
            }
            return true;
        }
    };

} // namespace impl

///////////////////////////////////////////////////////////////////////////////
//
//  numeric_action class implementation
//
///////////////////////////////////////////////////////////////////////////////
template <typename ParserT, typename ActionT>
inline numeric_action<ParserT, ActionT>::numeric_action(ActionT const& actor_)
:   unary<ParserT>(ParserT(0)), actor(actor_) {}

//////////////////////////////////
template <typename ParserT, typename ActionT>
template <typename IteratorT>
inline match
numeric_action<ParserT, ActionT>::
    parse(IteratorT& first, IteratorT const& last) const
{
    match hit = this->subject().parse(first, last);
    if (hit)
        actor(this->subject().n);
    return hit;
}

//////////////////////////////////
template <typename ParserT, typename ActionT>
template <typename IteratorT, typename MatchTraitsT>
inline typename MatchTraitsT::match_t
numeric_action<ParserT, ActionT>::
    ast_parse(IteratorT& first, IteratorT const& last, 
            MatchTraitsT const& mt) const
{
    typename MatchTraitsT::match_t hit = 
        this->subject().ast_parse(first, last, mt);
    if (hit)
        actor(this->subject().n);
    return hit;
}

///////////////////////////////////////////////////////////////////////////////
//
//  uint_parser class implementation
//
///////////////////////////////////////////////////////////////////////////////
template <typename T, const int Radix>
inline uint_parser<T, Radix>::uint_parser(T n_)
:   n(n_) {}

//////////////////////////////////
template <typename T, const int Radix>
template <typename IteratorT>
inline match
uint_parser<T, Radix>::parse(IteratorT& first, IteratorT const& last) const
{
    n = 0;
    if (first == last)
        return match();

    typedef typename impl::strip_scanner<IteratorT>
        ::iterator_type plain_iter;
    plain_iter  i1 = impl::strip_scanner<IteratorT>::get(first);
    plain_iter  i2 = impl::strip_scanner<IteratorT>::get(last);

    unsigned count = 0;
    if (!impl::radix_traits<Radix>::isok(*i1))
        return match();

    if (impl::extract<Radix>::uint(i1, i2, n, count)) {
        first = i1;
        return match(count);
    }
    return match();        // return non-match if uint overflows
}

//////////////////////////////////
template <typename T, const int Radix>
template <typename IteratorT, typename MatchTraitsT>
inline typename MatchTraitsT::match_t
uint_parser<T, Radix>::ast_parse(IteratorT& first, IteratorT const& last, 
        MatchTraitsT const& mt) const
{
    n = 0;
    if (first == last)
        return MatchTraitsT::no_match;

    typedef typename impl::strip_scanner<IteratorT>
        ::iterator_type plain_iter;
    plain_iter  i1 = impl::strip_scanner<IteratorT>::get(first);
    plain_iter  i2 = impl::strip_scanner<IteratorT>::get(last);

    unsigned count = 0;
    if (!impl::radix_traits<Radix>::isok(*i1))
        return MatchTraitsT::no_match;

    if (impl::extract<Radix>::uint(i1, i2, n, count)) {
        IteratorT s = first;
        first = i1;
        return MatchTraitsT::create_match(count, s, i1);
    }
    return MatchTraitsT::no_match; // return non-match if uint overflows
}

///////////////////////////////////////////////////////////////////////////////
//
//  int_parser class implementation
//
///////////////////////////////////////////////////////////////////////////////
template <typename T, const int Radix>
inline int_parser<T, Radix>::int_parser(T n_)
:   n(n_) {}

//////////////////////////////////
template <typename T, const int Radix>
template <typename IteratorT>
inline match
int_parser<T, Radix>::parse(IteratorT& first, IteratorT const& last) const
{
    n = 0;
    if (first == last)
        return match();

    typedef typename impl::strip_scanner<IteratorT>
        ::iterator_type plain_iter;
    plain_iter  i1 = impl::strip_scanner<IteratorT>::get(first);
    plain_iter  i2 = impl::strip_scanner<IteratorT>::get(last);

    unsigned    count;
    bool        neg = impl::extract_sign(i1, count);

    bool matched = impl::extract<Radix>::sint(i1, i2, n, count, neg);
    match hit = match(count);

    if (matched && hit)
    {
        if (neg) n = -n;
        first = i1;
        return match(hit.length() + count);
    }
    return match();
}

//////////////////////////////////
template <typename T, const int Radix>
template <typename IteratorT, typename MatchTraitsT>
inline typename MatchTraitsT::match_t
int_parser<T, Radix>::ast_parse(IteratorT& first, IteratorT const& last, 
        MatchTraitsT const& mt) const
{
    n = 0;
    if (first == last)
        return MatchTraitsT::no_match;

    typedef typename impl::strip_scanner<IteratorT>
        ::iterator_type plain_iter;
    plain_iter  i1 = impl::strip_scanner<IteratorT>::get(first);
    plain_iter  i2 = impl::strip_scanner<IteratorT>::get(last);

    unsigned    count;
    bool        neg = impl::extract_sign(i1, count);

    bool matched = impl::extract<Radix>::sint(i1, i2, n, count, neg);
    typename MatchTraitsT::match_t hit = 
        MatchTraitsT::create_match(count, first, i1);

    if (matched && hit)
    {
        if (neg) n = -n;
        IteratorT s = first;
        first = i1;
        return hit;
    }
    return MatchTraitsT::no_match;
}

///////////////////////////////////////////////////////////////////////////////
namespace impl {

    template <typename T> struct real_traits {};
    template <typename T> struct real_traits<T&> : public real_traits<T> {};

#ifdef __MWERKS__

    //////////////////////////////////
    template <> struct real_traits<float> {

        static float pow(float x, float y)
        { return std::powf(x, y); }
    };

    //////////////////////////////////
    template <> struct real_traits<double> {

        static double pow(double x, double y)
        { return std::pow(x, y); }
    };

    //////////////////////////////////
    template <> struct real_traits<long double> {

        static double pow(long double x, long double y)
        { return std::powl(x, y); }
    };

#else

    //////////////////////////////////
    template <> struct real_traits<float> {

        static float pow(float x, float y)
        { return static_cast<float>(std::pow(x, y)); }
    };

    //////////////////////////////////
    template <> struct real_traits<double> {

        static double pow(double x, double y)
        { return std::pow(x, y); }
    };

    //////////////////////////////////
    template <> struct real_traits<long double> {

        static double pow(long double x, long double y)
        { return std::pow(x, y); }
    };

#endif
} // namespace impl

///////////////////////////////////////////////////////////////////////////////
//
//  ureal_parser class implementation
//
///////////////////////////////////////////////////////////////////////////////
template <typename T, typename TraitsT>
inline ureal_parser<T, TraitsT>::ureal_parser(T n_)
:   n(n_) {}

//////////////////////////////////
template <typename FloatT, typename TraitsT>
template <typename IteratorT>
match
ureal_parser<FloatT, TraitsT>::parse(IteratorT& first, 
        IteratorT const& last) const
{
    n = 0;
    if (first == last)
        return match();

    typedef typename impl::strip_scanner<IteratorT>
        ::iterator_type plain_iter;
    plain_iter  i1 = impl::strip_scanner<IteratorT>::get(first);
    plain_iter  i2 = impl::strip_scanner<IteratorT>::get(last);
    plain_iter    ittemp = i1;
    unsigned    counttemp = 0;

    if (!(isdigit(*ittemp) || 
                impl::traits_cmp<TraitsT>::is_ds(ittemp, i2, counttemp)))
        return match();

    unsigned    count = 0;
    unsigned    frac = 0;

    ittemp = i1;
    if (!impl::traits_cmp<TraitsT>::is_ds(ittemp, i2, count))
    {
        assert(0 == count);
        assert(ittemp == i1);

        //  Extract the digits before the decimal point: nnn.
        if (!impl::extract<10, TraitsT>::uint(ittemp, i2, n, count))
            return match();        // overflow, return non-match
        if (i1 == i2)
        {
            first = i1;
            return match(count);
        }
        i1 = ittemp;            // advance start position
    } 
    else
        count = 0;

    if (impl::traits_cmp<TraitsT>::is_ds(i1, i2, count))
    {
        //  Extract the digits after the decimal point .nnn
        if (!impl::extract<10>::uint(i1, i2, n, frac))
            return match();        // overflow, return non-match
        count += frac;
    }

// Extract the exponent
    match   hit(count);
    int     exp = 0;

    if (impl::traits_cmp<TraitsT>::is_exp(i1, i2, count)) {
        hit = match(count);
        if (match mexp = int_p[ref(exp)].parse(i1, i2))
            hit += mexp;
    }

    typedef typename remove_ref<FloatT>::type ref_type;
    typedef typename remove_wrap<ref_type>::type arg_type;

    n *= impl::real_traits<arg_type>::pow(10, exp - int(frac));
    first = i1;
    return hit;
}

//////////////////////////////////
template <typename FloatT, typename TraitsT>
template <typename IteratorT, typename MatchTraitsT>
inline typename MatchTraitsT::match_t
ureal_parser<FloatT, TraitsT>::ast_parse(IteratorT& first, 
        IteratorT const& last, MatchTraitsT const& mt) const
{
    n = 0;
    if (first == last)
        return MatchTraitsT::no_match;

    typedef typename impl::strip_scanner<IteratorT>
        ::iterator_type plain_iter;
    plain_iter  i1 = impl::strip_scanner<IteratorT>::get(first);
    plain_iter  i2 = impl::strip_scanner<IteratorT>::get(last);
    plain_iter    ittemp = i1;
    unsigned    counttemp = 0;

    if (!(isdigit(*ittemp) || 
                impl::traits_cmp<TraitsT>::is_ds(ittemp, i2, counttemp)))
        return MatchTraitsT::no_match;

    unsigned    count = 0;
    unsigned    frac = 0;

    ittemp = i1;
    if (!impl::traits_cmp<TraitsT>::is_ds(ittemp, i2, count))
    {
        assert(0 == count);
        assert(ittemp == i1);

        //  Extract the digits before the decimal point: nnn.
        if (!impl::extract<10, TraitsT>::uint(ittemp, i2, n, count))
            return MatchTraitsT::no_match;
        if (i1 == i2)
        {
            typename MatchTraitsT::match_t rval = 
                MatchTraitsT::create_match(count, first, i1);
            first = i1;
            return rval;
        }
        i1 = ittemp;            // advance start position
    } 
    else
        count = 0;

    if (impl::traits_cmp<TraitsT>::is_ds(i1, i2, count))
    {
        //  Extract the digits after the decimal point .nnn
        if (!impl::extract<10>::uint(i1, i2, n, frac))
            return MatchTraitsT::no_match;
        count += frac;
    }

// Extract the exponent
    typename MatchTraitsT::match_t hit = 
        MatchTraitsT::create_match(count, first, i1);
    int     exp = 0;

    if (impl::traits_cmp<TraitsT>::is_exp(i1, i2, count)) {
        typename MatchTraitsT::match_t temphit = 
            MatchTraitsT::create_match(count, first, i1);
        MatchTraitsT::extend(hit, temphit);
        if (typename MatchTraitsT::match_t mexp = 
                int_p[ref(exp)].ast_parse(i1, i2, mt))
        {
            MatchTraitsT::extend(hit, mexp);
        }
    }

    typedef typename remove_ref<FloatT>::type ref_type;
    typedef typename remove_wrap<ref_type>::type arg_type;

    n *= impl::real_traits<arg_type>::pow(10, exp - int(frac));
    first = i1;
    return hit;
}

///////////////////////////////////////////////////////////////////////////////
//
//  real_parser class implementation
//
///////////////////////////////////////////////////////////////////////////////
template <typename FloatT, typename TraitsT>
inline real_parser<FloatT, TraitsT>::real_parser(FloatT n_)
:   n(n_) {}

//////////////////////////////////
template <typename FloatT, typename TraitsT>
template <typename IteratorT>
inline match
real_parser<FloatT, TraitsT>::parse(IteratorT& first, 
        IteratorT const& last) const
{
    n = 0;
    if (first == last)
        return match();

    typedef typename impl::strip_scanner<IteratorT>
        ::iterator_type plain_iter;
    plain_iter  i1 = impl::strip_scanner<IteratorT>::get(first);
    plain_iter  i2 = impl::strip_scanner<IteratorT>::get(last);

    unsigned    count;
    bool        neg = impl::extract_sign(i1, count);
    ureal_parser_gen<TraitsT> parse_ureal;

    if (match hit = parse_ureal[ref(n)].parse(i1, i2))
    {
        if (neg) n = -n;
        first = i1;
        return match(hit.length() + count);
    }
    return match();
}

//////////////////////////////////
template <typename FloatT, typename TraitsT>
template <typename IteratorT, typename MatchTraitsT>
inline typename MatchTraitsT::match_t
real_parser<FloatT, TraitsT>::ast_parse(IteratorT& first, 
        IteratorT const& last, MatchTraitsT const& mt) const
{
    n = 0;
    if (first == last)
        return MatchTraitsT::no_match;

    typedef typename impl::strip_scanner<IteratorT>
        ::iterator_type plain_iter;
    plain_iter  i1 = impl::strip_scanner<IteratorT>::get(first);
    plain_iter  i2 = impl::strip_scanner<IteratorT>::get(last);

    unsigned    count;
    bool        neg = impl::extract_sign(i1, count);
    ureal_parser_gen<TraitsT> parse_ureal;

    if (typename MatchTraitsT::match_t hit = 
            parse_ureal[ref(n)].ast_parse(i1, i2, mt))
    {
        if (neg) n = -n;
        first = i1;
        return hit;
    }
    return MatchTraitsT::no_match;
}

///////////////////////////////////////////////////////////////////////////////
}   //  namespace Spirit

#endif
