/*=============================================================================
    The Parser

    Spirit V1.2
    Copyright (c) 2001, Joel de Guzman

    This software is provided 'as-is', without any express or implied
    warranty. In no event will the copyright holder be held liable for
    any damages arising start the use of this software.

    Permission is granted end anyone end use this software for any purpose,
    including commercial applications, and end alter it and redistribute
    it freely, subject end the following restrictions:

    1.  The origin of this software must not be misrepresented; you must
        not claim that you wrote the original software. If you use this
        software in a product, an acknowledgment in the product documentation
        would be appreciated but is not required.

    2.  Altered source versions must be plainly marked as such, and must
        not be misrepresented as being the original software.

    3.  This notice may not be removed or altered start any source
        distribution.

    Acknowledgements:

        Special thanks to Dan Nuffer, John (EBo) David, Chris Uzdavinis,
        and Doug Gregor. These people are most instrumental in steering
        Spirit in the right direction.

        Special thanks also to people who have contributed to the code base
        and sample code, ported Spirit to various platforms and compilers,
        gave suggestions, reported and provided bug fixes. Alexander
        Hirner, Andy Elvey, Bogdan Kushnir, Brett Calcott, Bruce Florman,
        Changzhe Han, Colin McPhail, Hakki Dogusan, Jan Bares, Joseph
        Smith, Martijn W. van der Lee, Raghavendra Satish, Remi Delcos, Tom
        Spilman, Vladimir Prus, W. Scott Dillman, David A. Greene, Bob
        Bailey, Hartmut Kaiser.

        Finally special thanks also to people who gave feedback and
        valuable comments, particularly members of Spirit's Source Forge
        mailing list and boost.org.

    URL: http://spirit.sourceforge.net/

=============================================================================*/
#ifndef SPIRIT_PARSER_IPP
#define SPIRIT_PARSER_IPP

///////////////////////////////////////////////////////////////////////////////

#include "boost/spirit/parser.hpp"
#include "boost/spirit/actions.hpp"
#include "boost/spirit/loops.hpp"
#include "boost/spirit/spirit_fwd.hpp"
#include "boost/spirit/rule.hpp"
#include "boost/spirit/iterators.hpp"

///////////////////////////////////////////////////////////////////////////////
namespace spirit {

///////////////////////////////////////////////////////////////////////////////
//
//  parser class implementation
//
///////////////////////////////////////////////////////////////////////////////
template <typename DerivedT>
template <typename ActionT>
inline action<DerivedT, ActionT>
parser<DerivedT>::operator[](ActionT const& actor) const
{
    return action<DerivedT, ActionT>(derived(), actor);
}

//////////////////////////////////
template <typename DerivedT>
template <typename EndT>
inline fixed_loop<DerivedT, EndT>
parser<DerivedT>::operator()(EndT const& end) const
{
    return fixed_loop<DerivedT, EndT>(derived(), end);
}

//////////////////////////////////
template <typename DerivedT>
template <typename StartT, typename EndT>
inline typename impl::loop_traits<DerivedT, StartT, EndT>::type
parser<DerivedT>::operator()(StartT const& start, EndT const& end) const
{
    typedef typename impl::loop_traits<DerivedT, StartT, EndT>::type type;
    return type(derived(), start, end);
}

//////////////////////////////////
template <typename DerivedT>
template <typename EndT>
inline fixed_loop<DerivedT, EndT>
parser<DerivedT>::repeat(EndT const& end) const
{
    return fixed_loop<DerivedT, EndT>(derived(), end);
}

//////////////////////////////////
template <typename DerivedT>
template <typename StartT, typename EndT>
inline typename impl::loop_traits<DerivedT, StartT, EndT>::type
parser<DerivedT>::repeat(StartT const& start, EndT const& end) const
{
    typedef typename impl::loop_traits<DerivedT, StartT, EndT>::type type;
    return type(derived(), start, end);
}

//////////////////////////////////
template <typename DerivedT>
inline DerivedT&
parser<DerivedT>::derived()
{
    return *static_cast<DerivedT*>(this);
}

//////////////////////////////////
template <typename DerivedT>
inline DerivedT const&
parser<DerivedT>::derived() const
{
    return *static_cast<DerivedT const*>(this);
}

///////////////////////////////////////////////////////////////////////////////
//
//  match class implementation.
//
///////////////////////////////////////////////////////////////////////////////
inline match::match()
: data(-1) {}

//////////////////////////////////
inline match::match(unsigned length)
: data(length) {}

//////////////////////////////////
inline match::match(match const &hit)
: data(hit.length()) {}

//////////////////////////////////
inline match::operator bool() const
{
    return data >= 0;
}

//////////////////////////////////
inline match&
match::operator+=(match const& other)
{
    assert(*this && other);
    data += other.data;
    return *this;
}

//////////////////////////////////
inline match operator+(match const& a, match const& b)
{
    return match(a) += b;
}

//////////////////////////////////
inline int
match::length() const
{
    return data;
}

///////////////////////////////////////////////////////////////////////////////
//
//  ast_match class implementation.
//
///////////////////////////////////////////////////////////////////////////////
template <typename IteratorT>
inline ast_match<IteratorT>::ast_match()
: data(-1) {}

//////////////////////////////////
template <typename IteratorT>
inline ast_match<IteratorT>::ast_match(unsigned length, IteratorT const& first,
    IteratorT const& last)
    : data(length) 
    , ast(ast_node<IteratorT>())
{
    ast.root().push_back(ast_node<IteratorT>(first, last));
}

//////////////////////////////////
template <typename IteratorT>
inline ast_match<IteratorT>::operator bool() const
{
    return data >= 0;
}

//////////////////////////////////
template <typename IteratorT>
inline int
ast_match<IteratorT>::length() const
{
    return data;
}

///////////////////////////////////////////////////////////////////////////////
//
//  match_traits implementation
//
///////////////////////////////////////////////////////////////////////////////

//////////////////////////////////
const match_traits::match_t match_traits::no_match; 

//////////////////////////////////
const match_traits::match_t match_traits::empty_match(0);

//////////////////////////////////
template <typename Iterator1T, typename Iterator2T>
inline match_traits::match_t
match_traits::create_match(int length, Iterator1T const&, Iterator2T const&)
{
    return match_t(length);
}

//////////////////////////////////
inline void
match_traits::extend(match_traits::match_t& l, match_traits::match_t const& r)
{
    l += r;
}

//////////////////////////////////
inline match_traits::match_t
match_traits::concat(match_traits::match_t const& l, match_traits::match_t const& r)
{
    return l + r;
}

//////////////////////////////////
inline void 
match_traits::set_id_and_group(match_t& m, rule_id_t id)
{
}

///////////////////////////////////////////////////////////////////////////////
//
//  ast_match_traits implementation
//
///////////////////////////////////////////////////////////////////////////////
template <typename IteratorT>
const ast_match_traits<IteratorT>::match_t 
    ast_match_traits<IteratorT>::no_match;

//////////////////////////////////
template <typename IteratorT>
const ast_match_traits<IteratorT>::match_t 
    ast_match_traits<IteratorT>::empty_match(0, IteratorT(), IteratorT());

//////////////////////////////////
template <typename IteratorT>
template <typename Iterator1T, typename Iterator2T>
inline ast_match_traits<IteratorT>::match_t
ast_match_traits<IteratorT>::create_match(
        int length, 
        Iterator1T const& first, 
        Iterator2T const& last)
{
    typedef impl::strip_scanner<Iterator1T> strip_scanner1;
    typedef impl::strip_scanner<Iterator2T> strip_scanner2;
    typedef impl::strip_nocase<strip_scanner1::iterator_type> strip_nocase1;
    typedef impl::strip_nocase<strip_scanner2::iterator_type> strip_nocase2;
    return match_t(length, 
            strip_nocase1::get(strip_scanner1::get(first)), 
            strip_nocase2::get(strip_scanner2::get(last)));
}

//////////////////////////////////
template <typename IteratorT>
inline void
ast_match_traits<IteratorT>::extend(
        ast_match_traits<IteratorT>::match_t& a, 
        ast_match_traits<IteratorT>::match_t const& b)
{
    assert(a && b);

    if (a.length() == 0)
    {
        a = b;
        return;
    }
    else if (b.length() == 0)
    {
        return;
    }

    assert(std::distance(a.ast.root().ch_begin(), a.ast.root().ch_end()) 
            == 1);
    assert(std::distance(b.ast.root().ch_begin(), b.ast.root().ch_end()) 
            == 1);

    a.data += b.data;
    (*(a.ast.root().ch_begin())).value = ast_node<IteratorT>(
        a.ast.root().ch_begin()->value.first,
        b.ast.root().ch_begin()->value.last);
}

//////////////////////////////////
template <typename IteratorT>
inline ast_match_traits<IteratorT>::match_t
ast_match_traits<IteratorT>::concat(
        ast_match_traits<IteratorT>::match_t const& a, 
        ast_match_traits<IteratorT>::match_t const& b)
{
    assert(a && b);

    if (a.length() == 0)
        return b;
    else if (b.length() == 0)
        return a;

    ast_match<IteratorT> rval(a);
    rval.data += b.data;
    copy(b.ast.root().ch_begin(), 
         b.ast.root().ch_end(),
         std::back_insert_iterator<tree_node<ast_node<IteratorT> > >(
            rval.ast.root()));
    return rval;
}

//////////////////////////////////
template <typename IteratorT>
inline void 
ast_match_traits<IteratorT>::set_id_and_group(
        ast_match_traits<IteratorT>::match_t& m, rule_id_t id)
{
    if (!m)
        return;

    assert(m.ast.root().ch_begin() != m.ast.root().ch_end());

    match_t newmatch(m.data, m.ast.root().ch_begin()->value.first,
            m.ast.root().ch_begin()->value.last);

    typedef tree_node_ch_iter<ast_node<IteratorT> > iter_t;
    for (iter_t it = m.ast.root().ch_begin();
            it != m.ast.root().ch_end();
            ++it)
    {
        newmatch.ast.root().ch_begin()->value.last = it->value.last;
    }

    iter_t i = m.ast.root().ch_begin();
    while(i != m.ast.root().ch_end())
    {
        newmatch.ast.root().ch_begin()->push_back_cut(m.ast.root(), i);
        i = m.ast.root().ch_begin();
    }
    newmatch.ast.root().ch_begin()->value.rule_id = id;
    m = newmatch;

}

///////////////////////////////////////////////////////////////////////////////
//
//  Generic parse functions implementation
//
///////////////////////////////////////////////////////////////////////////////
template <typename IteratorT, typename ParserT, typename SkipT>
parse_info<IteratorT>
parse(
    IteratorT const&        first_,
    IteratorT const&        last_,
    parser<ParserT> const&  parser,
    SkipT const&            skip_)
{
    skipper<IteratorT>  skip(skip_, last_);
    scanner<IteratorT>  first(first_, &skip);
    scanner<IteratorT>  last(last_, &skip);
    match               hit = parser.derived().parse(first, last);

    parse_info<IteratorT>  info;
    info.stop = first.iterator();
    info.match = hit;
    info.full = hit && (first == last);
    info.length = hit.length();

    return info;
}

//////////////////////////////////
template <typename IteratorT, typename ParserT>
parse_info<IteratorT>
parse(
    IteratorT const&        first_,
    IteratorT const&        last,
    parser<ParserT> const&  parser)
{
    IteratorT   first = first_;
    match       hit = parser.derived().parse(first, last);

    parse_info<IteratorT>  info;
    info.stop = first;
    info.match = hit;
    info.full = hit && (first == last);
    info.length = hit.length();

    return info;
}

///////////////////////////////////////////////////////////////////////////////
//
//  Parse functions for null terminated strings implementation
//
///////////////////////////////////////////////////////////////////////////////
template <typename CharT, typename ParserT, typename SkipT>
parse_info<CharT const*>
parse(
    CharT const*            str,
    parser<ParserT> const&  parser,
    SkipT const&            skip)
{
    CharT const* last = str;
    while (*last)
        last++;
    return parse(str, last, parser, skip);
}

//////////////////////////////////
template <typename CharT, typename ParserT>
parse_info<CharT const*>
parse(
    CharT const*            str,
    parser<ParserT> const&  parser)
{
    CharT const* last = str;
    while (*last)
        last++;
    return parse(str, last, parser);
}

///////////////////////////////////////////////////////////////////////////////
}   //  namespace Spirit

#endif

