/*=============================================================================
    The Production rule

    Spirit V1.2
    Copyright (c) 2001, Joel de Guzman

    This software is provided 'as-is', without any express or implied
    warranty. In no event will the copyright holder be held liable for
    any damages arising from the use of this software.

    Permission is granted to anyone to use this software for any purpose,
    including commercial applications, and to alter it and redistribute
    it freely, subject to the following restrictions:

    1.  The origin of this software must not be misrepresented; you must
        not claim that you wrote the original software. If you use this
        software in a product, an acknowledgment in the product documentation
        would be appreciated but is not required.

    2.  Altered source versions must be plainly marked as such, and must
        not be misrepresented as being the original software.

    3.  This notice may not be removed or altered from any source
        distribution.

    Acknowledgements:

        Special thanks to Dan Nuffer, John (EBo) David, Chris Uzdavinis,
        and Doug Gregor. These people are most instrumental in steering
        Spirit in the right direction.

        Special thanks also to people who have contributed to the code base
        and sample code, ported Spirit to various platforms and compilers,
        gave suggestions, reported and provided bug fixes. Alexander
        Hirner, Andy Elvey, Bogdan Kushnir, Brett Calcott, Bruce Florman,
        Changzhe Han, Colin McPhail, Hakki Dogusan, Jan Bares, Joseph
        Smith, Martijn W. van der Lee, Raghavendra Satish, Remi Delcos, Tom
        Spilman, Vladimir Prus, W. Scott Dillman, David A. Greene, Bob
        Bailey, Hartmut Kaiser.

        Finally special thanks also to people who gave feedback and
        valuable comments, particularly members of Spirit's Source Forge
        mailing list and boost.org.

    URL: http://spirit.sourceforge.net/

======================================================================================*/
#ifndef SPIRIT_RULE_IPP
#define SPIRIT_RULE_IPP

///////////////////////////////////////////////////////////////////////////////

#include <cassert>
#include "boost/spirit/rule.hpp"
#include "boost/spirit/primitives.hpp"

///////////////////////////////////////////////////////////////////////////////
namespace spirit {

///////////////////////////////////////////////////////////////////////////////
namespace impl {

    ///////////////////////////////////////////////////////////////////////////////
    //
    //  abstract_parser class
    //
    ///////////////////////////////////////////////////////////////////////////////
    template <typename IteratorT, typename MatchTraitsT>
    class abstract_parser {

    public:

        abstract_parser();
        virtual ~abstract_parser();

        virtual match
        parse(IteratorT& first, IteratorT const& last) const = 0;

        virtual typename MatchTraitsT::match_t
        ast_parse(IteratorT& first, IteratorT const& last, 
                MatchTraitsT const& mt) const = 0;
    };

    //////////////////////////////////
    template <typename IteratorT, typename MatchTraitsT>
    inline abstract_parser<IteratorT, MatchTraitsT>::abstract_parser() {}

    //////////////////////////////////
    template <typename IteratorT, typename MatchTraitsT>
    abstract_parser<IteratorT, MatchTraitsT>::~abstract_parser() {}

    ///////////////////////////////////////////////////////////////////////////////
    //
    //  concrete_parser class
    //
    ///////////////////////////////////////////////////////////////////////////////
    template <typename ParserT, typename IteratorT, typename MatchTraitsT>
    class concrete_parser
    :   public ParserT,
        public abstract_parser<IteratorT, MatchTraitsT> {

    public:

        concrete_parser(ParserT const& parser);
        virtual ~concrete_parser();

        virtual match
        parse(IteratorT& first, IteratorT const& last) const;

        virtual typename MatchTraitsT::match_t
        ast_parse(IteratorT& first, IteratorT const& last, 
                MatchTraitsT const& mt) const;
    };

    //////////////////////////////////
    template <typename ParserT, typename IteratorT, typename MatchTraitsT>
    inline concrete_parser<ParserT, IteratorT, MatchTraitsT>::concrete_parser(
        ParserT const& parser)
    :   ParserT(parser) {}

    //////////////////////////////////
    template <typename ParserT, typename IteratorT, typename MatchTraitsT>
    concrete_parser<ParserT, IteratorT, MatchTraitsT>::~concrete_parser() {}

    //////////////////////////////////
    template <typename ParserT, typename IteratorT, typename MatchTraitsT>
    match
    concrete_parser<ParserT, IteratorT, MatchTraitsT>::
        parse(IteratorT& first, IteratorT const& last) const
    {
        return ParserT::parse(first, last);
    }

    //////////////////////////////////
    template <typename ParserT, typename IteratorT, typename MatchTraitsT>
    typename MatchTraitsT::match_t
    concrete_parser<ParserT, IteratorT, MatchTraitsT>::
        ast_parse(IteratorT& first, IteratorT const& last, 
                MatchTraitsT const& mt) const
    {
        return ParserT::ast_parse(first, last, mt);
    }

    ///////////////////////////////////////////////////////////////////////////////
    //
    //  rule_alias class
    //
    ///////////////////////////////////////////////////////////////////////////////
    template <typename IteratorT, typename MatchTraitsT, typename DerivedT>
    class rule_alias : public abstract_parser<IteratorT, MatchTraitsT> {

    public:

        rule_alias(base_rule<IteratorT, MatchTraitsT, DerivedT> const& alias);
        virtual ~rule_alias();

        virtual match
        parse(IteratorT& first, IteratorT const& last) const;

        virtual typename MatchTraitsT::match_t
        ast_parse(IteratorT& first, IteratorT const& last, 
                MatchTraitsT const& mt) const;

    private:

        base_rule<IteratorT, MatchTraitsT, DerivedT> const& alias;
    };

    //////////////////////////////////
    template <typename IteratorT, typename MatchTraitsT, typename DerivedT>
    inline rule_alias<IteratorT, MatchTraitsT, DerivedT>::rule_alias(
        base_rule<IteratorT, MatchTraitsT, DerivedT> const& alias_)
    :   alias(alias_) {}

    //////////////////////////////////
    template <typename IteratorT, typename MatchTraitsT, typename DerivedT>
    rule_alias<IteratorT, MatchTraitsT, DerivedT>::~rule_alias() {}

    //////////////////////////////////
    template <typename IteratorT, typename MatchTraitsT, typename DerivedT>
    match
    rule_alias<IteratorT, MatchTraitsT, DerivedT>::
        parse(IteratorT& first, IteratorT const& last) const
    {
        return alias.parse(first, last);
    }

    //////////////////////////////////
    template <typename IteratorT, typename MatchTraitsT, typename DerivedT>
    typename MatchTraitsT::match_t
    rule_alias<IteratorT, MatchTraitsT, DerivedT>::
        ast_parse(IteratorT& first, IteratorT const& last, 
                MatchTraitsT const& mt) const
    {
        return alias.ast_parse(first, last, mt);
    }

    ///////////////////////////////////////////////////////////////////////////////
    //
    //  alt_parser class
    //
    ///////////////////////////////////////////////////////////////////////////////
    template <typename ParserT, typename IteratorT, typename MatchTraitsT>
    class alt_parser : public abstract_parser<IteratorT, MatchTraitsT> {

    public:

        alt_parser(abstract_parser<IteratorT, MatchTraitsT>* left, 
                ParserT const& right);
        virtual ~alt_parser();

        virtual match
        parse(IteratorT& first, IteratorT const& last) const;

        virtual typename MatchTraitsT::match_t
        ast_parse(IteratorT& first, IteratorT const& last, 
                MatchTraitsT const& mt) const;

    private:

        abstract_parser<IteratorT, MatchTraitsT>* left;
        ParserT right;
    };

    //////////////////////////////////
    template <typename ParserT, typename IteratorT, typename MatchTraitsT>
    inline alt_parser<ParserT, IteratorT, MatchTraitsT>::alt_parser(
        abstract_parser<IteratorT, MatchTraitsT>* left_, ParserT const& right_)
    :   left(left_),
        right(right_) {}

    //////////////////////////////////
    template <typename ParserT, typename IteratorT, typename MatchTraitsT>
    alt_parser<ParserT, IteratorT, MatchTraitsT>::~alt_parser()
    {
        delete left;
    }

    //////////////////////////////////
    template <typename ParserT, typename IteratorT, typename MatchTraitsT>
    match
    alt_parser<ParserT, IteratorT, MatchTraitsT>::
        parse(IteratorT& first, IteratorT const& last) const
    {
        if (match hit = left->parse(first, last))
            return hit;
        return right.parse(first, last);
    }

    //////////////////////////////////
    template <typename ParserT, typename IteratorT, typename MatchTraitsT>
    typename MatchTraitsT::match_t
    alt_parser<ParserT, IteratorT, MatchTraitsT>::
        ast_parse(IteratorT& first, IteratorT const& last, 
                MatchTraitsT const& mt) const
    {
        if (typename MatchTraitsT::match_t hit = 
                left->ast_parse(first, last, mt))
            return hit;
        return right.ast_parse(first, last, mt);
    }

} // namespace impl

///////////////////////////////////////////////////////////////////////////////
//
//  rule class implementation.
//
///////////////////////////////////////////////////////////////////////////////
template <typename IteratorT, typename MatchTraitsT, typename DerivedT>
inline base_rule<IteratorT, MatchTraitsT, DerivedT>::base_rule()
:   meta(0) {}

//////////////////////////////////
template <typename IteratorT, typename MatchTraitsT, typename DerivedT>
inline base_rule<IteratorT, MatchTraitsT, DerivedT>::~base_rule()
{
    delete meta;
}

//////////////////////////////////
template <typename IteratorT, typename MatchTraitsT, typename DerivedT>
inline base_rule<IteratorT, MatchTraitsT, DerivedT>
        ::base_rule(base_rule const& other)
:   parser<DerivedT>()
,   meta(new impl::rule_alias<IteratorT, MatchTraitsT, DerivedT>(other))
{
}

//////////////////////////////////
template <typename IteratorT, typename MatchTraitsT, typename DerivedT>
template <typename ParserT>
inline base_rule<IteratorT, MatchTraitsT, DerivedT>
        ::base_rule(ParserT const& parser)
:   meta(new impl::concrete_parser<ParserT, IteratorT, MatchTraitsT>(parser))
{
}

//////////////////////////////////
template <typename IteratorT, typename MatchTraitsT, typename DerivedT>
inline base_rule<IteratorT, MatchTraitsT, DerivedT>&
base_rule<IteratorT, MatchTraitsT, DerivedT>
    ::operator = (base_rule<IteratorT, MatchTraitsT, DerivedT> const& other)
{
    if (this != &other)
    {
        if (meta == 0)
            meta = new impl::rule_alias<IteratorT, MatchTraitsT, DerivedT>(other);
        else
            meta = new impl::alt_parser<DerivedT, IteratorT, MatchTraitsT>(meta, 
                    other);
    }
    return *this;
}

//////////////////////////////////
template <typename IteratorT, typename MatchTraitsT, typename DerivedT>
template <typename ParserT>
inline base_rule<IteratorT, MatchTraitsT, DerivedT>&
base_rule<IteratorT, MatchTraitsT, DerivedT>
    ::operator = (ParserT const& parser)
{
    if (meta == 0)
        meta = new impl::concrete_parser<ParserT, IteratorT, MatchTraitsT>(
                parser);
    else
        meta = new impl::alt_parser<ParserT, IteratorT, MatchTraitsT>(meta, 
                parser);
    return *this;
}

//////////////////////////////////
#ifdef SPIRIT_DEBUG
template <typename IteratorT>
void
print_rule_info(
    bool                hit,
    int                 level,
    bool                close,
    std::string const&  name,
    IteratorT           first,
    IteratorT           last)
{
    if (!name.empty())
    {
        for (int i = 0; i < level; ++i)
            SPIRIT_DEBUG_OUT << "  ";
        if (close)
        {
            if (hit)
                SPIRIT_DEBUG_OUT << "/";
            else
                SPIRIT_DEBUG_OUT << "#";
        }
        SPIRIT_DEBUG_OUT << name << ":\t\"";
        typedef impl::strip_scanner<IteratorT> strip_scanner;
        typename strip_scanner::iterator_type
            iter = strip_scanner::get(first);
        typename strip_scanner::iterator_type
            ilast = strip_scanner::get(last);
        for (int j = 0; j < 20; ++j)
        {
            if (iter == ilast)
                break;

            switch (*iter) {

                case '\r': SPIRIT_DEBUG_OUT << "\\r";  break;
                case '\n': SPIRIT_DEBUG_OUT << "\\n";  break;
                default: SPIRIT_DEBUG_OUT << *iter;    break;
            }
            ++iter;
        }
        SPIRIT_DEBUG_OUT << " \"\n";
    }
}
#endif

template <typename IteratorT, typename MatchTraitsT, typename DerivedT>
inline match
base_rule<IteratorT, MatchTraitsT, DerivedT>::parse(IteratorT& first, 
        IteratorT const& last) const
{
#if defined(SPIRIT_DEBUG) && (SPIRIT_DEBUG_LEVEL >= SPIRIT_DEBUG_LEVEL_LEVEL2)

    print_rule_info(false, debug.level, false, debug.name(), first, last);

    int save_level = debug.level;
    debug.level++;
    match hit = meta ? meta->parse(first, last) : match();
    debug.level = save_level;

    print_rule_info(hit, debug.level, true, debug.name(), first, last);
    return hit;

#else
    return meta ? meta->parse(first, last) : match();
#endif
}

template <typename IteratorT, typename MatchTraitsT, typename DerivedT>
inline typename MatchTraitsT::match_t
base_rule<IteratorT, MatchTraitsT, DerivedT>::ast_parse(IteratorT& first, 
        IteratorT const& last, MatchTraitsT const& mt) const
{
#ifdef SPIRIT_DEBUG

    print_rule_info(false, debug.level, false, debug.name(), first, last);

    int save_level = debug.level;
    debug.level++;

    typename MatchTraitsT::match_t hit;
    if (meta)
    {
        hit = meta->ast_parse(first, last, mt);
        MatchTraitsT::set_id_and_group(hit, this);
    }
    else
    {
        hit = MatchTraitsT::no_match;
    }

    debug.level = save_level;

    print_rule_info(hit, debug.level, true, debug.name(), first, last);
    return hit;

#else
    if (meta)
    {
        typename MatchTraitsT::match_t hit = meta->ast_parse(first, last, mt);
        MatchTraitsT::set_id_and_group(hit, this);
        return hit;
    }
    else
    {
        return MatchTraitsT::no_match;
    }
#endif
}

template <typename IteratorT, typename MatchTraitsT>
inline rule<IteratorT, MatchTraitsT>::rule()
{
}

template <typename IteratorT, typename MatchTraitsT>
inline rule<IteratorT, MatchTraitsT>::rule(rule const& other) :
    base_rule<IteratorT, MatchTraitsT, rule<IteratorT, MatchTraitsT> >(other)
{
}

template <typename IteratorT, typename MatchTraitsT>
template <typename ParserT>
inline rule<IteratorT, MatchTraitsT>::rule(ParserT const& parser) :
    base_rule<IteratorT, MatchTraitsT, rule<IteratorT, MatchTraitsT> >(parser)
{
}

template <typename IteratorT, typename MatchTraitsT>
inline rule<IteratorT, MatchTraitsT>&
rule<IteratorT, MatchTraitsT>::operator = (rule const& other)
{
    if (this != &other)
        *static_cast<base_rule_t *>(this) = other;
    return *this;
}

template <typename IteratorT, typename MatchTraitsT>
template <typename ParserT>
inline rule<IteratorT, MatchTraitsT>&
rule<IteratorT, MatchTraitsT>::operator=(ParserT const& parser)
{
    *static_cast<base_rule_t *>(this) = parser;
    return *this;
}

///////////////////////////////////////////////////////////////////////////////
//
//  skipper class implementation.
//
///////////////////////////////////////////////////////////////////////////////
template <typename IteratorT, typename MatchTraitsT>
template <typename ParserT>
inline skipper<IteratorT, MatchTraitsT>::skipper(
    ParserT const&      skip_rule_,
    IteratorT const&    last_)
:
    last(last_),
    skip_rule(skip_rule_)
{
}

//////////////////////////////////
template <typename IteratorT, typename MatchTraitsT>
inline skipper<IteratorT, MatchTraitsT>::skipper(IteratorT const& last_)
:
    last(last_),
    skip_rule(space)
{
}

//////////////////////////////////
template <typename IteratorT, typename MatchTraitsT>
inline void
skipper<IteratorT, MatchTraitsT>::skip(IteratorT& current) const
{
    while (skip_rule.parse(current, last))
    {}
}

///////////////////////////////////////////////////////////////////////////////
}   //  namespace Spirit

#endif
