/*=============================================================================
    Utilities

    Spirit V1.2
    Copyright (c) 2001, Daniel Nuffer
    Copyright (c) 2001, Hartmut Kaiser

    This software is provided 'as-is', without any express or implied
    warranty. In no event will the copyright holder be held liable for
    any damages arising from the use of this software.

    Permission is granted to anyone to use this software for any purpose,
    including commercial applications, and to alter it and redistribute
    it freely, subject to the following restrictions:

    1.  The origin of this software must not be misrepresented; you must
        not claim that you wrote the original software. If you use this
        software in a product, an acknowledgment in the product documentation
        would be appreciated but is not required.

    2.  Altered source versions must be plainly marked as such, and must
        not be misrepresented as being the original software.

    3.  This notice may not be removed or altered from any source
        distribution.

    Acknowledgements:

        Special thanks to Dan Nuffer, John (EBo) David, Chris Uzdavinis,
        and Doug Gregor. These people are most instrumental in steering
        Spirit in the right direction.

        Special thanks also to people who have contributed to the code base
        and sample code, ported Spirit to various platforms and compilers,
        gave suggestions, reported and provided bug fixes. Alexander
        Hirner, Andy Elvey, Bogdan Kushnir, Brett Calcott, Bruce Florman,
        Changzhe Han, Colin McPhail, Hakki Dogusan, Jan Bares, Joseph
        Smith, Martijn W. van der Lee, Raghavendra Satish, Remi Delcos, Tom
        Spilman, Vladimir Prus, W. Scott Dillman, David A. Greene, Bob
        Bailey, Hartmut Kaiser.

        Finally special thanks also to people who gave feedback and
        valuable comments, particularly members of Spirit's Source Forge
        mailing list and boost.org.

    URL: http://spirit.sourceforge.net/

=============================================================================*/
#ifndef SPIRIT_UTILITIES_IPP
#define SPIRIT_UTILITIES_IPP

///////////////////////////////////////////////////////////////////////////////
#include <cassert>

#include "boost/spirit/utilities.hpp"

///////////////////////////////////////////////////////////////////////////////
//
//  These seem to creep in as macros when using gcc 2.95.2.
//
///////////////////////////////////////////////////////////////////////////////
#ifdef __GNUC__
#ifdef isdigit
#undef isdigit
#endif

#ifdef isxdigit
#undef isxdigit
#endif
#endif

///////////////////////////////////////////////////////////////////////////////
namespace spirit {

/*=============================================================================

    Escape char parser classes implementation [Dan Nuffer]

=============================================================================*/

namespace impl {

///////////////////////////////////////////////////////////////////////////////
//
//    escape_char_action_traits class used to determine the type of the
//    argument to the action functor.
//
///////////////////////////////////////////////////////////////////////////////

    template <typename T>
    struct escape_char_action_traits {
        typedef typename T::argument_type argument_type;
    };

    template <typename T>
    struct escape_char_action_traits<void (*)(T)> {
        typedef T argument_type;
    };

    template <typename T>
    struct escape_char_action_traits<reference_wrapper<T> > {
        typedef T argument_type;
    };

} // namespace impl

///////////////////////////////////////////////////////////////////////////////
//
//  escape_char_action class implementation
//
///////////////////////////////////////////////////////////////////////////////
template <typename ParserT, typename ActionT, unsigned long Flags>
inline escape_char_action<ParserT, ActionT, Flags>::escape_char_action(
    ParserT const& subject, ActionT const& actor_)
:   unary<ParserT>(subject),
    actor(actor_) {}

//////////////////////////////////
template <typename ParserT, typename ActionT, unsigned long Flags>
template <typename IteratorT>
inline match
escape_char_action<ParserT, ActionT, Flags>::
parse(IteratorT& first, IteratorT const& last) const
{
    // Actually decode the escape char.
    typedef typename impl::escape_char_action_traits<ActionT>
        ::argument_type char_t;

    if (first != last)
    {
        IteratorT s = first;
        if (match hit = this->subject().parse(s, last))
        {
            char_t unescaped;
            if (*first == '\\')
            {
                ++first;
                switch (*first)
                {
                    case 'b':   unescaped = '\b';   ++first; break;
                    case 't':   unescaped = '\t';   ++first; break;
                    case 'n':   unescaped = '\n';   ++first; break;
                    case 'f':   unescaped = '\f';   ++first; break;
                    case 'r':   unescaped = '\r';   ++first; break;
                    case '"':   unescaped = '"';    ++first; break;
                    case '\'':  unescaped = '\'';   ++first; break;
                    case '\\':  unescaped = '\\';   ++first; break;
                    case 'x': case 'X':
                        {
                            char_t hex = 0;
                            ++first;
                            char_t lim = 
                                std::numeric_limits<char_t>::max() >> 4;
                            while (first != last)
                            {
                                char_t c = *first;
                                if (hex > lim && std::isxdigit(c))
                                {
                                    return match(); // overflow detected
                                }
                                if (std::isdigit(c))
                                {
                                    hex <<= 4;
                                    hex |= c - '0';
                                    ++first;
                                }
                                else if (std::isxdigit(c))
                                {
                                    hex <<= 4;
                                    c = toupper(c);
                                    hex |= c - 'A' + 0xA;
                                    ++first;
                                }
                                else
                                {
                                    break; // reached the end of the number
                                }
                            }
                            unescaped = hex;
                        }
                        break;

                    case '0': case '1': case '2': case '3':
                    case '4': case '5': case '6': case '7':
                        {
                            char_t hex = 0;
                            char_t lim = 
                                std::numeric_limits<char_t>::max() >> 3;
                            while (first != last)
                            {
                                char_t c = *first;
                                if (hex > lim && (c >= '0' && c <= '7'))
                                {
                                    return match(); // overflow detected
                                }
                                if (c >= '0' && c <= '7')
                                {
                                    hex <<= 3;
                                    hex |= c - '0';
                                    ++first;
                                }
                                else
                                {
                                    break; // reached end of digits
                                }
                            }
                            unescaped = hex;
                        }

                        break;

                    default:
                        if (Flags & escape_flags::c_escapes)
                        {
                            return match();
                        }
                        else
                        {
                            unescaped = *first;
                            ++first;
                        }
                        break;
                }
            }
            else
            {
                unescaped = *first;
                ++first;
            }

            actor(unescaped);
            return hit;
        }
    }
    return match();
}

//////////////////////////////////
template <typename ParserT, typename ActionT, unsigned long Flags>
template <typename IteratorT, typename MatchTraitsT>
inline typename MatchTraitsT::match_t
escape_char_action<ParserT, ActionT, Flags>::
ast_parse(IteratorT& first, IteratorT const& last, MatchTraitsT const& mt) const
{
    // Actually decode the escape char.
    typedef typename impl::escape_char_action_traits<ActionT>
        ::argument_type char_t;

    if (first != last)
    {
        IteratorT s = first;
        if (typename MatchTraitsT::match_t hit = 
                this->subject().ast_parse(s, last, mt))
        {
            char_t unescaped;
            if (*first == '\\')
            {
                ++first;
                switch (*first)
                {
                    case 'b':   unescaped = '\b';   ++first; break;
                    case 't':   unescaped = '\t';   ++first; break;
                    case 'n':   unescaped = '\n';   ++first; break;
                    case 'f':   unescaped = '\f';   ++first; break;
                    case 'r':   unescaped = '\r';   ++first; break;
                    case '"':   unescaped = '"';    ++first; break;
                    case '\'':  unescaped = '\'';   ++first; break;
                    case '\\':  unescaped = '\\';   ++first; break;
                    case 'x': case 'X':
                        {
                            char_t hex = 0;
                            ++first;
                            char_t lim = 
                                std::numeric_limits<char_t>::max() >> 4;
                            while (first != last)
                            {
                                char_t c = *first;
                                if (hex > lim && std::isxdigit(c))
                                {
                                    // overflow detected
                                    return MatchTraitsT::no_match; 
                                }
                                if (std::isdigit(c))
                                {
                                    hex <<= 4;
                                    hex |= c - '0';
                                    ++first;
                                }
                                else if (std::isxdigit(c))
                                {
                                    hex <<= 4;
                                    c = toupper(c);
                                    hex |= c - 'A' + 0xA;
                                    ++first;
                                }
                                else
                                {
                                    break; // reached the end of the number
                                }
                            }
                            unescaped = hex;
                        }
                        break;

                    case '0': case '1': case '2': case '3':
                    case '4': case '5': case '6': case '7':
                        {
                            char_t hex = 0;
                            char_t lim = 
                                std::numeric_limits<char_t>::max() >> 3;
                            while (first != last)
                            {
                                char_t c = *first;
                                if (hex > lim && (c >= '0' && c <= '7'))
                                {
                                    // overflow detected
                                    return MatchTraitsT::no_match; 
                                }
                                if (c >= '0' && c <= '7')
                                {
                                    hex <<= 3;
                                    hex |= c - '0';
                                    ++first;
                                }
                                else
                                {
                                    break; // reached end of digits
                                }
                            }
                            unescaped = hex;
                        }

                        break;

                    default:
                        if (Flags & escape_flags::c_escapes)
                        {
                            // undefined escape invalid for c
                            return MatchTraitsT::no_match; 
                        }
                        else
                        {
                            unescaped = *first;
                            ++first;
                        }
                        break;
                }
            }
            else
            {
                unescaped = *first;
                ++first;
            }

            actor(unescaped);
            return hit;
        }
    }
    return MatchTraitsT::no_match; // overflow detected
}

//////////////////////////////////
template <unsigned long Flags>
template <typename IteratorT>
inline match
escape_char_parser<Flags>::parse(IteratorT& first, IteratorT const& last) const
{
    static range<> octal('0','7');
    static rule<IteratorT> escape
        =   +octal
        |   nocase['x'] >> +xdigit
        |   (anychar - nocase['x'] - octal);

    return ((anychar - '\\') | ('\\' >> escape)).parse(first, last);
}

//////////////////////////////////
template <unsigned long Flags>
template <typename IteratorT, typename MatchTraitsT>
inline typename MatchTraitsT::match_t
escape_char_parser<Flags>::ast_parse(IteratorT& first, IteratorT const& last, 
        MatchTraitsT const& mt) const
{
    static range<> octal('0','7');
    static rule<IteratorT> escape
        =   +octal
        |   nocase['x'] >> +xdigit
        |   (anychar - nocase['x'] - octal);

    return ((anychar - '\\') | ('\\' >> escape)).ast_parse(first, last, mt);
}

/*=============================================================================

    confix parsers implementation [Hartmut Kaiser]

=============================================================================*/

///////////////////////////////////////////////////////////////////////////////
//
//  confix_parser class implementation
//
///////////////////////////////////////////////////////////////////////////////
namespace impl {

    template <typename NestedT, typename CategoryT>
    struct confix_parser_type;

///////////////////////////////////////////////////////////////////////////////
// There is nothing special for plain_parser_category types as ExprT.
    template <>
    struct confix_parser_type<nested, plain_parser_category> {

        template <
            typename IteratorT, typename ThisT,
            typename OpenT, typename ExprT, typename CloseT
        >
        static match
        parse(
            IteratorT& first, IteratorT const& last, ThisT const& this_,
            OpenT const& open, ExprT const& expr, CloseT const& close)
        {
            return (open >> *(this_ | expr - close) >> close).parse(
                    first, last);
        }

        template <
            typename IteratorT, typename MatchTraitsT, typename ThisT,
            typename OpenT, typename ExprT, typename CloseT
        >
        static typename MatchTraitsT::match_t
        ast_parse(
            IteratorT& first, IteratorT const& last, MatchTraitsT const& mt, 
            ThisT const& this_, OpenT const& open, ExprT const& expr, 
            CloseT const& close)
        {
            return (open >> *(this_ | expr - close) >> close).ast_parse(
                    first, last, mt);
        }
    };

    template <>
    struct confix_parser_type<non_nested, plain_parser_category> {

        template <
            typename IteratorT, typename ThisT,
            typename OpenT, typename ExprT, typename CloseT
        >
        static match
        parse(
            IteratorT& first, IteratorT const& last, ThisT const& /*this_*/,
            OpenT const& open, ExprT const& expr, CloseT const& close)
        {
            return (open >> *(expr - close) >> close).parse(first, last);
        }

        template <
            typename IteratorT, typename MatchTraitsT, typename ThisT,
            typename OpenT, typename ExprT, typename CloseT
        >
        static typename MatchTraitsT::match_t
        ast_parse(
            IteratorT& first, IteratorT const& last, MatchTraitsT const& mt, 
            ThisT const& /*this_*/, OpenT const& open, ExprT const& expr, 
            CloseT const& close)
        {
            return (open >> *(expr - close) >> close).ast_parse(
                    first, last, mt);
        }
    };

///////////////////////////////////////////////////////////////////////////////
// action_parser_category types as ExprT require some special handling
    template <>
    struct confix_parser_type<nested, action_parser_category> {

        template <
            typename IteratorT, typename ThisT,
            typename OpenT, typename ExprT, typename CloseT
        >
        static match
        parse(
            IteratorT& first, IteratorT const& last, ThisT const& this_,
            OpenT const& open, ExprT const& expr, CloseT const& close)
        {
            return (
                    open
                >>  (*(this_ | expr.subject() - close))[expr.predicate()]
                >>  close
            ).parse(first, last);
        }

        template <
            typename IteratorT, typename MatchTraitsT, typename ThisT,
            typename OpenT, typename ExprT, typename CloseT
        >
        static typename MatchTraitsT::match_t
        ast_parse(
            IteratorT& first, IteratorT const& last, MatchTraitsT const& mt, 
            ThisT const& this_, OpenT const& open, ExprT const& expr, 
            CloseT const& close)
        {
            return (
                    open
                >>  (*(this_ | expr.subject() - close))[expr.predicate()]
                >>  close
            ).ast_parse(first, last, mt);
        }
    };

    template <>
    struct confix_parser_type<non_nested, action_parser_category> {

        template <
            typename IteratorT, typename ThisT,
            typename OpenT, typename ExprT, typename CloseT
        >
        static match
        parse(
            IteratorT& first, IteratorT const& last, ThisT const& /*this_*/,
            OpenT const& open, ExprT const& expr, CloseT const& close)
        {
            return (
                    open
                >>  (*(expr.subject() - close))[expr.predicate()]
                >>  close
            ).parse(first, last);
        }

        template <
            typename IteratorT, typename MatchTraitsT, typename ThisT,
            typename OpenT, typename ExprT, typename CloseT
        >
        static typename MatchTraitsT::match_t
        ast_parse(
            IteratorT& first, IteratorT const& last, MatchTraitsT const& mt, 
            ThisT const& /*this_*/, OpenT const& open, ExprT const& expr, 
            CloseT const& close)
        {
            return (
                    open
                >>  (*(expr.subject() - close))[expr.predicate()]
                >>  close
            ).ast_parse(first, last, mt);
        }
    };
}   // end of namespace impl

///////////////////////////////////////////////////////////////////////////////
//
// confix_parser class implementation
//
///////////////////////////////////////////////////////////////////////////////
template <
    typename OpenT, typename ExprT, typename CloseT,
    typename NestedT, typename CategoryT
>
inline confix_parser<OpenT, ExprT, CloseT, NestedT, CategoryT>::confix_parser(
    OpenT const &open_, ExprT const &expr_, CloseT const &close_) :
    open(open_), expr(expr_), close(close_)
{
}

template <
    typename OpenT, typename ExprT, typename CloseT,
    typename NestedT, typename CategoryT
>
template <typename IteratorT>
inline match
confix_parser<OpenT, ExprT, CloseT, NestedT, CategoryT>
    ::parse(IteratorT& first, IteratorT const& last) const
{
    return impl::confix_parser_type<NestedT, CategoryT>::
        parse(first, last, *this, open, expr, close);
}

template <
    typename OpenT, typename ExprT, typename CloseT,
    typename NestedT, typename CategoryT
>
template <typename IteratorT, typename MatchTraitsT>
inline typename MatchTraitsT::match_t
confix_parser<OpenT, ExprT, CloseT, NestedT, CategoryT>
    ::ast_parse(IteratorT& first, IteratorT const& last, 
            MatchTraitsT const& mt) const
{
    return impl::confix_parser_type<NestedT, CategoryT>::
        ast_parse(first, last, mt, *this, open, expr, close);
}

///////////////////////////////////////////////////////////////////////////////
//
//  Generic generator function for creation of concrete confix parsers
//
///////////////////////////////////////////////////////////////////////////////
template<typename NestedT>
template<typename StartT, typename ExprT, typename EndT>
inline confix_parser<
    typename impl::parser_type<StartT>::parser_t,
    typename impl::parser_type<ExprT>::parser_t,
    typename impl::parser_type<EndT>::parser_t,
    NestedT,
    typename impl::parser_type<ExprT>::parser_t::parser_category
>
confix_parser_gen<NestedT>::operator()(
    StartT const &start_, ExprT const &expr_, EndT const &end_) const
{
    typedef typename impl::parser_type<StartT>::parser_t start_t;
    typedef typename impl::parser_type<ExprT>::parser_t expr_t;
    typedef typename impl::parser_type<EndT>::parser_t end_t;
    typedef
        typename impl::parser_type<ExprT>::parser_t::parser_category
        parser_category;

    typedef
        confix_parser<start_t, expr_t, end_t, NestedT, parser_category>
        return_t;

    return return_t(
                impl::parser_type<StartT>::get(start_),
                impl::parser_type<ExprT>::get(expr_),
                impl::parser_type<EndT>::get(end_)
           );
}

///////////////////////////////////////////////////////////////////////////////
//
//  Generic generator function for creation of concrete confix parsers
//  which have a action directly attached to the ExprT part of the parser
//  (see comment in utilities.hpp).
//
///////////////////////////////////////////////////////////////////////////////
template<typename NestedT>
template<typename StartT, typename ExprT, typename EndT>
inline confix_parser<
    typename impl::parser_type<StartT>::parser_t,
    typename impl::parser_type<ExprT>::parser_t,
    typename impl::parser_type<EndT>::parser_t,
    NestedT,
    plain_parser_category    // do not re-attach the actor
>
confix_parser_gen<NestedT>::direct(
    StartT const &start_, ExprT const &expr_, EndT const &end_) const
{
    typedef typename impl::parser_type<StartT>::parser_t start_t;
    typedef typename impl::parser_type<ExprT>::parser_t expr_t;
    typedef typename impl::parser_type<EndT>::parser_t end_t;
    typedef plain_parser_category parser_category;

    typedef
        confix_parser<start_t, expr_t, end_t, NestedT, parser_category>
        return_t;

    return return_t(
                impl::parser_type<StartT>::get(start_),
                impl::parser_type<ExprT>::get(expr_),
                impl::parser_type<EndT>::get(end_)
           );
}

///////////////////////////////////////////////////////////////////////////////
//
//  Generic generator function for creation of concrete comment parsers
//
///////////////////////////////////////////////////////////////////////////////
template<typename NestedT>
template<typename StartT, typename EndT>
inline confix_parser<
    typename impl::parser_type<StartT>::parser_t,
    anychar_,
    typename impl::parser_type<EndT>::parser_t,
    NestedT
>
comment_parser_gen<NestedT>::operator()(
    StartT const &start_, EndT const &end_) const
{
    typedef typename impl::parser_type<StartT>::parser_t start_t;
    typedef typename impl::parser_type<EndT>::parser_t end_t;

    typedef confix_parser<start_t, anychar_, end_t, NestedT> return_t;

    return return_t(
        impl::parser_type<StartT>::get(start_),
        anychar,
        impl::parser_type<EndT>::get(end_)
    );
}

///////////////////////////////////////////////////////////////////////////////
// 
//  list_action class implementation
//
///////////////////////////////////////////////////////////////////////////////
template <typename ParserT, typename ActionT>
inline 
list_action<ParserT, ActionT>::list_action(
    ParserT const& subject, ActionT const& actor_) :
    unary<ParserT>(subject), actor(actor_)
{
}

template <typename ParserT, typename ActionT>
template <typename IteratorT>
inline match
list_action<ParserT, ActionT>::parse(IteratorT& first, 
        IteratorT const& last) const
{
    typedef impl::strip_scanner<IteratorT> strip_scanner;
    typedef list_action_iterator<ParserT, IteratorT, 
        typename ParserT::string_t> iterator_t;

    typename strip_scanner::iterator_type begin = strip_scanner::get(first);
    iterator_t list_it (&subject(), first, last);

    actor(list_it, iterator_t());
    if (list_it.matched()) {
        first = list_it.actposition();
        return match (std::distance(begin, strip_scanner::get(first)));
    }
    return match();
}

template <typename ParserT, typename ActionT>
template <typename IteratorT, typename MatchTraitsT>
inline typename MatchTraitsT::match_t
list_action<ParserT, ActionT>::ast_parse(IteratorT& first, IteratorT const& last, MatchTraitsT const& mt) const
{
    typedef impl::strip_scanner<IteratorT> strip_scanner;
    typedef list_action_iterator<ParserT, IteratorT, 
        typename ParserT::string_t> iterator_t;

    typename strip_scanner::iterator_type begin = strip_scanner::get(first);
    IteratorT savedfirst(first);
    iterator_t list_it (&subject(), first, last);

    actor(list_it, iterator_t());
    if (list_it.matched()) {
        first = list_it.actposition();
        return MatchTraitsT::create_match(
                std::distance(begin, strip_scanner::get(first)), 
                savedfirst, 
                first);
    }
    return MatchTraitsT::no_match;
}

///////////////////////////////////////////////////////////////////////////////
// 
//  list_action_iterator class implementation
//
///////////////////////////////////////////////////////////////////////////////
template <typename ParserT, typename IteratorT, typename BaseT>
inline 
list_action_iterator<ParserT, IteratorT, BaseT>::list_action_iterator() :
    parser(0), current(0), last(0), end_marker(true), saw_oneitem(false)
{
}

template <typename ParserT, typename IteratorT, typename BaseT>
inline 
list_action_iterator<ParserT, IteratorT, BaseT>::list_action_iterator(
    ParserT const *parser_, IteratorT &first_, IteratorT const &last_) :
    parser(parser_), current(first_), last(last_), end_marker(true), 
    saw_oneitem(false)
{
    end_marker = !read(true);     // get the first value
}

///////////////////////////////////////////////////////////////////////////////
// functions required for input iterators
template <typename ParserT, typename IteratorT, typename BaseT>
inline BaseT const &
list_action_iterator<ParserT, IteratorT, BaseT>::operator* () const 
{ 
    return value; 
}

template <typename ParserT, typename IteratorT, typename BaseT>
inline list_action_iterator<ParserT, IteratorT, BaseT> &
list_action_iterator<ParserT, IteratorT, BaseT>::operator++() 
{ 
    read(); 
    return *this; 
}

template <typename ParserT, typename IteratorT, typename BaseT>
inline list_action_iterator<ParserT, IteratorT, BaseT> 
list_action_iterator<ParserT, IteratorT, BaseT>::operator++ (int) 
{ 
list_action_iterator<ParserT, IteratorT, BaseT> tmp = *this; 

    read(); 
    return tmp; 
}

template <typename ParserT, typename IteratorT, typename BaseT>
inline bool 
list_action_iterator<ParserT, IteratorT, BaseT>::operator== (
    list_action_iterator<ParserT, IteratorT, BaseT> const &rhs) const
{ 
    return (current == rhs.current && 
                last == rhs.last && 
                end_marker == rhs.end_marker) ||
           (end_marker && rhs.end_marker);
}

template <typename ParserT, typename IteratorT, typename BaseT>
inline bool 
list_action_iterator<ParserT, IteratorT, BaseT>::operator!= (
    list_action_iterator<ParserT, IteratorT, BaseT> const &rhs) const
{ 
    return !(*this == rhs);
}

///////////////////////////////////////////////////////////////////////////////
// functions required for parser integration
template <typename ParserT, typename IteratorT, typename BaseT>
inline IteratorT const &
list_action_iterator<ParserT, IteratorT, BaseT>::actposition() const 
{ 
    return current; 
}

template <typename ParserT, typename IteratorT, typename BaseT>
inline bool
list_action_iterator<ParserT, IteratorT, BaseT>::matched() const 
{ 
    return saw_oneitem; 
}

template <typename ParserT, typename IteratorT, typename BaseT>
inline void 
list_action_iterator<ParserT, IteratorT, BaseT>::operator() (
    IteratorT const &begin_match, IteratorT const &end_match) const
{
    value = BaseT(begin_match, end_match-begin_match);
}

template <typename ParserT, typename IteratorT, typename BaseT>
inline void 
list_action_iterator<ParserT, IteratorT, BaseT>::operator() (
    typename std::iterator_traits<IteratorT>::value_type const &match_) const
{
    value += match_;
}

template <typename ParserT, typename IteratorT, typename BaseT>
inline ParserT const &
list_action_iterator<ParserT, IteratorT, BaseT>::subject() const 
{ 
    assert(0 != parser);
    return *parser; 
}

///////////////////////////////////////////////////////////////////////////////
// get the next item from the parsed list
namespace impl {

    template <typename CategoryT>
    struct list_action_parser_type;

///////////////////////////////////////////////////////////////////////////////
// There is nothing special for plain_parser_category types as ItemT.
    template <>
    struct list_action_parser_type<plain_parser_category>
    {
        template <typename ActionIterT, typename ParserT, typename IteratorT>
        static match read_item (
            ActionIterT const &list_action_it, ParserT const &subject, 
            IteratorT &current, IteratorT const &last)
        {
            return (
                *(subject.item_p() - (subject.delim_p() | subject.end_p()))
            )[list_action_it].parse(current, last);
        }
    };

///////////////////////////////////////////////////////////////////////////////
// action_parser_category types as ItemT require some special handling
    template <>
    struct list_action_parser_type<action_parser_category>
    {
        template <typename ActionIterT, typename ParserT, typename IteratorT>
        static match read_item (
            ActionIterT const &list_action_it, ParserT const &subject, 
            IteratorT &current, IteratorT const &last)
        {
            return (
                (*(subject.item_p().subject() 
                    - (subject.delim_p() | subject.end_p())
                ) )[subject.item_p().predicate()]
            )[list_action_it].parse(current, last);
        }
    };

} // end of namespace impl

template <typename ParserT, typename IteratorT, typename BaseT>
inline match 
list_action_iterator<ParserT, IteratorT, BaseT>::read (bool firstread)
{
    match hit;

    if (!firstread) {
    // match delimiter
        hit = subject().delim_p().parse (current, last);
        if (!hit) {
        // try to match the 'end' sequence
            end_marker = true;
            return subject().end_p().parse(current, last);
        }
    }

    // parse the next item
    value.empty();
    hit = impl::list_action_parser_type<typename ParserT::parser_category>
            ::read_item (*this, subject(), current, last);

    if (!saw_oneitem)
        saw_oneitem = hit;
    end_marker = !hit;

    return hit;
}

///////////////////////////////////////////////////////////////////////////////
// 
//  list_parser class
//
///////////////////////////////////////////////////////////////////////////////
namespace impl {

    template <typename CategoryT>
    struct list_parser_type;

///////////////////////////////////////////////////////////////////////////////
// There is nothing special for plain_parser_category types as ItemT.
    template <>
    struct list_parser_type<plain_parser_category> 
    {
    // match list with 'normal' syntax
        template <typename IteratorT, typename ItemT, typename DelimT>
        static match
        parse(IteratorT& first, IteratorT const& last, 
            ItemT const &item, DelimT const &delim)
        {
            return (
                    *(item - delim)
                >>  *(delim >> *(item - delim))
            ).parse(first, last);
        }

        template <typename IteratorT, typename MatchTraitsT, typename ItemT, 
            typename DelimT>
        static typename MatchTraitsT::match_t
        ast_parse(IteratorT& first, IteratorT const& last, 
                MatchTraitsT const& mt, ItemT const &item, DelimT const &delim)
        {
            return (
                    *(item - delim)
                >>  *(delim >> *(item - delim))
            ).ast_parse(first, last, mt);
        }

    // match list with 'extended' syntax
        template <
            typename IteratorT, 
            typename ItemT, typename DelimT, typename EndT
        >
        static match
        parse(IteratorT& first, IteratorT const& last, 
            ItemT const &item, DelimT const &delim, EndT const &end)
        {
            return (
                    *(item - (delim | end))
                >>  *(delim >> *(item - (delim | end)))
                >>  !end
            ).parse(first, last);
        }

        template <
            typename IteratorT, typename MatchTraitsT,
            typename ItemT, typename DelimT, typename EndT
        >
        static typename MatchTraitsT::match_t
        ast_parse(IteratorT& first, IteratorT const& last, 
                MatchTraitsT const& mt, ItemT const &item, DelimT const &delim,
                EndT const &end)
        {
            return (
                    *(item - (delim | end))
                >>  *(delim >> *(item - (delim | end)))
                >>  !end
            ).ast_parse(first, last, mt);
        }
    };

///////////////////////////////////////////////////////////////////////////////
// action_parser_category types as ItemT require some special handling
    template <>
    struct list_parser_type<action_parser_category> 
    {
    // match list with 'normal' syntax
        template <typename IteratorT, typename ItemT, typename DelimT>
        static match
        parse(IteratorT& first, IteratorT const& last, 
            ItemT const &item, DelimT const &delim)
        {
            return (
                    *(item.subject() - delim)[item.predicate()]
                >>  *(delim >> *(item.subject() - delim)[item.predicate()])
            ).parse(first, last);
        }

        template <typename IteratorT, typename MatchTraitsT, typename ItemT, 
                    typename DelimT>
        static typename MatchTraitsT::match_t
        ast_parse(IteratorT& first, IteratorT const& last, 
                MatchTraitsT const& mt, ItemT const &item, DelimT const &delim)
        {
            return (
                    *(item.subject() - delim)[item.predicate()]
                >>  *(delim >> *(item.subject() - delim)[item.predicate()])
            ).ast_parse(first, last, mt);
        }

    // match list with 'extended' syntax
        template <
            typename IteratorT, 
            typename ItemT, typename DelimT, typename EndT
        >
        static match
        parse(IteratorT& first, IteratorT const& last, 
            ItemT const &item, DelimT const &delim, EndT const &end)
        {
            return (
                    *(item.subject() - (delim | end))[item.predicate()]
                >>  *(delim >> *(item.subject() - (delim | end))[
                        item.predicate()])
                >>  !end
            ).parse(first, last);
        }

        template <
            typename IteratorT, typename MatchTraitsT, 
            typename ItemT, typename DelimT, typename EndT
        >
        static typename MatchTraitsT::match_t
        ast_parse(IteratorT& first, IteratorT const& last, 
                MatchTraitsT const& mt, ItemT const &item, DelimT const &delim,
                EndT const &end)
        {
            return (
                    *(item.subject() - (delim | end))[item.predicate()]
                >>  *(delim >> *(item.subject() - (delim | end))[
                        item.predicate()])
                >>  !end
            ).ast_parse(first, last, mt);
        }
    };

}   // end of namespace impl

template <
    typename ItemT, typename DelimT, 
    typename StringT, typename CategoryT
>
inline list_parser<ItemT, DelimT, StringT, CategoryT>
    ::list_parser(ItemT const &item_, DelimT const &delim_) :
    item(item_), delim(delim_)
{
}

template <
    typename ItemT, typename DelimT, 
    typename StringT, typename CategoryT
>
template <typename IteratorT>
inline match
list_parser<ItemT, DelimT, StringT, CategoryT>
    ::parse(IteratorT& first, IteratorT const& last) const
{
    return impl::list_parser_type<CategoryT>
            ::parse(first, last, item, delim);
}

template <
    typename ItemT, typename DelimT, 
    typename StringT, typename CategoryT
>
template <typename IteratorT, typename MatchTraitsT>
inline typename MatchTraitsT::match_t
list_parser<ItemT, DelimT, StringT, CategoryT>
    ::ast_parse(IteratorT& first, IteratorT const& last, 
            MatchTraitsT const& mt) const
{
    return impl::list_parser_type<CategoryT>
            ::ast_parse(first, last, mt, item, delim);
}

///////////////////////////////////////////////////////////////////////////////
// 
//  list_parser_ex class
//
///////////////////////////////////////////////////////////////////////////////
template <
    typename ItemT, typename DelimT, typename EndT, 
    typename StringT, typename CategoryT
>
inline list_parser_ex<ItemT, DelimT, EndT, StringT, CategoryT>
    ::list_parser_ex(ItemT const &item_, DelimT const &delim_, 
            EndT const &end_) :
    item(item_), delim(delim_), end(end_)
{
}

template <
    typename ItemT, typename DelimT, typename EndT, 
    typename StringT, typename CategoryT
>
template <typename IteratorT>
inline match
list_parser_ex<ItemT, DelimT, EndT, StringT, CategoryT>
    ::parse(IteratorT& first, IteratorT const& last) const
{
    return impl::list_parser_type<CategoryT>
            ::parse(first, last, item, delim, end);
}

template <
    typename ItemT, typename DelimT, typename EndT, 
    typename StringT, typename CategoryT
>
template <typename IteratorT, typename MatchTraitsT>
inline typename MatchTraitsT::match_t
list_parser_ex<ItemT, DelimT, EndT, StringT, CategoryT>
    ::ast_parse(IteratorT& first, IteratorT const& last, 
            MatchTraitsT const& mt) const
{
    return impl::list_parser_type<CategoryT>
            ::ast_parse(first, last, mt, item, delim, end);
}

///////////////////////////////////////////////////////////////////////////////
//
//  Generic generator function for creation of concrete list parsers for
//  'normal' list syntax: 
//
//      item >> *(delim >> item)
//
//  If item isn't given, everything between two delimiters is matched.
//
///////////////////////////////////////////////////////////////////////////////
template <typename StringT>
template <typename DelimT>
inline list_parser<
    anychar_,
    typename impl::parser_type<DelimT>::parser_t,
    StringT,
    plain_parser_category          // anychar is a plain_parser type
>
list_parser_gen<StringT>::operator()(DelimT const &delim_) const
{
    typedef typename impl::parser_type<DelimT>::parser_t delim_t;

    typedef 
        list_parser<anychar_, delim_t, StringT, plain_parser_category> 
        return_t;

    return return_t(
        anychar,
        impl::parser_type<DelimT>::get(delim_)
    );
}

template <typename StringT>
template <typename ItemT, typename DelimT>
inline list_parser<
    typename impl::parser_type<ItemT>::parser_t,
    typename impl::parser_type<DelimT>::parser_t,
    StringT
>
list_parser_gen<StringT>::operator()(ItemT const &item_, 
        DelimT const &delim_) const
{
    typedef typename impl::parser_type<ItemT>::parser_t item_t;
    typedef typename impl::parser_type<DelimT>::parser_t delim_t;

    typedef list_parser<item_t, delim_t, StringT> return_t;

    return return_t(
        impl::parser_type<ItemT>::get(item_),
        impl::parser_type<DelimT>::get(delim_)
    );
}

///////////////////////////////////////////////////////////////////////////////
//
//  Generic generator function for creation of concrete list parsers for
//  'extended' list syntax: 
//
//      item >> *(delim >> item) >> !end
//
///////////////////////////////////////////////////////////////////////////////
template <typename StringT>
template<typename ItemT, typename DelimT, typename EndT>
inline list_parser_ex<
    typename impl::parser_type<ItemT>::parser_t,
    typename impl::parser_type<DelimT>::parser_t,
    typename impl::parser_type<EndT>::parser_t,
    StringT
>
list_parser_gen<StringT>::operator()(
    ItemT const &item_, DelimT const &delim_, EndT const &end_) const
{
    typedef typename impl::parser_type<ItemT>::parser_t item_t;
    typedef typename impl::parser_type<DelimT>::parser_t delim_t;
    typedef typename impl::parser_type<EndT>::parser_t end_t;

    typedef list_parser_ex<item_t, delim_t, end_t, StringT> return_t;

    return return_t(
        impl::parser_type<ItemT>::get(item_),
        impl::parser_type<DelimT>::get(delim_),
        impl::parser_type<EndT>::get(end_)
    );
}

///////////////////////////////////////////////////////////////////////////////
//
//  Generic generator function for creation of concrete list parsers for
//  'normal' list syntax: 
//
//      item >> *(delim >> item)
//
//  If item isn't given, everything between two delimiters is matched.
//
//  The following function should be used, if the 'item' parser has an attached 
//  semantic action and this action should _not_ be re-attached during parser 
//  construction (see comment in utilities.hpp).
//
///////////////////////////////////////////////////////////////////////////////
template <typename StringT>
template <typename ItemT, typename DelimT>
inline list_parser<
    typename impl::parser_type<ItemT>::parser_t,
    typename impl::parser_type<DelimT>::parser_t,
    StringT,
    plain_parser_category        // inhibit action re-attachment
>
list_parser_gen<StringT>::direct(ItemT const &item_, DelimT const &delim_) const
{
    typedef typename impl::parser_type<ItemT>::parser_t item_t;
    typedef typename impl::parser_type<DelimT>::parser_t delim_t;

    typedef 
        list_parser<item_t, delim_t, StringT, plain_parser_category> 
        return_t;

    return return_t(
        impl::parser_type<ItemT>::get(item_),
        impl::parser_type<DelimT>::get(delim_)
    );
}

///////////////////////////////////////////////////////////////////////////////
//
//  Generic generator function for creation of concrete list parsers for
//  'extended' list syntax: 
//
//      item >> *(delim >> item) >> !end
//
///////////////////////////////////////////////////////////////////////////////
template <typename StringT>
template<typename ItemT, typename DelimT, typename EndT>
inline list_parser_ex<
    typename impl::parser_type<ItemT>::parser_t,
    typename impl::parser_type<DelimT>::parser_t,
    typename impl::parser_type<EndT>::parser_t,
    StringT,
    plain_parser_category        // inhibit action re-attachment
>
list_parser_gen<StringT>::direct(
    ItemT const &item_, DelimT const &delim_, EndT const &end_) const
{
    typedef typename impl::parser_type<ItemT>::parser_t item_t;
    typedef typename impl::parser_type<DelimT>::parser_t delim_t;
    typedef typename impl::parser_type<EndT>::parser_t end_t;

    typedef 
        list_parser_ex<item_t, delim_t, end_t, StringT, plain_parser_category> 
        return_t;

    return return_t(
        impl::parser_type<ItemT>::get(item_),
        impl::parser_type<DelimT>::get(delim_),
        impl::parser_type<EndT>::get(end_)
    );
}

} // namespace spirit

#endif

