/*=============================================================================
    Numeric parsers

    Spirit V1.2
    Copyright (c) 2001, Joel de Guzman

    This software is provided 'as-is', without any express or implied
    warranty. In no event will the copyright holder be held liable for
    any damages arising from the use of this software.

    Permission is granted to anyone to use this software for any purpose,
    including commercial applications, and to alter it and redistribute
    it freely, subject to the following restrictions:

    1.  The origin of this software must not be misrepresented; you must
        not claim that you wrote the original software. If you use this
        software in a product, an acknowledgment in the product documentation
        would be appreciated but is not required.

    2.  Altered source versions must be plainly marked as such, and must
        not be misrepresented as being the original software.

    3.  This notice may not be removed or altered from any source
        distribution.

    Acknowledgements:

        Special thanks to Dan Nuffer, John (EBo) David, Chris Uzdavinis,
        and Doug Gregor. These people are most instrumental in steering
        Spirit in the right direction.

        Special thanks also to people who have contributed to the code base
        and sample code, ported Spirit to various platforms and compilers,
        gave suggestions, reported and provided bug fixes. Alexander
        Hirner, Andy Elvey, Bogdan Kushnir, Brett Calcott, Bruce Florman,
        Changzhe Han, Colin McPhail, Hakki Dogusan, Jan Bares, Joseph
        Smith, Martijn W. van der Lee, Raghavendra Satish, Remi Delcos, Tom
        Spilman, Vladimir Prus, W. Scott Dillman, David A. Greene, Bob
        Bailey, Hartmut Kaiser.

        Finally special thanks also to people who gave feedback and
        valuable comments, particularly members of Spirit's Source Forge
        mailing list and boost.org.

    URL: http://spirit.sourceforge.net/

=============================================================================*/
#ifndef SPIRIT_NUMERICS_HPP
#define SPIRIT_NUMERICS_HPP

///////////////////////////////////////////////////////////////////////////////
namespace spirit {

///////////////////////////////////////////////////////////////////////////////
//
//  numeric_action class
//
//      Links a numeric parser with a user defined semantic action.
//      The semantic action may be a function or a functor. A function
//      should be compatible with the interface:
//
//          void f(NumT num);
//
//      A functor should have a member operator() with a compatible
//      signature as above. The matching number is passed into the
//      function/functor. This is the default class that numeric
//      parsers use when dealing with the construct:
//
//          p[f]
//
//      where p is a parser and f is a function or functor. The header
//      file <actions.hpp> has some useful predefined generic functors
//      that may be used to directly link variables with numeric parsers
//      (see reference_wrapper and context_wrapper classes.
//
///////////////////////////////////////////////////////////////////////////////
template <typename ParserT, typename ActionT>
class numeric_action
:   public unary<ParserT>,
    public parser<numeric_action<ParserT, ActionT> > {

public:

    typedef action_parser_category parser_category;

    numeric_action(ActionT const& actor);

    template <typename IteratorT>
    match
    parse(IteratorT& first, IteratorT const& last) const;

    template <typename IteratorT, typename MatchTraitsT>
    typename MatchTraitsT::match_t
    ast_parse(IteratorT& first, IteratorT const& last, 
            MatchTraitsT const&) const;

    ActionT const &predicate() const { return actor; }

private:

    typename embed_trait<ActionT>::type actor;
};

///////////////////////////////////////////////////////////////////////////////
// Simple traits class for customizing locale specifics. This allows for
// customization of the parsing algorithm in dependency of the used locale,
// format of the number etc.
// Use this traits template, if you need single character delimiters.
// The default separators used here are for parsing C-style numbers.
template<const int ds = '.', const int ts = '\0', const int exp = 'e'>
struct numeric_parser_traits
{
    enum real_parser_chars {
        decimal_sep_char = ds,   // character used as decimal point
        thousands_sep_char = ts, // character used as thousands separator
        exponent_char = exp      // character used as delimiter before exponent
    };

    typedef const int DecimalSepT;
    typedef const int ThousandsSepT;
    typedef const int ExponentSepT;
};

///////////////////////////////////////////////////////////////////////////////
// More complex traits class for customizing locale specifics. This allows for
// customization of the parsing algorithm in dependency of the used locale,
// format of the number etc.
// Use this traits template, if you need more complex delimiters such as static
// parsers et.al.
template<typename DecimalParserT, typename ThousandsParserT, 
        typename ExponentParserT>
struct numeric_parser_traits_ex
{
    typedef DecimalParserT DecimalSepT;
    typedef ThousandsParserT ThousandsSepT;
    typedef ExponentParserT ExponentSepT;

    static DecimalSepT decimal_sep_char;     // delimiter used as decimal point
    static ThousandsSepT thousands_sep_char; // delimiter used as thousands 
                                             // separator
    static ExponentSepT exponent_char;       // delimiter used before exponent
};

template<typename DecimalParserT, typename ThousandsParserT, 
    typename ExponentParserT>
numeric_parser_traits_ex<DecimalParserT, ThousandsParserT, ExponentParserT>
    ::DecimalSepT
numeric_parser_traits_ex<DecimalParserT, ThousandsParserT, ExponentParserT>
    ::decimal_sep_char;

template<typename DecimalParserT, typename ThousandsParserT, 
    typename ExponentParserT>
numeric_parser_traits_ex<DecimalParserT, ThousandsParserT, ExponentParserT>
    ::ThousandsSepT
numeric_parser_traits_ex<DecimalParserT, ThousandsParserT, ExponentParserT>
    ::thousands_sep_char;

template<typename DecimalParserT, typename ThousandsParserT, 
    typename ExponentParserT>
numeric_parser_traits_ex<DecimalParserT, ThousandsParserT, ExponentParserT>
    ::ExponentSepT
numeric_parser_traits_ex<DecimalParserT, ThousandsParserT, ExponentParserT>
    ::exponent_char;

///////////////////////////////////////////////////////////////////////////////
// forward declarations
template <typename T = unsigned, const int Radix = 10> struct uint_parser;
template <typename T = int, const int Radix = 10> struct int_parser;
template <typename T = double, typename TraitsT = numeric_parser_traits<> > 
    struct ureal_parser;
template <typename T = double, typename TraitsT = numeric_parser_traits<> > 
    struct real_parser;

///////////////////////////////////////////////////////////////////////////////
template <typename T, const int Radix = 10>
struct int_numeric_action_traits {

    typedef typename T::arg_type arg_type;
    typedef numeric_action<uint_parser<arg_type, Radix>, T>     uint_action;
    typedef numeric_action<int_parser<arg_type, Radix>, T>      int_action;
};

template <typename T, typename TraitsT = numeric_parser_traits<> >
struct real_numeric_action_traits {

    typedef typename T::arg_type arg_type;
    typedef numeric_action<ureal_parser<arg_type, TraitsT>, T>  ureal_action;
    typedef numeric_action<real_parser<arg_type, TraitsT>, T>   real_action;
};

///////////////////////////////////////////////////////////////////////////////
template <typename T, const int Radix>
struct int_numeric_action_traits<void (*)(T), Radix> {

    typedef void (*func_type)(T);
    typedef numeric_action<uint_parser<T, Radix>, func_type>    uint_action;
    typedef numeric_action<int_parser<T, Radix>, func_type>     int_action;
};

template <typename T, typename TraitsT>
struct real_numeric_action_traits<void (*)(T), TraitsT> {

    typedef void (*func_type)(T);
    typedef numeric_action<ureal_parser<T, TraitsT>, func_type>  ureal_action;
    typedef numeric_action<real_parser<T, TraitsT>, func_type>   real_action;
};

///////////////////////////////////////////////////////////////////////////////
template <typename T, const int Radix>
struct int_numeric_action_traits<reference_wrapper<T>, Radix> {

    typedef uint_parser<T&, Radix>      uint_action;
    typedef int_parser<T&, Radix>       int_action;
};

template <typename T, typename TraitsT>
struct real_numeric_action_traits<reference_wrapper<T>, TraitsT> {

    typedef ureal_parser<T&, TraitsT>   ureal_action;
    typedef real_parser<T&, TraitsT>    real_action;
};

///////////////////////////////////////////////////////////////////////////////
//
//  uint_parser class
//
//      Parses unsigned integers. A uint_parser is never directly
//      instantiated. An instance of the uint_parser_gen, uint_p,
//      generates a uint_parser through its [] operator:
//
//          uint_p[f]
//
//      where f is a semantic action which may be a function or a
//      functor. A function should be compatible with the interface:
//
//          void f(NumT num);
//
//      A functor should have a member operator() with a compatible
//      signature as above. The matching number is passed into the
//      function/functor (see numeric_action class above).
//
///////////////////////////////////////////////////////////////////////////////
template <typename T, const int Radix>
struct uint_parser : public parser<uint_parser<T, Radix> > {

    uint_parser(T _n);

    template <typename IteratorT>
    match
    parse(IteratorT& first, IteratorT const& last) const;

    template <typename IteratorT, typename MatchTraitsT>
    typename MatchTraitsT::match_t
    ast_parse(IteratorT& first, IteratorT const& last, 
            MatchTraitsT const&) const;

    mutable T n;
};

///////////////////////////////////////////////////////////////////////////////
template<const int Radix>
struct uint_parser_gen : public uint_parser<unsigned, Radix> {

    uint_parser_gen()
    :   uint_parser<unsigned, Radix>(0) {}

    template <typename T>
    typename int_numeric_action_traits<T, Radix>::uint_action
    operator [] (T const& actor) const
    {
        //  Borland 5.5 reports an internal compiler
        //  error if this is not defined here.
        typedef typename int_numeric_action_traits<T, Radix>::uint_action 
            action_t;
        return action_t(actor);
    }
};

///////////////////////////////////////////////////////////////////////////////
// predefined unsigned integer parsers
const uint_parser_gen<2>    ubin_p = uint_parser_gen<2>();
const uint_parser_gen<8>    uoct_p = uint_parser_gen<8>();
const uint_parser_gen<10>   uint_p = uint_parser_gen<10>();
const uint_parser_gen<16>   uhex_p = uint_parser_gen<16>();

///////////////////////////////////////////////////////////////////////////////
//
//  int_parser class
//
//      Parses signed integers. An int_parser is never directly
//      instantiated. An instance of the int_parser_gen, int_p,
//      generates an int_parser through its [] operator:
//
//          int_p[f]
//
//      where f is a semantic action which may be a function or a
//      functor. A function should be compatible with the interface:
//
//          void f(NumT num);
//
//      A functor should have a member operator() with a compatible
//      signature as above. The matching number is passed into the
//      function/functor (see numeric_action class above).
//
///////////////////////////////////////////////////////////////////////////////
template <typename T, const int Radix>
struct int_parser : public parser<int_parser<T, Radix> > {

    int_parser(T _n);

    template <typename IteratorT>
    match
    parse(IteratorT& first, IteratorT const& last) const;

    template <typename IteratorT, typename MatchTraitsT>
    typename MatchTraitsT::match_t
    ast_parse(IteratorT& first, IteratorT const& last, 
            MatchTraitsT const&) const;

    mutable T n;
};


///////////////////////////////////////////////////////////////////////////////
template <const int Radix>
struct int_parser_gen : public int_parser<int, Radix> {

    int_parser_gen()
    :   int_parser<int, Radix>(0) {}

    template <typename T>
    typename int_numeric_action_traits<T, Radix>::int_action
    operator [] (T const& actor) const
    {
        //  Borland 5.5 reports an internal compiler
        //  error if this is not defined here.
        typedef typename int_numeric_action_traits<T, Radix>::int_action 
            action_t;
        return action_t(actor);
    }
};

///////////////////////////////////////////////////////////////////////////////
// predefined signed integer parsers
const int_parser_gen<2>     bin_p = int_parser_gen<2>();
const int_parser_gen<8>     oct_p = int_parser_gen<8>();
const int_parser_gen<10>    int_p = int_parser_gen<10>();
const int_parser_gen<16>    hex_p = int_parser_gen<16>();

///////////////////////////////////////////////////////////////////////////////
//
//  ureal_parser class
//
//      Parses unsigned reals. A ureal_parser is never directly
//      instantiated. An instance of the ureal_parser_gen, ureal_p,
//      generates a ureal_parser through its [] operator:
//
//          ureal_p[f]
//
//      where f is a semantic action which may be a function or a
//      functor. A function should be compatible with the interface:
//
//          void f(NumT num);
//
//      A functor should have a member operator() with a compatible
//      signature as above. The matching number is passed into the
//      function/functor (see numeric_action class above).
//
///////////////////////////////////////////////////////////////////////////////
template <typename FloatT, typename TraitsT>
struct ureal_parser :
    public parser<ureal_parser<FloatT, TraitsT> >
{
    ureal_parser(FloatT n);

    template <typename IteratorT>
    match
    parse(IteratorT& first, IteratorT const& last) const;

    template <typename IteratorT, typename MatchTraitsT>
    typename MatchTraitsT::match_t
    ast_parse(IteratorT& first, IteratorT const& last, 
            MatchTraitsT const&) const;

    mutable FloatT n;
};

///////////////////////////////////////////////////////////////////////////////
//
template <typename TraitsT = numeric_parser_traits<> >
struct ureal_parser_gen :
    public ureal_parser<double, TraitsT>
{
    ureal_parser_gen()
    :   ureal_parser<double, TraitsT>(0) {}

    template <typename T>
    typename real_numeric_action_traits<T, TraitsT>::ureal_action
    operator [] (T const& actor) const
    {
        //  Borland 5.5 reports an internal compiler
        //  error if this is not defined here.
        typedef typename real_numeric_action_traits<T, TraitsT>::ureal_action 
            action_t;
        return action_t(actor);
    }

// Helper function for specifying a numeric_parser_traits type when using
// the predefined parser ureal_p. Now we could write ureal_p(mytraits())[...],
// where 'mytraits' is a class, which defines the same names as the
// 'numeric_parser_traits' template.
    template <typename NewTraitsT>
    ureal_parser_gen<NewTraitsT>
    operator ()(NewTraitsT) const
    {
        return ureal_parser_gen<NewTraitsT>();
    }
};

///////////////////////////////////////////////////////////////////////////////
// Predefined unsigned floating point parser
//
// The underlying base type (float, double etc.) is selected automatically at
// compile time through the type accociated with the attached semantic action.
// If no sematic action is used, double will be selected.
const ureal_parser_gen<> ureal_p = ureal_parser_gen<>();

///////////////////////////////////////////////////////////////////////////////
//
//  real_parser class
//
//      Parses signed reals. A real_parser is never directly
//      instantiated. An instance of the real_parser_gen, real_p,
//      generates a real_parser through its [] operator:
//
//          real_p[f]
//
//      where f is a semantic action which may be a function or a
//      functor. A function should be compatible with the interface:
//
//          void f(NumT num);
//
//      A functor should have a member operator() with a compatible
//      signature as above. The matching number is passed into the
//      function/functor (see numeric_action class above).
//
///////////////////////////////////////////////////////////////////////////////
template <typename FloatT, typename TraitsT>
struct real_parser :
    public parser<real_parser<FloatT, TraitsT> >
{
    real_parser(FloatT n);

    template <typename IteratorT>
    match
    parse(IteratorT& first, IteratorT const& last) const;

    template <typename IteratorT, typename MatchTraitsT>
    typename MatchTraitsT::match_t
    ast_parse(IteratorT& first, IteratorT const& last, 
            MatchTraitsT const&) const;

    mutable FloatT n;
};

///////////////////////////////////////////////////////////////////////////////
template <typename TraitsT = numeric_parser_traits<> >
struct real_parser_gen :
    public real_parser<double, TraitsT>
{
    real_parser_gen()
    :   real_parser<double, TraitsT>(0) {}

    template <typename T>
    typename real_numeric_action_traits<T, TraitsT>::real_action
    operator [] (T const& actor) const
    {
        //  Borland 5.5 reports an internal compiler
        //  error if this is not defined here.
        typedef typename real_numeric_action_traits<T, TraitsT>::real_action 
            action_t;
        return action_t(actor);
    }

// Helper function for specifying a numeric_parser_traits type when using
// the predefined parser real_p. Now we could write real_p(mytraits())[...],
// where 'mytraits' is a class, which defines the same names as the
// 'numeric_parser_traits' template.
    template <typename NewTraitsT>
    real_parser_gen<NewTraitsT>
    operator ()(NewTraitsT) const
    {
        return real_parser_gen<NewTraitsT>();
    }
};

///////////////////////////////////////////////////////////////////////////////
// Predefined signed floating point parser
//
// The underlying base type (float, double etc.) is selected automatically at
// compile time through the type associated with the attached semantic action.
// If no sematic action is used, double will be selected.
const real_parser_gen<> real_p = real_parser_gen<>();

///////////////////////////////////////////////////////////////////////////////
}   //  namespace Spirit

#endif
