/*=============================================================================
    Utilities

    Spirit V1.2
    Copyright (c) 2001, Daniel Nuffer
    Copyright (c) 2001, Hartmut Kaiser

    This software is provided 'as-is', without any express or implied
    warranty. In no event will the copyright holder be held liable for
    any damages arising from the use of this software.

    Permission is granted to anyone to use this software for any purpose,
    including commercial applications, and to alter it and redistribute
    it freely, subject to the following restrictions:

    1.  The origin of this software must not be misrepresented; you must
        not claim that you wrote the original software. If you use this
        software in a product, an acknowledgment in the product documentation
        would be appreciated but is not required.

    2.  Altered source versions must be plainly marked as such, and must
        not be misrepresented as being the original software.

    3.  This notice may not be removed or altered from any source
        distribution.

    Acknowledgements:

        Special thanks to Dan Nuffer, John (EBo) David, Chris Uzdavinis,
        and Doug Gregor. These people are most instrumental in steering
        Spirit in the right direction.

        Special thanks also to people who have contributed to the code base
        and sample code, ported Spirit to various platforms and compilers,
        gave suggestions, reported and provided bug fixes. Alexander
        Hirner, Andy Elvey, Bogdan Kushnir, Brett Calcott, Bruce Florman,
        Changzhe Han, Colin McPhail, Hakki Dogusan, Jan Bares, Joseph
        Smith, Martijn W. van der Lee, Raghavendra Satish, Remi Delcos, Tom
        Spilman, Vladimir Prus, W. Scott Dillman, David A. Greene, Bob
        Bailey, Hartmut Kaiser.

        Finally special thanks also to people who gave feedback and
        valuable comments, particularly members of Spirit's Source Forge
        mailing list and boost.org.

    URL: http://spirit.sourceforge.net/

=============================================================================*/
#ifndef SPIRIT_UTILITIES_HPP
#define SPIRIT_UTILITIES_HPP

#include <string>
#include <iterator>

///////////////////////////////////////////////////////////////////////////////
namespace spirit {

///////////////////////////////////////////////////////////////////////////////
//
//  escape_char_action class
//
//      Links an escape char parser with a user defined semantic action.
//      The semantic action may be a function or a functor. A function
//      should be compatible with the interface:
//
//          void f(CharT ch);
//
//      A functor should have a member operator() with a compatible signature
//      as above. The matching character is passed into the function/functor.
//      This is the default class that character parsers use when dealing with
//      the construct:
//
//          p[f]
//
//      where p is a parser and f is a function or functor.
//
///////////////////////////////////////////////////////////////////////////////
template <typename ParserT, typename ActionT, unsigned long Flags>
class escape_char_action
:   public unary<ParserT>,
    public parser<escape_char_action<ParserT, ActionT, Flags> > {

public:

    typedef action_parser_category parser_category;

    escape_char_action(
        ParserT const& subject,
        ActionT const& actor_);

    template <typename IteratorT>
    match
    parse(IteratorT& first, IteratorT const& last) const;

    template <typename IteratorT, typename MatchTraitsT>
    typename MatchTraitsT::match_t
    ast_parse(IteratorT& first, IteratorT const& last, 
            MatchTraitsT const&) const;

    ActionT const &predicate() const { return actor; }

private:

    typename embed_trait<ActionT>::type actor;
};

///////////////////////////////////////////////////////////////////////////////
//
//  escape_char_parser class
//
///////////////////////////////////////////////////////////////////////////////
struct escape_flags {

    static const unsigned long c_escapes = 1;
    static const unsigned long lex_escapes = c_escapes << 1;
};

//////////////////////////////////
template <unsigned long Flags>
class escape_char_parser : public parser<escape_char_parser<Flags> > {

public:

    template <typename ActionT>
    escape_char_action<escape_char_parser<Flags>, ActionT, Flags>
    operator[](ActionT const& actor) const
    {
        //  Borland internal compiler error if this
        //  is defined elsewhere (JDG)
        return escape_char_action<escape_char_parser, ActionT, Flags>(
            *this, actor);
    }

    template <typename IteratorT>
    match
    parse(IteratorT& first, IteratorT const& last) const;

    template <typename IteratorT, typename MatchTraitsT>
    typename MatchTraitsT::match_t
    ast_parse(IteratorT& first, IteratorT const& last, 
            MatchTraitsT const&) const;
};

///////////////////////////////////////////////////////////////////////////////
//
//  predefined escape_char_parser classes
//
///////////////////////////////////////////////////////////////////////////////
const escape_char_parser<escape_flags::lex_escapes> lex_escape_ch_p =
    escape_char_parser<escape_flags::lex_escapes>();

const escape_char_parser<escape_flags::c_escapes> c_escape_ch_p =
    escape_char_parser<escape_flags::c_escapes>();

///////////////////////////////////////////////////////////////////////////////
//
//  Types to distinguish nested and non-nested confix parsers
//
///////////////////////////////////////////////////////////////////////////////
struct nested {};
struct non_nested {};

///////////////////////////////////////////////////////////////////////////////
//
//  Helper templates to derive the parser type.
//
///////////////////////////////////////////////////////////////////////////////
namespace impl {

    template<typename T>
    struct parser_type
    {
        typedef T parser_t;
        static parser_t const& get(parser_t const& p) { return p; }
    };

    template<>
    struct parser_type<char>
    {
        typedef chlit<char> parser_t;
        static parser_t get(char ch) { return parser_t(ch); }
    };

    template<>
    struct parser_type<wchar_t>
    {
        typedef chlit<wchar_t> parser_t;
        static parser_t get(wchar_t ch) { return parser_t(ch); }
    };

    template<>
    struct parser_type<char const *>
    {
        typedef strlit<cstring<char> > parser_t;
        static parser_t get(char const* str) { return parser_t(str); }
    };

    template<>
    struct parser_type<wchar_t const *>
    {
        typedef strlit<cstring<wchar_t> > parser_t;
        static parser_t get(wchar_t const* str) { return parser_t(str); }
    };

    template<int N>
    struct parser_type<char[N]>
    {
        typedef strlit<cstring<char> > parser_t;
        static parser_t get(char const str[N]) { return parser_t(str); }
    };

    template<int N>
    struct parser_type<wchar_t[N]>
    {
        typedef strlit<cstring<wchar_t> > parser_t;
        static parser_t get(wchar_t const str[N]) { return parser_t(str); }
    };

} // namespace impl

///////////////////////////////////////////////////////////////////////////////
//
//  confix_parser class
//
//      Parses a (possibly nested) sequence of 3 sub-matches. This class may
//      be used to parse structures, where the opening part is possibly
//      contained in the expression part and the whole sequence is only
//      parsed after seeing the closing part matching the first opening
//      subsequence. Example: nested PASCAL-comments:
//
//      { This is a { nested } PASCAL-comment }
//
///////////////////////////////////////////////////////////////////////////////
template <
    typename OpenT, typename ExprT, typename CloseT,
    typename NestedT = non_nested, 
    typename CategoryT = 
        typename impl::parser_type<ExprT>::parser_t::parser_category
>
struct confix_parser :
    public parser<confix_parser<OpenT, ExprT, CloseT, NestedT, CategoryT> >
{
    confix_parser(OpenT const &open_, ExprT const &expr_, CloseT const &close_);

    template <typename IteratorT>
    match
    parse(IteratorT& first, IteratorT const& last) const;

    template <typename IteratorT, typename MatchTraitsT>
    typename MatchTraitsT::match_t
    ast_parse(IteratorT& first, IteratorT const& last, 
            MatchTraitsT const&) const;

private:
    typename embed_trait<OpenT>::type open;
    typename embed_trait<ExprT>::type expr;
    typename embed_trait<CloseT>::type close;
};

///////////////////////////////////////////////////////////////////////////////
//
// Confix parser generator template
//
//      This is a helper for generating a correct confix_parser<> from
//      auxilliary parameters. There are the following types supported as
//      parameters yet: parsers, single characters and strings (see
//      impl::parser_type).
//
//      If ExprT is an action_parser_category type (parser with an attached 
//      semantic action) we have to do something special. This happens, if the 
//      user wrote something like:
//
//          confix_p(open, body[f], close)
//
//      where 'body' is the parser matching the body of the confix sequence 
//      and 'f' is a functor to be called after matching the body. If we would 
//      do nothing, the resulting code would parse the sequence as follows:
//
//          start >> *(body[f] - close) >> close
//
//      what in most cases is not what the user expects. 
//      (If this _is_ what you've expected, then please use the confix_p or
//      confix_nest_p generator function 'direct', which will inhibit 
//      re-attaching the actor to the body parser).
//
//      To make the confix parser behave as expected:
//
//          start >> (*(body - close))[f] >> close
//
//      the actor attached to the 'body' parser has to be re-attached to the
//      *(body - close) parser construct, which will make the resulting confix
//      parser 'do the right thing'.
//
///////////////////////////////////////////////////////////////////////////////
template<typename NestedT>
struct confix_parser_gen
{
    // Generic generator function for creation of concrete confix parsers

    template<typename StartT, typename ExprT, typename EndT>
    confix_parser<
        typename impl::parser_type<StartT>::parser_t,
        typename impl::parser_type<ExprT>::parser_t,
        typename impl::parser_type<EndT>::parser_t,
        NestedT,
        typename impl::parser_type<ExprT>::parser_t::parser_category
    >
    operator()(
        StartT const &start_, ExprT const &expr_, EndT const &end_) const;

    // Generic generator function for creation of concrete confix parsers
    // which have a action directly attached to the ExprT part of the
    // parser (see comment above)

    template<typename StartT, typename ExprT, typename EndT>
    confix_parser<
        typename impl::parser_type<StartT>::parser_t,
        typename impl::parser_type<ExprT>::parser_t,
        typename impl::parser_type<EndT>::parser_t,
        NestedT,
        plain_parser_category // do not re-attach action
    >
    direct(
        StartT const &start, ExprT const &expr, EndT const &end) const;
};

///////////////////////////////////////////////////////////////////////////////
//
//  Predefined non_nested and nested confix parser generators
//
///////////////////////////////////////////////////////////////////////////////
confix_parser_gen<non_nested> confix_p = confix_parser_gen<non_nested>();
confix_parser_gen<nested> confix_nest_p = confix_parser_gen<nested>();

///////////////////////////////////////////////////////////////////////////////
//
//  Comments are special types of confix parsers
//
//      Comment parser generator template. This is a helper for generating a
//      correct confix_parser<> from auxilliary parameters, which is able to
//      parse comment constructs: (StartToken >> Comment text >> EndToken).
//      There are the following types supported as parameters yet: parsers,
//      single characters and strings (see impl::parser_type).
//
///////////////////////////////////////////////////////////////////////////////
template<typename NestedT>
struct comment_parser_gen
{
    // generic generator function for creation of concrete comment parsers

    template<typename StartT, typename EndT>
    confix_parser<
        typename impl::parser_type<StartT>::parser_t,
        anychar_,
        typename impl::parser_type<EndT>::parser_t,
        NestedT
    >
    operator() (StartT const &start_, EndT const &end_) const;
};

///////////////////////////////////////////////////////////////////////////////
//
//  Predefined non_nested and nested comment parser generator
//
///////////////////////////////////////////////////////////////////////////////
comment_parser_gen<non_nested> comment_p = comment_parser_gen<non_nested>();
comment_parser_gen<nested> comment_nest_p = comment_parser_gen<nested>();

///////////////////////////////////////////////////////////////////////////////
// 
//  list_action class
//
//      Links an list parser with a user defined semantic action.
//      The semantic action may be a function or a functor. A function
//      should be compatible with the interface:
//
//          template <typename IteratorT>
//          void f (IteratorT it);
//
//      Where 'it' is an input iterator for iteration over the matched items
//      in the list. Dereferencing this iterator will yield in getting a
//      list item as a StringT. 
//
//      A functor should have a member operator() with a compatible signature
//      as above. The matching item is passed into the function/functor.
//      This is the default class that list parsers use when dealing with
//      the construct:
//
//          p[f]
//
//      where p is a parser and f is a function or functor.
//
///////////////////////////////////////////////////////////////////////////////
template <typename ParserT, typename ActionT>
struct list_action :
    public unary<ParserT>,
    public parser<list_action<ParserT, ActionT> > 
{
    typedef action_parser_category parser_category;

    list_action(ParserT const& subject, ActionT const& actor_);

    template <typename IteratorT>
    match
    parse(IteratorT& first, IteratorT const& last) const;

    template <typename IteratorT, typename MatchTraitsT>
    typename MatchTraitsT::match_t
    ast_parse(IteratorT& first, IteratorT const& last, 
            MatchTraitsT const&) const;

    ActionT const &predicate() const { return actor; }

private:

    typename embed_trait<ActionT>::type actor;
};

///////////////////////////////////////////////////////////////////////////////
// 
//  class list_action_iterator
//
//      The list_action_iterator is an input iterator, which allows access
//      to the successfully parsed list items.
//
///////////////////////////////////////////////////////////////////////////////
template <typename ParserT, typename IteratorT, typename BaseT = std::string>
struct list_action_iterator 
{
    typedef std::input_iterator_tag iterator_category;
    typedef BaseT                   value_type;
    typedef ptrdiff_t               difference_type;
    typedef BaseT *                 pointer;
    typedef BaseT &                 reference;

    list_action_iterator();
    list_action_iterator(
        ParserT const *parser, IteratorT &first, IteratorT const &last);

    // functions required for input iterators 
    BaseT const &operator* () const;
    list_action_iterator<ParserT, IteratorT, BaseT> &operator++ ();
    list_action_iterator<ParserT, IteratorT, BaseT> operator++ (int);
    bool operator== (
        list_action_iterator<ParserT, IteratorT, BaseT> const &rhs) const;
    bool operator!= (
        list_action_iterator<ParserT, IteratorT, BaseT> const &rhs) const;

    // functions required for parser integration
    IteratorT const &actposition() const;
    bool matched() const;
    void operator() (
        typename std::iterator_traits<IteratorT>::value_type const &match
        ) const;
    void operator() (
        IteratorT const &begin_match, IteratorT const &end_match) const;
    ParserT const &subject() const;

private:
    match read(bool firstread = false);

    ParserT const *parser;
    mutable BaseT value;
    IteratorT current;
    IteratorT const last;
    bool end_marker;
    bool saw_oneitem;
};

///////////////////////////////////////////////////////////////////////////////
// 
//  list_parser class
//
//      List parsers allow to parse constructs like 
//
//          item >> *(delim >> item)
//
//      where 'item' is an auxilliary expression to parse and 'delim' is an
//      auxilliary delimiter to parse. 
//
//      If ItemT is an action_parser_category type (parser with an attached  
//      semantic action) we have to do something special. This happens, if the 
//      user wrote something like:
//
//          list_p(item[f], delim)
//
//      where 'item' is the parser matching one item of the list sequence and 
//      'f' is a functor to be called after matching one item. If we would do 
//      nothing, the resulting code would parse the sequence as follows:
//
//          *(item[f] - delim) >> *(delim >> *(item[f] - delim))
//
//      what in most cases is not what the user expects. 
//      (If this _is_ what you've expected, then please use one of the list_p 
//      generator functions 'direct', which will inhibit re-attaching 
//      the actor to the item parser).
//
//      To make the list parser behave as expected:
//
//          (*(item - delim))[f] >> *(delim >> (*(item - delim))[f])
//
//      the actor attached to the 'item' parser has to be re-attached to the
//      *(item - delim) parser construct, which will make the resulting list
//      parser 'do the right thing'.
//
///////////////////////////////////////////////////////////////////////////////
template <
    typename ItemT, typename DelimT, 
    typename StringT = std::string,
    typename CategoryT = 
        typename impl::parser_type<ItemT>::parser_t::parser_category
>
struct list_parser :
    public parser<list_parser<ItemT, DelimT, StringT, CategoryT> >
{
    typedef StringT string_t;
    typedef CategoryT parser_category;

    list_parser(ItemT const &item_, DelimT const &delim_);

    template <typename IteratorT>
    match
    parse(IteratorT& first, IteratorT const& last) const;

    template <typename IteratorT, typename MatchTraitsT>
    typename MatchTraitsT::match_t
    ast_parse(IteratorT& first, IteratorT const& last, 
            MatchTraitsT const&) const;

    template <typename ActionT>
    list_action<list_parser<ItemT, DelimT, StringT, CategoryT>, ActionT>
    operator[](ActionT const& actor) const
    {
        //  Borland internal compiler error if this
        //  is defined elsewhere (JDG)
        return list_action<list_parser, ActionT>(*this, actor);
    }

    ItemT const &item_p() const { return item; }
    DelimT const &delim_p() const { return delim; }
    epsilon_ const &end_p() const { return epsilon; }

private:
    typename embed_trait<ItemT>::type item;
    typename embed_trait<DelimT>::type delim;
};

///////////////////////////////////////////////////////////////////////////////
// 
//  list_parser_ex class
//
//      Extended list parsers allow to parse constructs like 
//
//          item >> *(delim >> item) >> !end
//
//      where 'item' is an auxilliary expression to parse and 'delim' is an
//      auxilliary delimiter to parse. The 'end' token represents an optional
//      closing delimiter.
//
//      Same comments regarding re-attaching an eventually action code attached 
//      to the 'item' parser apply as in 'list_parser' (see above).
//
///////////////////////////////////////////////////////////////////////////////
template <
    typename ItemT, typename DelimT, typename EndT, 
    typename StringT = std::string,
    typename CategoryT = 
        typename impl::parser_type<ItemT>::parser_t::parser_category
>
struct list_parser_ex :
    public parser<list_parser_ex<ItemT, DelimT, EndT, StringT, CategoryT> >
{
    typedef StringT string_t;
    typedef CategoryT parser_category;

    list_parser_ex(
        ItemT const &item_, DelimT const &delim_, EndT const &end_);

    template <typename IteratorT>
    match
    parse(IteratorT& first, IteratorT const& last) const;

    template <typename IteratorT, typename MatchTraitsT>
    typename MatchTraitsT::match_t
    ast_parse(IteratorT& first, IteratorT const& last, 
            MatchTraitsT const&) const;

    template <typename ActionT>
    list_action<list_parser_ex<ItemT, DelimT, EndT, StringT, CategoryT>, 
                ActionT>
    operator[](ActionT const& actor) const
    {
        //  Borland internal compiler error if this
        //  is defined elsewhere (JDG)
        return list_action<list_parser_ex, ActionT>(*this, actor);
    }

    ItemT const &item_p() const { return item; }
    DelimT const &delim_p() const { return delim; }
    EndT const &end_p() const { return end; }

private:
    typename embed_trait<ItemT>::type item;
    typename embed_trait<DelimT>::type delim;
    typename embed_trait<EndT>::type end;
};

///////////////////////////////////////////////////////////////////////////////
//
// List parser generator template
//
//      This is a helper for generating a correct list_parser<> from
//      auxilliary parameters. There are the following types supported as
//      parameters yet: parsers, single characters and strings (see
//      impl::parser_type).
//
//      The list_parser_gen by itself can be used for parsing comma separated
//      lists without item formatting. 
//
///////////////////////////////////////////////////////////////////////////////
template <typename StringT = std::string>
struct list_parser_gen :
    public list_parser<anychar_, chlit<typename StringT::value_type>, StringT>
{
    typedef typename StringT::value_type CharT;

// construct the list_parser_gen object as an list parser for comma separated
// lists without item formatting.
    list_parser_gen() :
        list_parser<anychar_, chlit<CharT>, StringT>(anychar, chlit<CharT>(','))
    {
    }

// attach an action to the list_parser_gen by itself
    template <typename ActionT>
    list_action<list_parser<anychar_, chlit<CharT>, StringT>, ActionT>
    operator[](ActionT const& actor) const
    {
        //  Borland internal compiler error if this
        //  is defined elsewhere (JDG)
        typedef 
            list_action<list_parser<anychar_, chlit<CharT>, StringT>, ActionT> 
            return_t;

        return return_t(*this, actor);
    }

// The following generator functions should be used under normal circumstances.
// (the operator()(...) functions)

    // Generic generator functions for creation of concrete list parsers, which 
    // support 'normal' syntax: 
    //
    //      item >> *(delim >> item)
    //
    // If item isn't given, everything between two delimiters is matched.

    template<typename DelimT>
    list_parser<
        anychar_,
        typename impl::parser_type<DelimT>::parser_t,
        StringT,
        plain_parser_category          // anychar is a plain_parser type
    >
    operator()(DelimT const &delim_) const;

    template<typename ItemT, typename DelimT>
    list_parser<
        typename impl::parser_type<ItemT>::parser_t,
        typename impl::parser_type<DelimT>::parser_t,
        StringT
    >
    operator()(ItemT const &item_, DelimT const &delim_) const;

    // Generic generator function for creation of concrete list parsers, which 
    // support 'extended' syntax: 
    //
    //      item >> *(delim >> item) >> !end
    
    template<typename ItemT, typename DelimT, typename EndT>
    list_parser_ex<
        typename impl::parser_type<ItemT>::parser_t,
        typename impl::parser_type<DelimT>::parser_t,
        typename impl::parser_type<EndT>::parser_t,
        StringT
    >
    operator()(
        ItemT const &item_, DelimT const &delim_, EndT const &end_) const;

// The following functions should be used, if the 'item' parser has an attached 
// semantic action and this action should _not_ be re-attached during parser 
// construction (see comment above).

    // Generic generator function for creation of concrete list parsers, which 
    // support 'normal' syntax: 
    //
    //      item >> *(delim >> item)

    template<typename ItemT, typename DelimT>
    list_parser<
        typename impl::parser_type<ItemT>::parser_t,
        typename impl::parser_type<DelimT>::parser_t,
        StringT,
        plain_parser_category        // inhibit action re-attachment
    >
    direct(ItemT const &item_, DelimT const &delim_) const;

    // Generic generator function for creation of concrete list parsers, which 
    // support 'extended' syntax: 
    //
    //      item >> *(delim >> item) >> !end
    
    template<typename ItemT, typename DelimT, typename EndT>
    list_parser_ex<
        typename impl::parser_type<ItemT>::parser_t,
        typename impl::parser_type<DelimT>::parser_t,
        typename impl::parser_type<EndT>::parser_t,
        StringT,
        plain_parser_category        // inhibit action re-attachment
    >
    direct(
        ItemT const &item_, DelimT const &delim_, EndT const &end_) const;
};

///////////////////////////////////////////////////////////////////////////////
//
//  Predefined list parser generator
//
//      The list_p parser generator can be used 
//        - by itself for parsing comma separated lists without item formatting 
//      or 
//        - for generating list parsers with auxilliary parser parameters 
//          for the 'item', 'delim' and 'end' subsequences.
//
///////////////////////////////////////////////////////////////////////////////
list_parser_gen<> list_p = list_parser_gen<>();

} // namespace spirit

#endif
