///////////////////////////////////////////////////////////////////////////////
// This sample shows, how to use attr_rule<>'s for implementing a simple
// (RPN style) calculator

// #define SPIRIT_DEBUG

#include <boost/spirit/spirit.hpp>
#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// used namespaces
using namespace spirit;
using namespace std;

///////////////////////////////////////////////////////////////////////////////
// helper for raising power
template <class _Tp>
struct raise_power :
    public binary_function<_Tp, _Tp, _Tp>
{
    _Tp operator()(const _Tp& __x, const _Tp& __y) const
    {
        return std::pow(__x, __y);
    }
};

///////////////////////////////////////////////////////////////////////////////
// helper for working with two values
template <typename ClosureT, typename PredT>
struct expr {
    expr (ClosureT const &clos_) : clos(clos_) {}

    template<typename ItT>
    void operator()(ItT &) const
    {
        clos.set(m0, PredT()(clos.get(m0), clos.get(m1)));
    }
    ClosureT const &clos;
};

///////////////////////////////////////////////////////////////////////////////
// helper for working with one value
template <typename ClosureT, typename PredT>
struct unary_expr {
    unary_expr (ClosureT const &clos_) : clos(clos_) {}

    template<typename ItT>
    void operator()(ItT &) const
    {
        clos.set(m0, PredT()(clos.get(m0)));
    }
    ClosureT const &clos;
};

///////////////////////////////////////////////////////////////////////////////
// The class 'RPNCalculator' implements a polish reverse notation calculator
// which is equivalent to the following YACC description.
//
//      exp:      NUM             { $$ = $1;         }
//              | exp exp '+'     { $$ = $1 + $2;    }
//              | exp exp '-'     { $$ = $1 - $2;    }
//              | exp exp '*'     { $$ = $1 * $2;    }
//              | exp exp '/'     { $$ = $1 / $2;    }
//            /* Exponentiation */
//              | exp exp '^'     { $$ = pow ($1, $2); }
//            /* Unary minus    */
//              | exp 'n'         { $$ = -$1;        }
//      ;
//
// The different notation results from the requirement of RD parsers not to
// contain left recursion in their grammar (would lead to endless recursion).
// Therefor the left recursion in the YACC script before is transformated into
// a equivalent right recursion notation, which is in most cases less
// intuitive.

// Note: the types in the closure<> template parameter used in the following
// class definition are selected different intentionally, because the
// IntelV5.0.1 compiler fails to generate correct code for closures with
// identical types as closure members. This may be (should be) no problem for
// other compilers.
struct RPNCalculator :
    public grammar<RPNCalculator, double>
{
    template <typename IteratorT>
    struct definition
    {
        typedef closure<long double, double> closure_t;
        typedef attr_rule<closure_t, IteratorT> attr_rule_t;

        attr_rule_t exp;

        definition()
        {
            expr<attr_rule_t, std::plus<return_t> >         plus(exp);
            expr<attr_rule_t, std::minus<return_t> >        minus(exp);
            expr<attr_rule_t, std::multiplies<return_t> >   times(exp);
            expr<attr_rule_t, std::divides<return_t> >      divide(exp);
            expr<attr_rule_t, raise_power<return_t> >  power(exp);
            unary_expr<attr_rule_t, std::negate<return_t> > negate(exp);

            exp =
                real_p[exp(m0)] >>
                *(
                    exp[exp(m1)] >>
                    (
                            ch_p('+')[plus]
                        |   ch_p('-')[minus]
                        |   ch_p('*')[times]
                        |   ch_p('/')[divide]
                        |   ch_p('^')[power]
                    )
                |   ch_p('n')[negate]
                )
                ;

            SPIRIT_DEBUG_RULE(exp);
        }

        attr_rule<closure_t, IteratorT> const& start() const
        {
            return exp;
        }
    };
};

///////////////////////////////////////////////////////////////////////////////
// main entry point
int main()
{
RPNCalculator exp;

    cout << "/////////////////////////////////////////////////////////\n\n";
    cout << "\t\tThe simplest working RPN calculator...\n\n";
    cout << "/////////////////////////////////////////////////////////\n\n";
    cout << "Type an expression...or [q or Q] to quit\n\n";

    while (true)
    {
        char str[256];
        cin.getline(str, 256);
        if (str[0] == 'q' || str[0] == 'Q')
            break;

        const char* str_begin = str;
        const char* str_end = str;
        while (*str_end)
            ++str_end;

        skipper<> skip(space, str_end);
        scanner<> first(str_begin, &skip);
        scanner<> last(str_end, &skip);

    double result = 0.0;

        match hit = exp[ref(result)].parse(first, last);

        if (hit && first == last)
        {
            cout << "Parsing succeeded!\n";
            cout << "Result = " << result << "\n\n";
        }
        else
        {
            cout << "Parsing failed!\n";
        }
    }

    cout << "Bye... :-) \n\n";
    return 0;
}

