//
// C Grammar checker implemented with Spirit (http://spirit.sourceforge.net/)
//
// Copyright 2001 Hartmut Kaiser, All rights reserved
// email: HartmutKaiser@t-online.de
// Created: 15.10.2001 09:25:56

// Adapted from:
//	http://www.lysator.liu.se/c/ANSI-C-grammar-y.html
//	http://www.lysator.liu.se/c/ANSI-C-grammar-l.html

///////////////////////////////////////////////////////////////////////////////
// version history
//
// TODO:	Handling typedef's, without that we can't correctly parse
//			auxilliary valid C sources. But hey it's a sample.
//
// Modified: 18.10.2001 16:36:07
//	Corrected: String literals may be splitted in several strings: "abc" "def"
//	Corrected: String literals may have length 0 ("")
//
// Modified: 17.10.2001 10:56:57
//	Rewritten:	*_CONSTANT rules
//	Introduced: operator '||' where possible
//	Corrected:	IDENTIFIER can start with '_' and can contain '$'
//	Added:		Skipping of '#line' and '#pragma' directives, which contained in
//				in preprocessed files of the Intel V5.0.1 compiler
//
// Modified: 16.10.2001 21:12:05
//	Corrected: missing lexeme[] statements for *_CONSTANT
//	Corrected: missing longest[] for CONSTANT

#include <iostream>
#include <fstream>
#include <vector>

#if defined(_DEBUG)
#define SPIRIT_DEBUG
#endif // defined(_DEBUG)

#include <boost/spirit/spirit.hpp>

///////////////////////////////////////////////////////////////////////////////
// used namespaces
using namespace spirit;
using namespace std;

///////////////////////////////////////////////////////////////////////////////
// parsing helper function
static void
parse(rule<> const& r, char const* filename)
{
ifstream in(filename);

	if (!in) {
		cerr << "Could not open file: " << filename << endl;
		return;
	}

// Turn of white space skipping on the stream
	in.unsetf(ios::skipws);

//  Here's our nested C comment rule
rule<char const *> comment
        =   "//" >> *(anychar - '\n') >> '\n'		//  C++ comment
        |   "/*" >> *(anychar - "*/") >> "*/"		//  C comment
		|	"#line" >> *(anychar - '\n') >> '\n'	// added for correctly handling preprocessed
		|	"#pragma" >> *(anychar - '\n') >> '\n'	// files from Intel V5.0.1 on W2K
        ;

// whoops, this is a function declaration!
//vector<char> vec (istream_iterator<char>(in),istream_iterator<char>());

vector<char> vec = vector<char>(istream_iterator<char>(in),istream_iterator<char>());
vector<char> const& cvec = vec;

parse_info<char const *> result = parse(&*cvec.begin(), &*cvec.end(), r, space | comment);

    if (result.full)
		cerr << filename << " Parses OK" << endl;
    else {
		cerr << filename << " Fails Parsing" << endl;
		for (int i = 0; i < 50; i++)
		{
		    if (result.stop == &*cvec.end())
		        break;
		    cerr << *result.stop++;
	    }
		cerr << endl;
    }
}

///////////////////////////////////////////////////////////////////////////////
// main entrypoint
int
main(int argc, char* argv[])
{
// Start grammar definition
	cerr << "C Grammar checker implemented with Spirit ..." << endl;

// C keywords
symbols<> keywords;

    keywords =
        "auto", "break", "case", "char", "const", "continue", "default", "do", "double",
        "else", "enum", "extern", "float", "for", "goto", "if", "int",
        "long", "register", "return", "short", "signed", "sizeof", "static",
        "struct", "switch", "typedef", "union", "unsigned", "void", "volatile", "while";

// C operators
strlit<> ELLIPSIS("...");
strlit<> RIGHT_ASSIGN(">>=");
strlit<> LEFT_ASSIGN("<<=");
strlit<> ADD_ASSIGN("+=");
strlit<> SUB_ASSIGN("-=");
strlit<> MUL_ASSIGN("*=");
strlit<> DIV_ASSIGN("/=");
strlit<> MOD_ASSIGN("%=");
strlit<> AND_ASSIGN("&=");
strlit<> XOR_ASSIGN("^=");
strlit<> OR_ASSIGN("|=");
strlit<> RIGHT_OP(">>");
strlit<> LEFT_OP("<<");
strlit<> INC_OP("++");
strlit<> DEC_OP("--");
strlit<> PTR_OP("->");
strlit<> AND_OP("&&");
strlit<> OR_OP("||");
strlit<> LE_OP("<=");
strlit<> GE_OP(">=");
strlit<> EQ_OP("==");
strlit<> NE_OP("!=");
chlit<> SEMICOLON(';');
rule<> LEFT_BRACE = chlit<>('{') | strlit<>("<%");
rule<> RIGHT_BRACE = chlit<>('}') | strlit<>("%>");
chlit<> COMMA(',');
chlit<> COLON(':');
chlit<> ASSIGN('=');
chlit<> LEFT_PAREN('(');
chlit<> RIGHT_PAREN(')');
rule<> LEFT_BRACKET = chlit<>('[') | strlit<>("<:");
rule<> RIGHT_BRACKET = chlit<>(']') | strlit<>(":>");
chlit<> DOT('.');
chlit<> ADDROF('&');
chlit<> BANG('!');
chlit<> TILDE('~');
chlit<> MINUS('-');
chlit<> PLUS('+');
chlit<> STAR('*');
chlit<> SLASH('/');
chlit<> PERCENT('%');
chlit<> LT_OP('<');
chlit<> GT_OP('>');
chlit<> XOR('^');
chlit<> OR('|');
chlit<> QUEST('?');

// Tokens
rule<> AUTO = strlit<>("auto");
rule<> BREAK = strlit<>("break");
rule<> CASE = strlit<>("case");
rule<> CHAR = strlit<>("char");
rule<> CONST = strlit<>("const");
rule<> CONTINUE = strlit<>("continue");
rule<> DEFAULT = strlit<>("default");
rule<> DO = strlit<>("do");
rule<> DOUBLE = strlit<>("double");
rule<> ELSE = strlit<>("else");
rule<> ENUM = strlit<>("enum");
rule<> EXTERN = strlit<>("extern");
rule<> FOR = strlit<>("for");
rule<> FLOAT = strlit<>("float");
rule<> GOTO = strlit<>("goto");
rule<> IF = strlit<>("if");
rule<> INT = strlit<>("int");
rule<> LONG = strlit<>("long");
rule<> REGISTER = strlit<>("register");
rule<> RETURN = strlit<>("return");
rule<> SHORT = strlit<>("short");
rule<> SIGNED = strlit<>("signed");
rule<> SIZEOF = strlit<>("sizeof");
rule<> STATIC = strlit<>("static");
rule<> STRUCT = strlit<>("struct");
rule<> SWITCH = strlit<>("switch");
rule<> TYPEDEF = strlit<>("typedef");
rule<> UNION = strlit<>("union");
rule<> UNSIGNED = strlit<>("unsigned");
rule<> VOID = strlit<>("void");
rule<> VOLATILE = strlit<>("volatile");
rule<> WHILE = strlit<>("while");

// C identifiers
rule<> IDENTIFIER =
	lexeme[
		((alpha | '_' | '$') >> *(alnum | '_' | '$')) - (keywords >> anychar - (alnum | '_' | '$'))
	];

// string literals
rule<> STRING_LITERAL_PART =
	lexeme[
		!chlit<>('L') >> chlit<>('\"') >>
		*( strlit<>("\"\"") | anychar - chlit<>('\"') ) >>
		chlit<>('\"')
	];
rule<> STRING_LITERAL = +STRING_LITERAL_PART;

// integer constants
rule<> INT_CONSTANT_HEX;		// hex constant
rule<> INT_CONSTANT_OCT;		// octal constant
rule<> INT_CONSTANT_DEC;		// decimal constant
rule<> INT_CONSTANT_CHAR;	// character constant

	INT_CONSTANT_HEX
		= lexeme[
			chlit<>('0') >> nocase[chlit<>('x')] >> +xdigit >> !nocase[chlit<>('l') | chlit<>('u')]
		]
		;

	INT_CONSTANT_OCT
		= lexeme[
			chlit<>('0') >> +range<>('0', '7') >> !nocase[chlit<>('l') | chlit<>('u')]
		]
		;

	INT_CONSTANT_DEC
		= lexeme[
			+digit >> !nocase[chlit<>('l') | chlit<>('u')]
		]
		;

	INT_CONSTANT_CHAR
		= lexeme[
			!chlit<>('L') >> chlit<>('\'') >>
			longest[
					anychar
				|	(chlit<>('\\') >> chlit<>('0') >> range<>('0', '7').repeat(0, 2))
				|	(chlit<>('\\') >> anychar)
			] >>
			chlit<>('\'')
		]
		;

rule<> INT_CONSTANT = INT_CONSTANT_HEX | INT_CONSTANT_OCT | INT_CONSTANT_DEC | INT_CONSTANT_CHAR;

// float constants
rule<> FLOAT_CONSTANT_1;
rule<> FLOAT_CONSTANT_2;
rule<> FLOAT_CONSTANT_3;

	FLOAT_CONSTANT_1	// 12345[eE][+-]123[lLfF]?
		= lexeme[
			+digit >>
			(chlit<>('e') | chlit<>('E')) >> !(chlit<>('+') | chlit<>('-')) >> +digit >>
			!nocase[chlit<>('l') | chlit<>('f')]
		]
		;

	FLOAT_CONSTANT_2	// .123([[eE][+-]123)?[lLfF]?
		= lexeme[
			*digit >> chlit<>('.') >> +digit >>
			!((chlit<>('e') | chlit<>('E')) >> !(chlit<>('+') | chlit<>('-')) >> +digit) >>
			!nocase[chlit<>('l') | chlit<>('f')]
		]
		;

	FLOAT_CONSTANT_3	// 12345.([[eE][+-]123)?[lLfF]?
		= lexeme[
			+digit >> chlit<>('.') >> *digit >>
			!((chlit<>('e') | chlit<>('E')) >> !(chlit<>('+') | chlit<>('-')) >> +digit) >>
			!nocase[chlit<>('l') | chlit<>('f')]
		]
		;

rule<> FLOAT_CONSTANT = FLOAT_CONSTANT_1 | FLOAT_CONSTANT_2 | FLOAT_CONSTANT_3;

rule<> CONSTANT = longest[FLOAT_CONSTANT | INT_CONSTANT];

// Rules forward declarations
rule<> primary_expression;
rule<> postfix_expression;
rule<> postfix_expression_helper;
rule<> argument_expression_list;
rule<> unary_expression;
rule<> unary_operator;
rule<> cast_expression;
rule<> multiplicative_expression;
rule<> multiplicative_expression_helper;
rule<> additive_expression;
rule<> additive_expression_helper;
rule<> shift_expression;
rule<> shift_expression_helper;
rule<> relational_expression;
rule<> relational_expression_helper;
rule<> equality_expression;
rule<> equality_expression_helper;
rule<> and_expression;
rule<> and_expression_helper;
rule<> exclusive_or_expression;
rule<> exclusive_or_expression_helper;
rule<> inclusive_or_expression;
rule<> inclusive_or_expression_helper;
rule<> logical_and_expression;
rule<> logical_and_expression_helper;
rule<> logical_or_expression;
rule<> logical_or_expression_helper;
rule<> conditional_expression;
rule<> conditional_expression_helper;
rule<> assignment_expression;
rule<> assignment_operator;
rule<> expression;
rule<> expression_helper;
rule<> constant_expression;
rule<> declaration;
rule<> declaration_specifiers;
rule<> init_declarator_list;
rule<> init_declarator;
rule<> storage_class_specifier;
rule<> type_specifier;
rule<> struct_or_union_specifier;
rule<> struct_or_union;
rule<> struct_declaration_list;
rule<> struct_declaration;
rule<> specifier_qualifier_list;
rule<> struct_declarator_list;
rule<> struct_declarator;
rule<> enum_specifier;
rule<> enumerator_list;
rule<> enumerator;
rule<> type_qualifier;
rule<> declarator;
rule<> direct_declarator;
rule<> direct_declarator_helper;
rule<> pointer;
rule<> type_qualifier_list;
rule<> parameter_type_list;
rule<> parameter_list;
rule<> parameter_declaration;
rule<> identifier_list;
rule<> type_name;
rule<> abstract_declarator;
rule<> direct_abstract_declarator;
rule<> direct_abstract_declarator_helper;
rule<> initializer;
rule<> initializer_list;
rule<> statement;
rule<> labeled_statement;
rule<> compound_statement;
rule<> declaration_list;
rule<> statement_list;
rule<> expression_statement;
rule<> selection_statement;
rule<> iteration_statement;
rule<> jump_statement;
rule<> translation_unit;
rule<> external_declaration;
rule<> function_definition;

// debug support for terminals
SPIRIT_DEBUG_RULE(AUTO);
SPIRIT_DEBUG_RULE(BREAK);
SPIRIT_DEBUG_RULE(CASE);
SPIRIT_DEBUG_RULE(CHAR);
SPIRIT_DEBUG_RULE(CONST);
SPIRIT_DEBUG_RULE(CONTINUE);
SPIRIT_DEBUG_RULE(DEFAULT);
SPIRIT_DEBUG_RULE(DO);
SPIRIT_DEBUG_RULE(DOUBLE);
SPIRIT_DEBUG_RULE(ELSE);
SPIRIT_DEBUG_RULE(ENUM);
SPIRIT_DEBUG_RULE(EXTERN);
SPIRIT_DEBUG_RULE(FOR);
SPIRIT_DEBUG_RULE(FLOAT);
SPIRIT_DEBUG_RULE(GOTO);
SPIRIT_DEBUG_RULE(IF);
SPIRIT_DEBUG_RULE(INT);
SPIRIT_DEBUG_RULE(LONG);
SPIRIT_DEBUG_RULE(REGISTER);
SPIRIT_DEBUG_RULE(RETURN);
SPIRIT_DEBUG_RULE(SHORT);
SPIRIT_DEBUG_RULE(SIGNED);
SPIRIT_DEBUG_RULE(SIZEOF);
SPIRIT_DEBUG_RULE(STATIC);
SPIRIT_DEBUG_RULE(STRUCT);
SPIRIT_DEBUG_RULE(SWITCH);
SPIRIT_DEBUG_RULE(TYPEDEF);
SPIRIT_DEBUG_RULE(UNION);
SPIRIT_DEBUG_RULE(UNSIGNED);
SPIRIT_DEBUG_RULE(VOID);
SPIRIT_DEBUG_RULE(VOLATILE);
SPIRIT_DEBUG_RULE(WHILE);
SPIRIT_DEBUG_RULE(IDENTIFIER);
SPIRIT_DEBUG_RULE(STRING_LITERAL);
SPIRIT_DEBUG_RULE(INT_CONSTANT_HEX);
SPIRIT_DEBUG_RULE(INT_CONSTANT_OCT);
SPIRIT_DEBUG_RULE(INT_CONSTANT_DEC);
SPIRIT_DEBUG_RULE(INT_CONSTANT_CHAR);
SPIRIT_DEBUG_RULE(INT_CONSTANT);
SPIRIT_DEBUG_RULE(FLOAT_CONSTANT_1);
SPIRIT_DEBUG_RULE(FLOAT_CONSTANT_2);
SPIRIT_DEBUG_RULE(FLOAT_CONSTANT_3);
SPIRIT_DEBUG_RULE(FLOAT_CONSTANT);
SPIRIT_DEBUG_RULE(CONSTANT);
// debug support for non terminals
SPIRIT_DEBUG_RULE(primary_expression);
SPIRIT_DEBUG_RULE(postfix_expression);
SPIRIT_DEBUG_RULE(postfix_expression_helper);
SPIRIT_DEBUG_RULE(argument_expression_list);
SPIRIT_DEBUG_RULE(unary_expression);
SPIRIT_DEBUG_RULE(unary_operator);
SPIRIT_DEBUG_RULE(cast_expression);
SPIRIT_DEBUG_RULE(multiplicative_expression);
SPIRIT_DEBUG_RULE(multiplicative_expression_helper);
SPIRIT_DEBUG_RULE(additive_expression);
SPIRIT_DEBUG_RULE(additive_expression_helper);
SPIRIT_DEBUG_RULE(shift_expression);
SPIRIT_DEBUG_RULE(shift_expression_helper);
SPIRIT_DEBUG_RULE(relational_expression);
SPIRIT_DEBUG_RULE(relational_expression_helper);
SPIRIT_DEBUG_RULE(equality_expression);
SPIRIT_DEBUG_RULE(equality_expression_helper);
SPIRIT_DEBUG_RULE(and_expression);
SPIRIT_DEBUG_RULE(and_expression_helper);
SPIRIT_DEBUG_RULE(exclusive_or_expression);
SPIRIT_DEBUG_RULE(exclusive_or_expression_helper);
SPIRIT_DEBUG_RULE(inclusive_or_expression);
SPIRIT_DEBUG_RULE(inclusive_or_expression_helper);
SPIRIT_DEBUG_RULE(logical_and_expression);
SPIRIT_DEBUG_RULE(logical_and_expression_helper);
SPIRIT_DEBUG_RULE(logical_or_expression);
SPIRIT_DEBUG_RULE(logical_or_expression_helper);
SPIRIT_DEBUG_RULE(conditional_expression);
SPIRIT_DEBUG_RULE(conditional_expression_helper);
SPIRIT_DEBUG_RULE(assignment_expression);
SPIRIT_DEBUG_RULE(assignment_operator);
SPIRIT_DEBUG_RULE(expression);
SPIRIT_DEBUG_RULE(constant_expression);
SPIRIT_DEBUG_RULE(declaration);
SPIRIT_DEBUG_RULE(declaration_specifiers);
SPIRIT_DEBUG_RULE(init_declarator_list);
SPIRIT_DEBUG_RULE(init_declarator);
SPIRIT_DEBUG_RULE(storage_class_specifier);
SPIRIT_DEBUG_RULE(type_specifier);
SPIRIT_DEBUG_RULE(struct_or_union_specifier);
SPIRIT_DEBUG_RULE(struct_or_union);
SPIRIT_DEBUG_RULE(struct_declaration_list);
SPIRIT_DEBUG_RULE(struct_declaration);
SPIRIT_DEBUG_RULE(specifier_qualifier_list);
SPIRIT_DEBUG_RULE(struct_declarator_list);
SPIRIT_DEBUG_RULE(struct_declarator);
SPIRIT_DEBUG_RULE(enum_specifier);
SPIRIT_DEBUG_RULE(enumerator_list);
SPIRIT_DEBUG_RULE(enumerator);
SPIRIT_DEBUG_RULE(type_qualifier);
SPIRIT_DEBUG_RULE(declarator);
SPIRIT_DEBUG_RULE(direct_declarator);
SPIRIT_DEBUG_RULE(direct_declarator_helper);
SPIRIT_DEBUG_RULE(pointer);
SPIRIT_DEBUG_RULE(type_qualifier_list);
SPIRIT_DEBUG_RULE(parameter_type_list);
SPIRIT_DEBUG_RULE(parameter_list);
SPIRIT_DEBUG_RULE(parameter_declaration);
SPIRIT_DEBUG_RULE(identifier_list);
SPIRIT_DEBUG_RULE(type_name);
SPIRIT_DEBUG_RULE(abstract_declarator);
SPIRIT_DEBUG_RULE(direct_abstract_declarator);
SPIRIT_DEBUG_RULE(direct_abstract_declarator_helper);
SPIRIT_DEBUG_RULE(initializer);
SPIRIT_DEBUG_RULE(initializer_list);
SPIRIT_DEBUG_RULE(statement);
SPIRIT_DEBUG_RULE(labeled_statement);
SPIRIT_DEBUG_RULE(compound_statement);
SPIRIT_DEBUG_RULE(declaration_list);
SPIRIT_DEBUG_RULE(statement_list);
SPIRIT_DEBUG_RULE(expression_statement);
SPIRIT_DEBUG_RULE(selection_statement);
SPIRIT_DEBUG_RULE(iteration_statement);
SPIRIT_DEBUG_RULE(jump_statement);
SPIRIT_DEBUG_RULE(translation_unit);
SPIRIT_DEBUG_RULE(external_declaration);
SPIRIT_DEBUG_RULE(function_definition);

// Rules
	primary_expression
		= IDENTIFIER
		| CONSTANT
		| STRING_LITERAL
		| LEFT_PAREN >> expression >> RIGHT_PAREN
		;

// left recursion --> right recursion
//	postfix_expression
//		= primary_expression
//		| postfix_expression >>
//			(
//				LEFT_BRACKET >> expression >> RIGHT_BRACKET
//			|	LEFT_PAREN >> !argument_expression_list >> RIGHT_PAREN
//			|	DOT >> IDENTIFIER
//			|	PTR_OP >> IDENTIFIER
//			|	INC_OP
//			|	DEC_OP
//			)
//		;
	postfix_expression
		= primary_expression >> postfix_expression_helper
		;

	postfix_expression_helper
		= 	(
				LEFT_BRACKET >> expression >> RIGHT_BRACKET
			|	LEFT_PAREN >> !argument_expression_list >> RIGHT_PAREN
			|	DOT >> IDENTIFIER
			|	PTR_OP >> IDENTIFIER
			|	INC_OP
			|	DEC_OP
			) >>
			postfix_expression_helper
		| epsilon
		;

	argument_expression_list
		= assignment_expression >> *(COMMA >> assignment_expression)
		;

	unary_expression
		= postfix_expression
		| INC_OP >> unary_expression
		| DEC_OP >> unary_expression
		| unary_operator >> cast_expression
		| SIZEOF >>
			(
				unary_expression
			|	LEFT_PAREN >> type_name >> RIGHT_PAREN
			)
		;

	unary_operator
		= ADDROF
		| STAR
		| PLUS
		| MINUS
		| TILDE
		| BANG
		;

	cast_expression
		= LEFT_PAREN >> type_name >> RIGHT_PAREN >> cast_expression
		| unary_expression
		;

// left recursion --> right recursion
//	multiplicative_expression
//		= cast_expression
//		| multiplicative_expression >>
//		(
//				STAR >> cast_expression
//			|	SLASH >> cast_expression
//			|	PERCENT >> cast_expression
//		)
//		;
	multiplicative_expression
		= cast_expression >> multiplicative_expression_helper
		;

	multiplicative_expression_helper
		=	(
				STAR >> cast_expression
			|	SLASH >> cast_expression
			|	PERCENT >> cast_expression
			) >>
			multiplicative_expression_helper
		| epsilon
		;

// left recursion --> right recursion
//	additive_expression
//		= multiplicative_expression
//		| additive_expression >>
//		(
//				PLUS >> multiplicative_expression
//			|	MINUS >> multiplicative_expression
//		)
//		;
	additive_expression
		= multiplicative_expression >> additive_expression_helper
		;

	additive_expression_helper
		=	(
				PLUS >> multiplicative_expression
			|	MINUS >> multiplicative_expression
			) >>
			additive_expression_helper
		| epsilon
		;

// left recursion --> right recursion
//	shift_expression
//		= additive_expression
//		| shift_expression >>
//		(
//				LEFT_OP >> additive_expression
//			|	RIGHT_OP >> additive_expression
//		)
//		;
	shift_expression
		= additive_expression >> shift_expression_helper
		;

	shift_expression_helper
		=	(
				LEFT_OP >> additive_expression
			|	RIGHT_OP >> additive_expression
			) >>
			shift_expression_helper
		| epsilon
		;

// left recursion --> right recursion
//	relational_expression
//		= shift_expression
//		| relational_expression >>
//		(
//				LT_OP >> shift_expression
//			|	GT_OP >> shift_expression
//			|	LE_OP >> shift_expression
//			|	GE_OP >> shift_expression
//		)
//		;
	relational_expression
		= shift_expression >> relational_expression_helper
		;

	relational_expression_helper
		=	(
				LT_OP >> shift_expression
			|	GT_OP >> shift_expression
			|	LE_OP >> shift_expression
			|	GE_OP >> shift_expression
			) >>
			relational_expression_helper
		| epsilon
		;

// left recursion --> right recursion
//	equality_expression
//		= relational_expression
//		| equality_expression >>
//		(
//				EQ_OP >> relational_expression
//			|	NE_OP >> relational_expression
//		)
//		;
	equality_expression
		= relational_expression >> equality_expression_helper
		;

	equality_expression_helper
		=	(
				EQ_OP >> relational_expression
			|	NE_OP >> relational_expression
			) >>
			equality_expression_helper
		| epsilon
		;

// left recursion --> right recursion
//	and_expression
//		= equality_expression
//		| and_expression >> ADDROF >> equality_expression
//		;
	and_expression
		= equality_expression >> and_expression_helper
		;

	and_expression_helper
		= ADDROF >> equality_expression >> and_expression_helper
		| epsilon
		;

// left recursion --> right recursion
//	exclusive_or_expression
//		= and_expression
//		| exclusive_or_expression >> XOR >> and_expression
//		;
	exclusive_or_expression
		= and_expression >> exclusive_or_expression_helper
		;

	exclusive_or_expression_helper
		= XOR >> and_expression >> exclusive_or_expression_helper
		| epsilon
		;

// left recursion --> right recursion
//	inclusive_or_expression
//		= exclusive_or_expression
//		| inclusive_or_expression >> OR >> exclusive_or_expression
//		;
	inclusive_or_expression
		= exclusive_or_expression >> inclusive_or_expression_helper
		;

	inclusive_or_expression_helper
		= OR >> exclusive_or_expression >> inclusive_or_expression_helper
		| epsilon
		;

// left recursion --> right recursion
//	logical_and_expression
//		= inclusive_or_expression
//		| logical_and_expression >> AND_OP >> inclusive_or_expression
//		;
	logical_and_expression
		= inclusive_or_expression >> logical_and_expression_helper
		;

	logical_and_expression_helper
		= AND_OP >> inclusive_or_expression >> logical_and_expression_helper
		| epsilon
		;

// left recursion --> right recursion
//	logical_or_expression
//		= logical_and_expression
//		| logical_or_expression >> OR_OP >> logical_and_expression
//		;
	logical_or_expression
		= logical_and_expression >> logical_or_expression_helper
		;

	logical_or_expression_helper
		= OR_OP >> logical_and_expression >> logical_or_expression_helper
		| epsilon
		;

// left recursion --> right recursion
//	conditional_expression
//		= logical_or_expression
//		| logical_or_expression >> QUEST >> expression >> COLON >> conditional_expression
//		;
	conditional_expression
		= logical_or_expression >> conditional_expression_helper
		;

	conditional_expression_helper
		= QUEST >> expression >> COLON >> conditional_expression >> conditional_expression_helper
		| epsilon
		;

	assignment_expression
		= unary_expression >> assignment_operator >> assignment_expression
		| conditional_expression
		;

	assignment_operator
		= ASSIGN
		| MUL_ASSIGN
		| DIV_ASSIGN
		| MOD_ASSIGN
		| ADD_ASSIGN
		| SUB_ASSIGN
		| LEFT_ASSIGN
		| RIGHT_ASSIGN
		| AND_ASSIGN
		| XOR_ASSIGN
		| OR_ASSIGN
		;

// left recursion --> right recursion
//	expression
//		= assignment_expression
//		| expression >> COMMA >> assignment_expression
//		;
	expression
		= assignment_expression >> expression_helper
		;

	expression_helper
		= COMMA >> assignment_expression >> expression_helper
		| epsilon
		;

	constant_expression
		= conditional_expression
		;

	declaration
		= declaration_specifiers >> !init_declarator_list >> SEMICOLON
		;

	declaration_specifiers
		=	(
				storage_class_specifier
			|	type_specifier
			|	type_qualifier
			) >>
			!declaration_specifiers
		;

	init_declarator_list
		= init_declarator >> *(COMMA >> init_declarator)
		;

	init_declarator
		= declarator >> !(ASSIGN >> initializer)
		;

	storage_class_specifier
		= TYPEDEF
		| EXTERN
		| STATIC
		| AUTO
		| REGISTER
		;

	type_specifier
		= VOID
		| CHAR
		| SHORT
		| INT
		| LONG
		| FLOAT
		| DOUBLE
		| SIGNED
		| UNSIGNED
		| struct_or_union_specifier
		| enum_specifier
//		| TYPE_NAME
		;

	struct_or_union_specifier
		= struct_or_union >>
			(
				IDENTIFIER >> !(LEFT_BRACE >> struct_declaration_list >> RIGHT_BRACE)
			|	LEFT_BRACE >> struct_declaration_list >> RIGHT_BRACE
			)
		;

	struct_or_union
		= STRUCT
		| UNION
		;

	struct_declaration_list
		= +struct_declaration
		;

	struct_declaration
		= specifier_qualifier_list >> struct_declarator_list >> SEMICOLON
		;

	specifier_qualifier_list
		=	(
				type_specifier
			|	type_qualifier
			) >>
			!specifier_qualifier_list
		;

	struct_declarator_list
		= struct_declarator >> *(COMMA >> struct_declarator)
		;

	struct_declarator
		= declarator || (COLON >> constant_expression)
		;

	enum_specifier
		= ENUM >> !IDENTIFIER >> LEFT_BRACE >> enumerator_list >> RIGHT_BRACE
		;

	enumerator_list
		= enumerator >> *(COMMA >> enumerator)
		;

	enumerator
		= IDENTIFIER >> !(ASSIGN >> constant_expression)
		;

	type_qualifier
		= CONST
		| VOLATILE
		;

	declarator
		= !pointer >> direct_declarator
		;

// left recursion --> right recursion
//	direct_declarator
//		= IDENTIFIER
//		| LEFT_PAREN >> declarator >> RIGHT_PAREN
//		| direct_declarator >>
//			(
//				LEFT_BRACKET >> !constant_expression >> RIGHT_BRACKET
//			|	LEFT_PAREN >>
//				(
//					parameter_type_list >> RIGHT_PAREN
//				|	identifier_list >> RIGHT_PAREN
//				|	RIGHT_PAREN
//				)
//			)
//		;
	direct_declarator
		=	(
				IDENTIFIER
			|	LEFT_PAREN >> declarator >> RIGHT_PAREN
			) >>
			direct_declarator_helper
		;

	direct_declarator_helper
		=	(
				LEFT_BRACKET >> !constant_expression >> RIGHT_BRACKET
			|	LEFT_PAREN >>
				!(
					parameter_type_list
				|	identifier_list
				) >> RIGHT_PAREN
			) >> direct_declarator_helper
		| epsilon
		;

	pointer
		= STAR >> !(type_qualifier_list || pointer)
		;

	type_qualifier_list
		= +type_qualifier
		;

	parameter_type_list
		= parameter_list >> !(COMMA >> ELLIPSIS)
		;

	parameter_list
		= parameter_declaration >> *(COMMA >> parameter_declaration)
		;

	parameter_declaration
		= declaration_specifiers >>
			!(
				declarator
			|	abstract_declarator
			)
		;

	identifier_list
		= IDENTIFIER >> *(COMMA >> IDENTIFIER)
		;

	type_name
		= specifier_qualifier_list >> !abstract_declarator
		;

	abstract_declarator
		= pointer || direct_abstract_declarator
		;

// left recursion --> right recursion
//	direct_abstract_declarator
//		= LEFT_PAREN >>
//			(
//				abstract_declarator >> RIGHT_PAREN
//			|	!parameter_type_list >> RIGHT_PAREN
//			)
//		| LEFT_BRACKET >> !constant_expression >> RIGHT_BRACKET
//		| direct_abstract_declarator >>
//			(
//				LEFT_BRACKET >> !constant_expression >> RIGHT_BRACKET
//			|	LEFT_PAREN >> !parameter_type_list >> RIGHT_PAREN
//			)
//		;
	direct_abstract_declarator
		=	(
				LEFT_PAREN >>
				(
					abstract_declarator >> RIGHT_PAREN
				|	!parameter_type_list >> RIGHT_PAREN
				)
			|	LEFT_BRACKET >> !constant_expression >> RIGHT_BRACKET
			) >>
			direct_abstract_declarator_helper
		;

	direct_abstract_declarator_helper
		= 	(
				LEFT_BRACKET >> !constant_expression >> RIGHT_BRACKET
			|	LEFT_PAREN >> !parameter_type_list >> RIGHT_PAREN
			) >>
			direct_abstract_declarator_helper
		| epsilon
		;

	initializer
		= assignment_expression
		| LEFT_BRACE >> initializer_list >> !COMMA >> RIGHT_BRACE
		;

	initializer_list
		= initializer >> *(COMMA >> initializer)
		;

	statement
		= labeled_statement
		| compound_statement
		| expression_statement
		| selection_statement
		| iteration_statement
		| jump_statement
		;

	labeled_statement
		= IDENTIFIER >> COLON >> statement
		| CASE >> constant_expression >> COLON >> statement
		| DEFAULT >> COLON >> statement
		;

	compound_statement
		= LEFT_BRACE >> !(declaration_list || statement_list) >> RIGHT_BRACE
		;

	declaration_list
		= +declaration
		;

	statement_list
		= +statement
		;

	expression_statement
		= !expression >> SEMICOLON
		;

	selection_statement
		= IF >> LEFT_PAREN >> expression >> RIGHT_PAREN >> statement >> !(ELSE >> statement)
		| SWITCH >> LEFT_PAREN >> expression >> RIGHT_PAREN >> statement
		;

	iteration_statement
		= WHILE >> LEFT_PAREN >> expression >> RIGHT_PAREN >>  statement
		| DO >> statement >> WHILE >> LEFT_PAREN >> expression >> RIGHT_PAREN >> SEMICOLON
		| FOR >> LEFT_PAREN >> expression_statement >> expression_statement >> !expression >> RIGHT_PAREN >> statement
		;

	jump_statement
		= GOTO >> IDENTIFIER >> SEMICOLON
		| CONTINUE >> SEMICOLON
		| BREAK >> SEMICOLON
		| RETURN >> !expression >> SEMICOLON
		;

// parser start symbol
	translation_unit
		= *external_declaration
		;

	external_declaration
		= function_definition
		| declaration
		;

	function_definition
		= !declaration_specifiers >> declarator >> !declaration_list >> compound_statement
		;

// main driver code
	if (2 == argc)
		parse(translation_unit, argv[1]);
	else
		cerr << "No filename given" << endl;

    return 0;
}
