///////////////////////////////////////////////////////////////////////////////
// This example shows:
// 1. Parsing of different comment styles
//		parsing C/C++-style comment
//		parsing C++-style comment
//		parsing PASCAL-style comment
// 2. Parsing tagged data with the help of the confix_parser
// 3. Parsing tagged data with the help of the confix_parser but the semantic
//	  action is directly attached to the body sequence parser

#include <string>

#include <boost/spirit/spirit.hpp>

#include <iostream>
#include <cassert>

///////////////////////////////////////////////////////////////////////////////
// used namespaces
using namespace std;
using namespace spirit;

///////////////////////////////////////////////////////////////////////////////
// actor called after successfully matching a single character
class actor_string
{
public:
	actor_string(std::string &rstr) :
		matched(rstr)
	{
	}

	void operator() (const char *pbegin, const char *pend) const
	{
		matched += std::string(pbegin, pend-pbegin);
	}

private:
	std::string &matched;
};

///////////////////////////////////////////////////////////////////////////////
// actor called after successfully matching a C++-comment
void actor_cpp (const char *pfirst, const char *plast)
{
	cout << "Parsing C++-comment" <<endl;
	cout << "Matched (" << plast-pfirst << ") characters: ";

char cbbuffer[128];

	strncpy(cbbuffer, pfirst, plast-pfirst);
	cbbuffer[plast-pfirst] = '\0';

	cout << "\"" << cbbuffer << "\"" << endl;
}

///////////////////////////////////////////////////////////////////////////////
// main entry point
int main ()
{
///////////////////////////////////////////////////////////////////////////////
// 1. Parsing different comment styles
//		parsing C/C++-style comment (non-nested!)
char const* pCComment = "/* This is a /* nested */ C-comment */";

rule<char const *> cpp_comment;

	cpp_comment
		= comment_p("/*", "*/")					// rule for C-comments
		| comment_p("//", ch_p('\r') || '\n')	// rule for C++ comments
		;

std::string comment_c;
parse_info<char const *> result = parse (pCComment, cpp_comment[actor_string(comment_c)]);

	if (result.match) {
		cout << "Parsed C/C++-comment successfully!" << endl;
		cout << "Matched (" << comment_c.size() << ") characters: ";
		cout << "\"" << comment_c << "\"" << endl;
	} else
		cout << "Failed to parse C/C++-comment!" << endl;
	cout << endl;

//		parsing C++-style comment
char const* pCPPComment = "// This is a C++-comment\n";
std::string comment_cpp;

	result = parse (pCPPComment, cpp_comment[&actor_cpp]);
	if (result.match)
		cout << "Parsed C++-comment successfully!" << endl;
	else
		cout << "Failed to parse C++-comment!" << endl;
	cout << endl;

//		parsing PASCAL-style comment (nested!)
char const* pPComment = "{ This is a (* nested *) PASCAL-comment }";

rule<char const *> pascal_comment;

	pascal_comment						// in PASCAL we have two comment styles
		= comment_nest_p('{', '}')		// both may be nested
		| comment_nest_p("(*", "*)")
		;

std::string comment_pascal;

	result = parse (pPComment, pascal_comment[actor_string(comment_pascal)]);
	if (result.match) {
		cout << "Parsed PASCAL-comment successfully!" << endl;
		cout << "Matched (" << comment_pascal.size() << ") characters: ";
		cout << "\"" << comment_pascal << "\"" << endl;
	} else
		cout << "Failed to parse PASCAL-comment!" << endl;
	cout << endl;

///////////////////////////////////////////////////////////////////////////////
// 2. Parsing tagged data with the help of the confix_parser
std::string tag, body;
rule<const char *> html_tag, body_text, tag_name;

	body_text
		= anychar
		;

	tag_name
		= +chset<>("a-z")
		;

	html_tag
		= confix_p (
			ch_p('<') >> tag_name[actor_string(tag)] >> '>',
			body_text[actor_string(body)],
			str_p("</") >> strlit<std::string &>(tag) >> '>'
		)
		;

char const* pTag = "<b>Body text</b>";

	result = parse (pTag, html_tag);
	if (result.match) {
		cout << "Parsed HTML snippet \"<b>Body text</b>\" successfully (with re-attached actor)!" << endl;
		cout << "Found tag (" << tag.size() << ") characters: ";
		cout << "\"" << tag << "\"" << endl;
		cout << "Found body (" << body.size() << ") characters: ";
		cout << "\"" << body << "\"" << endl;
	} else
		cout << "Failed to parse HTML snippet (with re-attached actor)!" << endl;
	cout << endl;

///////////////////////////////////////////////////////////////////////////////
// 3. Parsing tagged data with the help of the confix_parser but the semantic
//	  action is directly attached to the body sequence parser (see comment in
//	  utilities.hpp)
rule<const char *> html_tag_direct;

std::string tagdirect, bodydirect;

	html_tag_direct
		= confix_p.direct(
			ch_p('<') >> tag_name[actor_string(tagdirect)] >> '>',
			body_text[actor_string(bodydirect)],
			str_p("</") >> strlit<std::string &>(tagdirect) >> '>'
		)
		;

char const* pTagDirect = "<b>Body text</b>";

	result = parse (pTagDirect, html_tag_direct);
	if (result.match) {
		cout << "Parsed HTML snippet \"<b>Body text</b>\" successfully (with direct actor)!" << endl;
		cout << "Found tag (" << tagdirect.size() << ") characters: ";
		cout << "\"" << tagdirect << "\"" << endl;
		cout << "Found body (" << bodydirect.size() << ") characters: ";
		cout << "\"" << bodydirect << "\"" << endl;
	} else
		cout << "Failed to parse HTML snippet (with direct actor)!" << endl;
	cout << endl;

	return 0;
}
