///////////////////////////////////////////////////////////////////////////////
// This sample shows the complete encapsulation of a grammar and getting back
// of a parsing result value from this grammar.
//
// It shows further, how a grammar can use an embedded closure for various 
// parsing tasks.

#include <boost/spirit/spirit.hpp>
#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// used namespaces
using namespace spirit;
using namespace std;

///////////////////////////////////////////////////////////////////////////////
//
// This grammar class returns a complete C style string literal from possibly
// separated string literals. I.e. when the input stream looks like
//
//      "stringliteral1" ... (some whitespace) "stringliteral2" etc.
//
// the grammar returns the complete stringliteral as
//
//      "stringliteral1stringliteral2".
//
// The grammar in this example is configured to embed a closure containing two 
// string variables. The closure generated by the starting rule of the grammar
// 'string_literal' and is tightly associated with the parsing context of the 
// related grammar. 
//
// These closure members are accessible through the 'string_literal' variable. 
// In this sample the access to the closure variables is done through the 
// predefined closure_member's 'm0' (closure_member<0>) and 'm1' 
// (closure_member<1>), you can use your own, though.
//
// Closure parsers and grammars have a return value, which is by definition
// equivalent to the value of closure_member<0>. This value is accessible 
// outside the closure or grammar with the help of an attached action (shown
// below for the grammar and in the sample closure.cpp for a closure). If you 
// do not want to have a return type from grammars, you can specify 
// the type spirit::closure_null_type as the first template parameter.

struct StringLiteral : 
    public grammar<StringLiteral, string>
{
    template <typename IteratorT>
    struct definition
    {
        typedef attr_rule<closure<string, std::string> > attr_rule_t;

        attr_rule_t string_literal;
        rule<IteratorT> string_literal_part;

        definition()
        {
            string_literal_part = lexeme[
                    '"' 
                >>  (*(str_p("\\\"") | anychar - '"'))[string_literal(m1)] 
                >>  '"'
            ];

            string_literal = +(string_literal_part[*this]);
        }

        attr_rule_t const& start() const
        {
            return string_literal;
        }

        template <typename ItT>
        void operator()(ItT &, ItT const&) const
        {
            string_literal.set(
                m0, string_literal.get(m0) + string_literal.get(m1));
        }
    };
};

///////////////////////////////////////////////////////////////////////////////
// The following specialization of the template embed_trait is required, because
// the class Stringliteral::definition<> should not be copied around and it 
// should be saved by the library solely as a reference. 
// Note: This is work in progress and should be removed in the future.
namespace spirit {
    template <typename IteratorT>
    struct embed_trait<StringLiteral::definition<IteratorT> >
    {
        typedef StringLiteral::definition<IteratorT> const& type;
    };
}

///////////////////////////////////////////////////////////////////////////////
// main entry point
int main()
{
StringLiteral Lit;
std::string s;

char const *pbegin = "\"part1 \" \"part2 \"";

    if (parse (pbegin, Lit[ref(s)], space).full) {
        cout 
            << "Parsed string literal parts successfully! Matched: " 
            << "\'" << s << "\'"
            << endl;
    } else {
        cout 
            << "Parsing the input stream failed!"
            << endl;
    }
    return 0;
}

