///////////////////////////////////////////////////////////////////////////////
// This sample shows the usage of the list_p utility parser
//  1. parsing a simple ',' delimited list w/o item formatting
//  2. parsing a CSV list (comma separated values - strings, integers or reals)
//  3. parsing a token list (token separated values - strings, integers or reals)
//     with an action parser directly attached to the item part of the list_p
//     generated parser

#include <string>

#include <boost/spirit/spirit.hpp>

#include <iostream>
#include <cassert>

///////////////////////////////////////////////////////////////////////////////
// used namespaces
using namespace std;
using namespace spirit;

///////////////////////////////////////////////////////////////////////////////
// actor, attached to the list_p parser
class list_actor 
{
public:
    list_actor (std::vector<std::string> &vec_) : vec(vec_) {}

// The following operator() is called by the action parser generated by
// attaching this actor to a list_p generated list parser.
//
// Note: the iterator 'it' should be passed by reference to get the correct
// matched character count back from the parse function
// 
// Additional Note: The iterators in this function are input iterators not to 
// single matched characters but to hole matched list items (in this case 
// std::string)

    template <typename ActionIterT>
    void operator() (ActionIterT &it, ActionIterT const &end) const
    {
        for (/**/; it != end; ++it) 
            vec.push_back(*it);
    }

private:
    std::vector<std::string> &vec;
};

///////////////////////////////////////////////////////////////////////////////
// actor, attached to the item part of a list_p generated list parser
class item_actor 
{
public:
    item_actor (std::vector<std::string> &vec_) : vec(vec_) {}

// The following operator() is called by the action parser generated by
// attaching this actor to the 'item' part of the list_p generated list parser.
//
// Note: here the signature of the actor function must match the signature 
// required by the item parser not (as above) as required by the list parser.
// So the iterators will iterate over matched input characters.
    template <typename IteratorT>
    void operator() (IteratorT &first, IteratorT const &last) const
    {
        vec.push_back(std::string(first, last-first));
    }

private:
    std::vector<std::string> &vec;
};

///////////////////////////////////////////////////////////////////////////////
// actor, attached directly to the item part of a list parser generated 
// by 'list_p.direct'
class item_actor_direct 
{
public:
    item_actor_direct (std::vector<std::string> &vec_) : vec(vec_) 
    {
        vec.push_back(std::string());
    }

// The following operator() is called by the action parser generated by
// attaching this actor to the 'item' part of the list parser generated by
// 'list_p.direct()'.
    template <typename IteratorT>
    void operator() (IteratorT &first, IteratorT const &last) const
    {
        assert(vec.size() > 0);
        vec[vec.size()-1] += std::string(first, last-first);
        if ('<' == *first)
            vec.push_back(std::string());
    }

private:
    std::vector<std::string> &vec;
};

///////////////////////////////////////////////////////////////////////////////
// main entry point
int main ()
{
// 1. parsing a simple ',' delimited list w/o item formatting
char const *plist_wo_item = "element1,element2,element3";
rule<char const *> list_wo_item;
std::vector<std::string> vec_list;

    list_wo_item 
        = list_p[list_actor(vec_list)]
        ;

parse_info<char const *> result = parse (plist_wo_item, list_wo_item);

    cout << "-----------------------------------------------------------------" << endl;
    if (result.match) {
        cout 
            << "Parsing simple list" << endl 
            << "\t" << plist_wo_item << endl 
            << "Parsed successfully!" << endl << endl;

        if (result.full)
            cout << "Matched " << vec_list.size() << " list elements (full list): " << endl;
        else
            cout << "Matched " << vec_list.size() << " list elements: " << endl;

        cout << "Items got from the list parser:" << endl;
        for (std::vector<std::string>::iterator it = vec_list.begin(); it != vec_list.end(); ++it)
            cout << *it << endl;

    } else
        cout << "Failed to parse simple list!" << endl;

	cout << endl;

// 2. parsing a CSV list (comma separated values - strings, integers or reals)
char const *plist_csv = "\"string\",\"string with an embedded \\\"\",12345,0.12345e4";
rule<char const *> list_csv, list_csv_direct, list_csv_item;
std::vector<std::string> vec_item;

    vec_list.clear();
    
    list_csv_item
        = confix_p('\"', c_escape_ch_p, '\"')
        | longest[real_p | int_p]
        ;

    list_csv
        = list_p(list_csv_item[item_actor(vec_item)], ',')[list_actor(vec_list)]
        ;

    result = parse (plist_csv, list_csv);

    cout << "-----------------------------------------------------------------" << endl;
    if (result.match) {
        assert(vec_list.size() == vec_item.size());

        cout 
            << "Parsing CSV list (comma separated values) " << endl 
            << "\t" << plist_csv << endl 
            << "Parsed successfully!" << endl << endl;

        if (result.full)
            cout << "Matched " << vec_list.size() << " list elements (full list): " << endl;
        else
            cout << "Matched " << vec_list.size() << " list elements: " << endl;

        cout << "Items got from the list parser:" << endl;
        for (std::vector<std::string>::iterator itl = vec_list.begin(); itl != vec_list.end(); ++itl)
            cout << *itl << endl;

        cout << endl << "Items got directly from the item parser:" << endl;
        for (std::vector<std::string>::iterator it = vec_item.begin(); it != vec_item.end(); ++it)
            cout << *it << endl;

    } else
        cout << "Failed to parse CSV list!" << endl;

	cout << endl;

// 3. parsing a token list (token separated values - strings, integers or reals) 
// with an action parser directly attached to the item part of the list_p
// generated parser
char const *plist_csv_direct = "\"string\"<par>\"string with an embedded \\\"\"<par>12345<par>0.12345e4";
rule<char const *> list_csv_item_direct;

    vec_list.clear();
    vec_item.clear();

    list_csv_item_direct
        = anychar
        ;

// Note: the list parser is here generated through the list_p.direct() 
// generator function. This inhibits re-attachment of the item_actor_direct during
// parser construction (see: comment in utilities.hpp)
    list_csv_direct
        = list_p.direct(list_csv_item_direct[item_actor_direct(vec_item)], "<par>")[list_actor(vec_list)]
        ;
        
    result = parse (plist_csv_direct, list_csv_direct);

    cout << "-----------------------------------------------------------------" << endl;
    if (result.match) {
        assert(vec_list.size() == vec_item.size());

        cout 
            << "Parsing CSV list (comma separated values)" << endl
            << "The list parser was generated with 'list_p.direct()'" << endl 
            << "\t" << plist_csv << endl 
            << "Parsed successfully!" << endl << endl;

        if (result.full)
            cout << "Matched " << vec_list.size() << " list elements (full list): " << endl;
        else
            cout << "Matched " << vec_list.size() << " list elements: " << endl;

        cout << "Items got from the list parser:" << endl;
        for (std::vector<std::string>::iterator itl = vec_list.begin(); itl != vec_list.end(); ++itl)
            cout << *itl << endl;

        cout << endl << "Items got directly from the item parser:" << endl;
        for (std::vector<std::string>::iterator it = vec_item.begin(); it != vec_item.end(); ++it)
            cout << *it << endl;

    } else
        cout << "Failed to parse CSV list!" << endl;

	cout << endl;

    return 0;
}
