// This file contains the following samples:
// 1. parsing real numbers formatted in different ways (standard parsing traits - C-like reals)
// 2. parsing real numbers formatted in different ways (special parsing traits - German locale)
// 3. parsing real numbers formatted in different ways (more complex parsing traits)
// 4. integer parsers
// 5. some tests of int_p et.al. for extreme values

#include <boost/spirit/spirit.hpp>
#include <boost/spirit/static_parsers.hpp>

#include <iostream>
#include <assert.h>

///////////////////////////////////////////////////////////////////////////////
using namespace std;
using namespace spirit;

///////////////////////////////////////////////////////////////////////////////
struct ftor {

    typedef double arg_type; //  Required

    void
    operator () (arg_type const& n) const { cout << n << endl; }
};

void
foo(float n) { cout << n << endl; }

// integer parsers
///////////////////////////////////////////////////////////////////////////////
struct ftor_oct {

    typedef int arg_type; //  Required

    void
    operator () (arg_type const& n) const 
	{ 
		cout << "as octal: " << oct << n << endl; 
	}
};

void
foo_int(int n) 
{ 
	cout << "as decimal: " << n << endl; 
}

const char *as_bin (int bin)
{
static char buffer[42];
char *ptr = buffer;
char *firstdig = buffer;

	do {
		*ptr = bin % 2 + '0';
		bin /= 2;
		ptr++;
	} while (bin > 0);
	*ptr-- = '\0';

// reverse buffer
    do {
	char temp = *ptr;

        *ptr = *firstdig;
        *firstdig = temp;
        --ptr;
        ++firstdig;
    } while (firstdig < ptr);

	return buffer;
}

///////////////////////////////////////////////////////////////////////////////
// Helper functions for testing integer parsers
bool MatchInt10 (const char *str, int &n)
{
	char const* end = str + strlen(str);
	return int_parser_gen<10>()[ref(n)].parse(str, end);
}

bool MatchInt16 (const char *str, int &n)
{
	char const* end = str + strlen(str);
	return int_parser_gen<16>()[ref(n)].parse(str, end);
}

bool MatchUInt10 (const char *str, unsigned int &n)
{
	char const* end = str + strlen(str);
	return uint_parser_gen<10>()[ref(n)].parse(str, end);
}

bool MatchUInt16 (const char *str, unsigned int &n)
{
	char const* end = str + strlen(str);
	return uint_parser_gen<16>()[ref(n)].parse(str, end);
}

///////////////////////////////////////////////////////////////////////////////
// definition of "exp" as an external entity for use as template argument
extern const char exponent[] = "exp";

///////////////////////////////////////////////////////////////////////////////
int
main()
{
    cout << "/////////////////////////////////////////////////////////\n\n";
    cout << "\t\tNumeric actions...\n\n";
    cout << "/////////////////////////////////////////////////////////\n\n";

// float/double parsers
// numbers to scan (test maximum of possible formats)
	const struct scan_data {
		const char *toscan;
		const char *toscan2;
		const char *toscan3;
		double expected_result;
	} toscan[] = {
		{ "123456.789E6",	"123456,789E6",		"123456,789exp6",	123456.789E6 },
		{ "123456.78E-5",	"123.456,78E-5",	"123456,78exp-5",	123456.78E-5 },
		{ "123456",			"123456",			"123456",			123456 },
		{ "123456.",		"123456,",			"123456,",			123456 },
		{ "123456E6",		"123456E6",			"123456exp6",		123456e6 },
		{ "123456.E6",		"123456,E6",		"123456,exp6",		123456e6 },
		{ ".789",			",789",				",789",				.789 },
		{ ".789E6",			",789E6",			",789exp6",			.789e6 },
		{ "-123456.789E6",	"-123456,789E6",	"-123456,789exp6",	-123456.789E6 },
		{ "-123456.78E-5",	"-123.456,78E-5",	"-123456,78exp-5",	-123456.78E-5 },
		{ "-123456",		"-123456",			"-123456",			-123456 },
		{ "-123456.",		"-123456,",			"-123456,",			-123456 },
		{ "-123456E6",		"-123456E6",		"-123456exp6",		-123456e6 },
		{ "-123456.E-6",	"-123456,E-6",		"-123456,exp-6",	-123456e-6 },
		{ "-.789",			"-,789",			"-,789",			-.789 },
		{ "-.789E6",		"-,789E6",			"-,789exp6",		-.789e6 },
	};

///////////////////////////////////////////////////////////////////////////////
// 1. parsing real numbers formatted in different ways (standard parsing traits - C-like reals)
	cout << "Parsing real numbers formatted in different ways (standard parsing traits)" << endl;
	for (int loop = 0; loop < sizeof(toscan)/sizeof(toscan[0]); ++loop) 
	{
		char const* str = toscan[loop].toscan;
		char const* end = str + strlen(str);

		char const* scan1(str);
		real_p[&foo].parse(scan1, end);

		char const* scan2(str);
		real_p[ftor()].parse(scan2, end);

		long double n;
		char const* scan3 = str;
		
		if (real_p[ref(n)].parse(scan3, end) && n == toscan[loop].expected_result)
			cout << toscan[loop].toscan << " = " << n << endl;
		else
			cout << toscan[loop].toscan << " not parsed" << endl;
	}
	cout << endl;

///////////////////////////////////////////////////////////////////////////////
// 2. parsing real numbers formatted in different ways (special parsing traits - German locale)
	cout << "Parsing real numbers formatted in different ways (special parsing traits)" << endl;

// Parses real numbers with ',' as decimal point, '.' as thousands delimiter
	typedef numeric_parser_traits<',', '.'> german_traits;

	for (int loop2 = 0; loop2 < sizeof(toscan)/sizeof(toscan[0]); ++loop2) 
	{
		char const* str = toscan[loop2].toscan2;
		char const* end = str + strlen(str);

		char const* scan1(str);
		real_p(german_traits())[&foo].parse(scan1, end);

		char const* scan2(str);
		real_p(german_traits())[ftor()].parse(scan2, end);

		long double n;
		char const* scan3 = str;
		
		if (real_p(german_traits())[ref(n)].parse(scan3, end) && n == toscan[loop2].expected_result)
			cout << toscan[loop2].toscan2 << " = " << n << endl;
		else
			cout << toscan[loop2].toscan2 << " not parsed" << endl;
	}
	cout << endl;

///////////////////////////////////////////////////////////////////////////////
// 3. parsing real numbers formatted in different ways (more complex parsing traits)
	cout << "Parsing real numbers formatted in different ways (more complex parsing traits)" << endl;

// Parses real numbers with ',' as decimal point, no thousands delimiter and 
// "exp" as exponent marker
	typedef numeric_parser_traits_ex<s_chlit<','>, s_nothing, s_strlit<exponent> > complex_traits;

	for (int loop3 = 0; loop3 < sizeof(toscan)/sizeof(toscan[0]); ++loop3) 
	{
		char const* str = toscan[loop3].toscan3;
		char const* end = str + strlen(str);

		char const* scan1(str);
		real_p(complex_traits())[&foo].parse(scan1, end);

		char const* scan2(str);
		real_p(complex_traits())[ftor()].parse(scan2, end);

		long double n;
		char const* scan3 = str;
		
		if (real_p(complex_traits())[ref(n)].parse(scan3, end) && n == toscan[loop3].expected_result)
			cout << toscan[loop3].toscan3 << " = " << n << endl;
		else
			cout << toscan[loop3].toscan3 << " not parsed" << endl;
	}
	cout << endl;

///////////////////////////////////////////////////////////////////////////////
// 4. integer parsers
	{
		char const* str = "101238A";
		char const* end = str + strlen(str);

		cout << "Parsing string: " << str << endl;

		char const* scan1(str);
		if (!int_p[&foo_int].parse(scan1, end))
			cout << "as decimal: not matched" << endl;

		char const* scan2(str);
		if (!oct_p[ftor_oct()].parse(scan2, end))
			cout << "as octal: not matched" << endl;

		int n;
		char const* scan3(str);
		if (hex_p[ref(n)].parse(scan3, end))
			cout << "as hexadecimal: " << hex << n << endl;
		else
			cout << "as hexadecimal: not matched" << endl;

		int bin;
		char const* scan4(str);
		if (bin_p[ref(bin)].parse(scan4, end))
			cout << "as dual: " << as_bin(bin) << endl;
		else
			cout << "as dual: not matched" << endl;
	}	

///////////////////////////////////////////////////////////////////////////////
// 5. some tests of int_p et.al. for extreme values
int i;

	assert(MatchInt10("2147483647", i) && 2147483647 == i);
	assert(MatchInt10("-2147483648", i) && -2147483647-1 == i);
	assert(!MatchInt10("2147483648", i));
	assert(!MatchInt10("-2147483649", i));

	assert(MatchInt16("7FFFFFFF", i) && 0x7FFFFFFF == i);
	assert(!MatchInt16("80000000", i));

// some tests of uint_p et.al. for extreme values
unsigned int ui;

	assert(MatchUInt10("4294967295", ui) && ui == 4294967295UL);
	assert(!MatchUInt10("4294967296", ui));

	assert(MatchUInt16("FFFFFFFF", ui) && 0xFFFFFFFF == ui);
	assert(!MatchUInt16("100000000", ui));
    return 0;
}
