///////////////////////////////////////////////////////////////////////////////
// This sample show the usage of parser parameters.
// 
// Parser parameters are used to pass some values from the outer parsing scope
// to the next inner scope. They can be imagined as the opposite to the return
// value paradigm, which returns some value from the inner scope to the next 
// outer.

#define SPIRIT_DEBUG

#include <boost/spirit/spirit.hpp>
#include <iostream>

#ifdef SPIRIT_DEBUG
// REVIEW: should we add this file from the boost distribution to spirit?
// copied from boost/libs/tuple/src/tuple.cpp
namespace boost {
namespace tuples {
namespace detail {

const int format_info::stream_index[number_of_manipulators] 
   = { std::ios::xalloc(), std::ios::xalloc(), std::ios::xalloc() };

}
}
}
#endif

///////////////////////////////////////////////////////////////////////////////
// used namespaces
using namespace spirit;
using namespace std;

///////////////////////////////////////////////////////////////////////////////
// Helper class for encapsulation of the type for the parsed variable names
class declaration_type
{
public:
    enum vartype {
        vartype_unknown = 0,        // unknown variable type
        vartype_int = 1,            // 'int'
        vartype_real = 2,           // 'real'
    };

    declaration_type() : type(vartype_unknown) 
    {
    }
    declaration_type(string const &type_) : type(vartype_unknown)
    {
        init(type_);
    }
    template <typename ItT>
    declaration_type(ItT const &first, ItT const &last) 
    {
        init(string(first, last-first-1));
    }
    declaration_type(declaration_type const &type_) : type(type_.type) 
    {
    }

// assignment of type
//    declaration_type &operator= (string const &type_)
//    {
//        init(type_);
//        return *this;
//    }

// access to the variable type
    //operator vartype &() { return type; }
    operator vartype const &() const { return type; }
    operator string ()
    {
        switch(type) {
        default:
        case vartype_unknown:   break;
        case vartype_int:       return string("int");
        case vartype_real:      return string("real");
        }
        return string ("unknown");
    }

    void swap(declaration_type &s) { std::swap(type, s.type); }

protected:
    void init (string const &type_)
    {
        if (type_ == "int")
            type = vartype_int;
        else if (type_ == "real")
            type = vartype_real;
        else
            type = vartype_unknown;
    }

private:
    vartype type;
};

///////////////////////////////////////////////////////////////////////////////
// Helper class for inserting an item with data into a symbol table
template <typename T, typename InitT>
class symbols_with_data 
{
public:
    typedef symbol_inserter<T, tst<T, char> > symbol_inserter_t;

    symbols_with_data(symbol_inserter_t const &add_, InitT const &data_) : 
        add(add_), data(data_) 
    {
    }

    template <typename IteratorT>
    T *operator()(IteratorT first, IteratorT const &last) const
    {
        return add(first, last, data);
    }

private:
    InitT data;
    symbol_inserter_t const &add;
};

template <typename T, typename CharT, typename InitT>
inline
symbols_with_data<T, InitT>
symbols_gen(symbol_inserter<T, tst<T, CharT> > const &add_, InitT const &data_)
{
    return symbols_with_data<T, InitT>(add_, data_);
}

///////////////////////////////////////////////////////////////////////////////
// The VarDeclList grammar parses variable declaration list 
struct VarDeclList : 
    public grammar<VarDeclList, declaration_type>
{
    typedef closure<declaration_type> closure_t;

    template <typename IteratorT>
    struct definition
    {
        symbols<declaration_type> symtab;
        attr_rule<closure_t> decl;
        attr_rule<closure_t> type;
        attr_rule<closure_t> list;
        attr_rule<closure_t> ident;

        definition()
        {
        // This must preceed all rule definitions to allow propagation of the
        // rule name to the attached action parsers.
            SPIRIT_DEBUG_RULE(decl);
            SPIRIT_DEBUG_RULE(list);
            SPIRIT_DEBUG_RULE(ident);
            SPIRIT_DEBUG_RULE(type);

        // pass variable type returned from 'type' to list closure member 0
	        decl = type[decl(m0)] >> list.init(decl(m0));  

        // m0 to access arg 0 of list --> passing variable type down to ident
	        list = ident.init(list(m0)) >> *(',' >> ident.init(list(m0))); 

        // store identifier and type into the symbol table
	        ident = (*alnum)[symbols_gen(symtab.add, ident(m0))];

        // the type of the decl is returned in type's closure member 0
	        type = str_p("int")[type(m0)] | str_p("real")[type(m0)];
        }

        attr_rule<closure_t> const& start() const
        {
            return decl;
        }
    };
};

///////////////////////////////////////////////////////////////////////////////
// main entry point
int main()
{
VarDeclList Decl;
declaration_type type;
char const *pbegin = "int var1";

//symbols<declaration_type> symtab;
//attr_rule<closure<declaration_type, string> > Decl;
//
//    Decl = (*alnum)[symbols_with_data<declaration_type>(symtab.add, Decl(m0))];

    if (parse (pbegin, Decl[ref(type)], space).full) {
        cout << endl
             << "Parsed variable declarations successfully!" << endl
             << "Detected type: " << declaration_type::vartype(type) << " (" << string(type) << ")"
             << endl;
    } else {
        cout << endl
             << "Parsing the input stream failed!"
             << endl;
    }
    return 0;
}


